package cbsw;

import java.nio.file.Paths;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.net.URL;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import java.io.*;
import cbsw.*;

/**
<p>A collection of miscellaneous methods</p>
 */
public class Cbsw
{
  public Cbsw(){}

/**
<p>JAXB unmarshaller</p>
<p>Unmarshal classes from a repository.  Unmarshalling creates objects
and initialises members of a class instance structure from an XML
file.</p>
<p>The URL may point to a file on the web or a local file using the
usual URL syntax.  The file is cached in the current directory with
prefix www- and is used automatically if the remote connection
fails.</p>
<p>To prevent overwriting a previously cached file, remove the www- prefix.
The unmarshalling always first tries to use the file with no prefix.</p>
@param xml URL of XML data
@param classes The classes that provide the unmarshalling context
@return The unmarshalled object
*/
  public static Object unmarshal(String xml,Class... classes)
  {
    Object obj = null;
      
    try
    {
      URL url = new URL(xml);
      String [] pathTokens = url.getPath().split("/");
      String fileName = pathTokens[pathTokens.length-1];

// Look for file in current directory and use that otherwise copy from URL.
      File file = new File(fileName);
      if (!file.canRead())
      {
        InputStream is = url.openStream();
        Files.copy(is,Paths.get("www-"+fileName),StandardCopyOption.REPLACE_EXISTING);
        file = new File("www-"+fileName);
      }

      JAXBContext jaxbContext = JAXBContext.newInstance(classes);
      Unmarshaller jaxbUnmarshaller = jaxbContext.createUnmarshaller();
      jaxbUnmarshaller.setEventHandler(new javax.xml.bind.helpers.DefaultValidationEventHandler());
      System.out.println("Unmarshalling configuration from "+file.getAbsolutePath());
      obj = jaxbUnmarshaller.unmarshal(file);
    }
    catch (JAXBException e) { e.printStackTrace(); System.exit(0);  }
    catch ( IOException ioe ){ System.err.println("Cbsw.unmarshal() Warning: " + ioe.getMessage() ); }

    return obj;
  }

/**
<p>Calls a method using Java reflection.</p>
<p>Supports only the following method signatures:</p>
<code>
static void method()
static void method(String)
static void method(String,String)
</code>
@param cl The class
@param cmd The method and parameters cmd[[|arg1[|arg2]]
*/
  public static void exec(Class<?> cl, String cmd)
  {
    System.out.println("Cbsw.exec() ["+cmd+"]");
    String [] cmdTokens = cmd.split("\\|"); // Have to escape the vert bar in regex

    try
    {
      if (cmdTokens.length == 1)
      {
        java.lang.reflect.Method method = cl.getMethod(cmdTokens[0]);
        method.invoke(cl);
      }
      else if (cmdTokens.length == 2)
      {
        java.lang.reflect.Method method = cl.getMethod(cmdTokens[0],String.class);
        method.invoke(cl,cmdTokens[1]);
      }
      else if (cmdTokens.length == 3)
      {
        java.lang.reflect.Method method = cl.getMethod(cmdTokens[0],String.class,String.class);
        method.invoke(cl,cmdTokens[1],cmdTokens[2]);
      }
    }
    catch (Exception e) {System.out.println("Cbsw.exec(): "+e.toString()+" for "+cl.toString()+"["+cmd+"]");}
  }

/**
<p>Launch a JFileChooser to choose a file</p>
@param dir The directory to navigate from
@param filter A file filter
@return The File
*/
  public static File chooseFile(String dir, javax.swing.filechooser.FileFilter filter)
  {
    javax.swing.JFileChooser chooser = new javax.swing.JFileChooser(dir);
    chooser.setFileFilter(filter);
    int returnVal = chooser.showOpenDialog(null);
    if(returnVal != javax.swing.JFileChooser.APPROVE_OPTION) return null;
    File file = chooser.getSelectedFile();
    if (!file.isFile()) return null;
    return file;
  }
}
