package cbsw;

import java.util.prefs.Preferences;
import javax.xml.bind.annotation.*;
import java.io.*;
import java.util.*;

/**
<h1>Bit-oriented register descriptor collection</h1>
<p>Hardware interfaces are typically manipulated through a set of
addressable byte-oriented registers. This class is intended to
simplify bit-oriented access to such registers. It encapsulates a
collection of descriptors for fields containing consecutive bits. Each
field is assigned a key, a bit offset from the start of the register
array and the length of consecutive bits belonging to the
field. Internally, the class maintains a copy of the current register
settings as a BitSet. Application code can manipulate the contents of
the cache using get() and set() methods. Writing the hardware is
achieved through the put() methods using the RegisterIo interface.</p>
<p>The class unmarshalls the collection of bitfields from an XML
file. An example bit-field descriptor might appear like this in the
XML:</p>
<code>
&lt;register name="gt.0.precursor" offset="536" length="5"&gt;GT TX precursor value&lt;/register&gt;
</code>
<p>The name attribute is used as the HashMap key.</p>
<p>Several of the methods are generic. Where the generic vaue
represents a bit-field value, the types Integer, Boolean or String are
allowed. All others result in the method throwing a
RuntimeException.</p>
<p>To support triplicated registers (such as used by the GBTX ASIC),
the offset attribute may be a comma-separated list of (usually 3)
offsets. The bit-field values are replicated to each offset in the
list.</p>
@see RegisterDescriptor
@see RegisterIo
 */
@XmlAccessorType(XmlAccessType.NONE)
@XmlRootElement(name="csr")
public class Csr
{
BitSet csrBits;
@XmlAttribute
int baseAddress;
@XmlAttribute
int csrSize;
@XmlElement(name="register")
ArrayList<RegisterDescriptor> descriptors;
HashMap<String,RegisterDescriptor> map;

/**
<p>Returns the bit-field descriptor list</p>
@return The list
*/
  public ArrayList<RegisterDescriptor> getDescriptors(){return descriptors;}

/**
<p>No-argument constructor called by unmarshaller.</p>
*/
  public Csr(){csrBits = new BitSet();}

/**
<p>Called after unmarshalling to complete the initialisation.</p>
<p>Initialises the HashMap used to index the ResisterDescriptors.</p>
*/
  protected void init()
  {
    map = new HashMap<String,RegisterDescriptor>();
    for (RegisterDescriptor reg : descriptors){ map.put(reg.getName(),reg); }
  }

/**
<p>Unmarshalls register descriptors into members of this Csr class.</p>
@param xml URL of unmarshalling XML data
@return A new instance of Csr
*/
  public static Csr unmarshall(String xml)
  {
    Csr csr = (Csr)Cbsw.unmarshal(xml,Csr.class);
    csr.init();  
    return csr;
  }

/**
<p>Gets the register base address</p>
@return The address
*/
  public int getBaseAddress(){return baseAddress;}
/**
<p>Gets the cached BitSet</p>
@return The BitSet
*/
  public BitSet getCsrBits(){return csrBits;}

/**
<p>Returns a Collection stream.</p>
@return The stream
 */
public java.util.stream.Stream<RegisterDescriptor> stream(){return descriptors.stream();}

/**
<p>Gets hex string representation of bit-field</p>
@param key The bit-field key
@return String containg hex representation of bit-field
 */
  public String get( String key ) { return map.get(key).get(csrBits); }
/**
<p>Gets hex string representation of bit-field</p>
<p>This variant allows a bit-field to be retrieved from a BitSet that is not necessarily the internally cached version.</p>
@param key The bit-field key
@param bits A BitSet containing the register values
@return String containg hex representation of bit-field
 */
  public String get( String key, BitSet bits ){ return map.get(key).get(bits); }
/**
<p>Gets hex string representation of bit-field</p>
<p>This variant allows a bit-field to be retrieved from a byte array
that is assumed to have the same shape as the register array described
by this Csr.</p>
@param key The bit-field key
@param regs A byte array containing the register values
@return String containg hex representation of bit-field
 */
  public String get( String key, byte [] regs )
  {
    BitSet bits = BitSet.valueOf(regs);
    return map.get(key).get(bits);
  }

/**
<p>Initialises the Csr from binary file</p>
<p>The BitSet cache is initialised from the binary data. The previous
BitSet, if any, is discarded. The new values are not written to the
hardware.</p>
@param cf A File containing the binary data
 */
  public void set(File cf)
  {
    byte [] buf = new byte [csrSize];

    try
    {
      FileInputStream fi = new FileInputStream(cf);
      fi.read( buf );
      fi.close();
    }
    catch (IOException ioe) { System.out.println("Csr.set(File): IOException"); System.out.println(ioe.getMessage()); }

    csrBits = BitSet.valueOf(buf);

    return;
  }

/**
<p>Converts value into String representation</p>
<p>Types String, Integer and Boolean are allowed. Throws a RuntimeException for any other type.</p>
@param obj The value
@return The hex String representation
 */
static public String valueOf(Object obj)
{
  if (obj instanceof Boolean) return (Boolean)obj ? "1" : "0";
  else if (obj instanceof Integer) return String.format("%X",(Integer)obj);
  else if (obj instanceof String) return (String)obj;
  else throw new RuntimeException("Csr.valueOf: Unexpected type "+obj.getClass().getName());
}

/**
<p>Converts a value from a Preference into String representation</p>
<p>Types String, Integer and Boolean are allowed. Throws a RuntimeException for any other type.</p>
@param key The key in the Preferences node
@param prefs The Preferences node
@param defaultValue The default value if the Preference does not exist
@return The hex String representation
*/
protected String getPrefsValue(String key, Preferences prefs, Object defaultValue)
{
  if (defaultValue instanceof Boolean) return prefs.getBoolean(key,(Boolean)defaultValue) ? "1" : "0";
  else if (defaultValue instanceof Integer) return String.format("%X",prefs.getInt(key,(Integer)defaultValue));
  else if (defaultValue instanceof String) return prefs.get(key,(String)defaultValue);
  else throw new RuntimeException("Csr.getPrefsValue: Unexpected type "+defaultValue.getClass().getName());
}

/**
<p>Sets bit-field</p>
@param <T> Type of the value to set
@param key The bit-field key
@param value The value to set
 */
  public <T> void set( String key, T value ) { map.get(key).set(csrBits,valueOf(value)); }

/**
<p>Sets the register bits from value stored in Java Preference</p>
<p>This is a convenience method for when the bit-field should be set according to a value stored as a Java Preference.</p>
<p>key is both the XML key and the preference key name</p>
@param key The bit-field key
@param prefs The Preferences node
@deprecated This call always uses "0" as the default for the preference value. This is almost never what you want.
 */
@Deprecated
  public void set( String key, Preferences prefs) { map.get(key).set(csrBits,prefs.get(key,"0")); }

/**
<p>Sets the register bits from value stored in Java Preference</p>
<p>This is a convenience method for when the bit-field should be set according to a value stored as a Java Preference.</p>
<p>key is both the XML key and the preference key name</p>
<p>The default value is applied if no Preference is found.</p>
@param <T> Type of the value to set
@param key The bit-field key
@param prefs The Preferences node
@param defaultValue Default value
 */
public <T> void set( String key, Preferences prefs, T defaultValue) { map.get(key).set(csrBits,getPrefsValue(key,prefs,defaultValue)); }

/**
<p>Returns a new byte array containing the register values</p>
<p>The method creates an array of bytes containing the cached register
values starting arrayOffset bytes from the start of the register
array.</p>
@param arrayOffset Offset from register array start
@return The register values
 */
  public byte [] toByteArray(int arrayOffset)
  {
    byte [] buf = new byte [arrayOffset+csrSize];
    System.arraycopy(csrBits.toByteArray(),0,buf,arrayOffset,csrBits.toByteArray().length);
    return buf;
  }

/**
<p>Returns the RegisterDescriptor for the given key</p>
@param key The bit-field key
@return The register descriptor
 */
  public RegisterDescriptor getRegisterDescriptor(String key){ return map.get(key); }

/**
<p>Returns a new byte array containing the value of the specified bit-field</p>
<p>The method creates an array of bytes containing the cached
bit-field value. The value is offset arrayOffset bytes from the start
of the returned array.</p>
@param key The bit-field key
@param arrayOffset Offset from register array start
@return The register values
 */
@Deprecated
  public byte [] getBytes(String key, int arrayOffset){ return map.get(key).getBytes(csrBits,arrayOffset); }
    
/**
<p>Displays the values of all bit-fields on System.out</p>
 */
  public void decode()
  {
    if (csrBits==null) return;
    for (RegisterDescriptor reg : descriptors){ System.out.println(reg.getName()+": "+reg.get(csrBits)); } 
    return;
  }

/**
<p>Initialise the bitset from hardware</p>
<p>Any existing internally cached BitSet is replaced by this operation.</p>
@param io The hardware I/O interface
@throws cbsw.RegisterIoException RegisterIo exception
 */
  public void cache(RegisterIo io) throws cbsw.RegisterIoException
  {
    try
    {
      byte [] rdBuf = new byte [csrSize];
      io.get(baseAddress,0,rdBuf);
      csrBits = BitSet.valueOf(rdBuf);
    }
    catch ( RegisterIoException e ){ throw e; }
  }
    
/**
<p>Writes bytes corresponding to bit-field to hardware</p>
@param name The bit-field key
@param io The hardware I/O interface
@throws cbsw.RegisterIoException RegisterIo exception
 */
  public void put(String name, RegisterIo io) throws cbsw.RegisterIoException
  {
    try
    {
      RegisterDescriptor reg = map.get(name);
      String [] offsets = reg.offset.split(",");
      for (String offsetString : offsets)
      {
        byte [] regBuf = reg.getBytes(csrBits);
        java.nio.ByteBuffer bb = java.nio.ByteBuffer.allocate(4+regBuf.length);
        bb.order(java.nio.ByteOrder.LITTLE_ENDIAN);
        bb.putShort((short)regBuf.length);
        bb.putShort((short)(Short.valueOf(offsetString)>>3));
        bb.put(regBuf);
        bb.flip();
        io.put(baseAddress,bb);
      }
    }
    catch ( Exception ex ){ throw ex; }
  }
    
/**
<p>Writes new bit-field value to hardware</p>
<p>Updates the internal BitSet and writes to the hardware</p>
@param <T> The type of the value to write
@param name The bit-field key
@param val The value to write into the bit-field
@param io The hardware I/O interface
@throws cbsw.RegisterIoException RegisterIo exception
 */
public <T> void put(String name, T val, RegisterIo io) throws cbsw.RegisterIoException
  {
    try
    {
      RegisterDescriptor reg = map.get(name);
      reg.set(csrBits,valueOf(val));
      String [] offsets = reg.offset.split(",");
      for (String offsetString : offsets)
      {
        byte [] regBuf = reg.getBytes(csrBits);
        java.nio.ByteBuffer bb = java.nio.ByteBuffer.allocate(4+regBuf.length);
        bb.order(java.nio.ByteOrder.LITTLE_ENDIAN);
        bb.putShort((short)regBuf.length);
        bb.putShort((short)(Short.valueOf(offsetString)>>3));
        bb.put(regBuf);
        bb.flip();
        io.put(baseAddress,bb);
      }
    }
    catch ( RegisterIoException e ){ throw e; }
  }
    
/**
<p>Writes new bit-field value to hardware</p>
<p>Updates the internal BitSet and writes to the hardware. Gets the
value from Preferences node and applies default value if Preference not
found.</p>
@param <T> The type of the value to write
@param name The bit-field key
@param prefs The Preferences node
@param defaultValue The default value of the bit-field
@param io The hardware I/O interface
@throws cbsw.RegisterIoException RegisterIo exception
 */
  public <T> void put(String name, Preferences prefs, T defaultValue, RegisterIo io) throws cbsw.RegisterIoException
  {
    try { put(name,getPrefsValue(name,prefs,defaultValue),io); }
    catch ( RegisterIoException ex ){ throw ex; }
  }
    
/**
<p>A simple demo how to create a Csr</p>
@param args Command line parameters
*/
  public static void main( String[] args )
  {
    Csr csr = (Csr)Cbsw.unmarshal("https://hep.phy.cam.ac.uk/lhcb/Chimaera2/xml/tbtrigger.xml");
    csr.init();
    for (RegisterDescriptor reg : csr.getDescriptors()){ System.out.println(reg.getName()); }  
  }

}

