package cbsw;

import java.util.BitSet;
import java.util.Arrays;
import javax.xml.bind.annotation.*;

/**
<p>A bit-field descriptor</p>
 */
@XmlAccessorType(XmlAccessType.NONE)
public class RegisterDescriptor
{
  @XmlValue
  protected String description;
  @XmlAttribute
  protected String name;
  @XmlAttribute
  protected String offset; // Can be comma-separated list
  @XmlAttribute
  protected int length;

/**
<p>No-argument constructor called by unmarshaller</p>
 */
  public RegisterDescriptor(){}

/**
<p>Returns the bit-field description</p>
@return The descriptive text
 */
  public String getDescription(){return description;}
/**
<p>Returns the bit-field name</p>
@return The name
 */
  public String getName(){return name;}
/**
<p>Returns the bit-field offset</p>
@return The offset
 */
  public String getOffset(){return offset;}
/**
<p>Returns the bit-field length</p>
@return The length
 */
  public int getLength(){return length;}

/*
<p>Returns an array of bytes spanning the register bits offset from start of new array</p>
@param regBits The register bits
@param arrayOffset Offset where to put data in returned array
@return The array of bytes containing the register values

  public byte [] getBytes(BitSet regBits, int arrayOffset)
  {
    int bitLow = offset&0xfffffff8;
    int bitHigh = ((offset+length-1)&0xfffffff8)+8;
    BitSet bits = regBits.get(bitLow,bitHigh);
    byte [] result = new byte [arrayOffset+(bitHigh-bitLow)/8];
    System.arraycopy(bits.toByteArray(),0,result,arrayOffset,bits.toByteArray().length);
    return result;
  }
 */
    
/**
<p>Returns an array of bytes spanning the register bits</p>
@param regBits The register bits
@return The array of bytes containing the register values
 */
  public byte [] getBytes(BitSet regBits)
  {
    String [] offsets = offset.split(",");
    return getBytes(regBits,Integer.valueOf(offsets[0]));
  }
    
/**
<p>Returns an array of bytes spanning the register bits</p>
@param regBits The register bits
@param offset The bit-field offset
@return The array of bytes containing the register values
 */
  public byte [] getBytes(BitSet regBits, int offset)
  {
    int bitLow = offset&0xfffffff8;
    int bitHigh = ((offset+length-1)&0xfffffff8)+8;
    BitSet bits = regBits.get(bitLow,bitHigh);
    byte [] result = Arrays.copyOfRange(bits.toByteArray(),0,(bitHigh-bitLow)/8);
    return result;
  }

/**
<p>Returns a hex String representation of the register bits</p>
@param regBits The register bits
@return Hex String representation of register bits
 */
  public String get(BitSet regBits)
  {
    int off = Integer.valueOf(offset.split(",")[0]);
    BitSet bits = regBits.get(off,off+length);
    byte [] result = Arrays.copyOfRange(bits.toByteArray(),0,(length+7)/8);
    String resultString = new String();
    for ( byte b: result ) resultString = String.format("%1$02x",b) + resultString;
    return resultString;
  }

/**
<p>Set the register bits for all offsets</p>
@param regBits The register bits
@param flag boolean representation of value to set
 */
  public void set(BitSet regBits, boolean flag)
  {
    if (length != 1) return;
    String [] offsets = offset.split(",");
    for (String offsetString : offsets)
    {
      int off = Integer.valueOf(offsetString);
      if (flag)
      {
        regBits.set(off);
      }
      else
      {
        regBits.clear(off);
      }
    }
  }

/**
<p>Set the register bits for all offsets</p>
@param regBits The register bits
@param regString Hex String representation of value to set
 */
  public void set(BitSet regBits, String regString)
  {
    String [] offsets = offset.split(",");
    for (String offsetString : offsets)
    {
      int off = Integer.valueOf(offsetString);
      int end = off + length;

      regBits.clear(off,end);

      for (int c=regString.length()-1;c>=0;c--)
      {
        if (off >= end) break;
        int reg = Integer.parseInt(regString.substring(c,c+1),16);
        for (int i=0;i<4;i++)
        {
	  if (off >= end) break;
          if ((reg>>i & 1) == 1) regBits.set(off);
	  off++;
        }
      }
    }
  }

/**
<p>Writes a register bit-field to the hardware</p>
@param bits The register bits
@param regValue Hex representation of value to write
@param io The IO interface
  */
public void put(BitSet bits, String regValue, Io io)
{
    System.err.println(name+" "+regValue);

    set(bits,regValue);

    for (String offsetString : offset.split(","))
    {
      int off = Integer.valueOf(offsetString);
      byte [] regBuf = getBytes(bits,off);
      io.put(off/8,regBuf);
    }
}

}
