package cbsw;

import java.io.*;
import java.nio.*;
import java.nio.channels.*;

/**
<p>A class that encapsulates a socket connection.</p>
@author Steve Wotton, Cavendish Lab (HEP), 2018-
**/
public class SocketTxRx
{
static final int bufferSize = 4096;
ByteBuffer rxBuffer;
ByteBuffer txBuffer;

ServerSocketChannel serverSocketChannel;
SocketChannel socketChannel;
ReadableByteChannel rxChannel;
WritableByteChannel txChannel;
    
public SocketTxRx(){}
{
  rxBuffer = ByteBuffer.allocate(bufferSize);
  rxBuffer.order(ByteOrder.LITTLE_ENDIAN);
  txBuffer = ByteBuffer.allocate(bufferSize);
  txBuffer.order(ByteOrder.LITTLE_ENDIAN);
}
    
/**
<p>Reads a message from the rx channel.</p>
@param timeout Socket timeout in milliseconds. 0 for no timeout.
@return The message string
 */
public String rx(int timeout)
{
  try
  {
    socketChannel.socket().setSoTimeout(timeout);
    rxBuffer.clear();
    rxChannel.read(rxBuffer);
    rxBuffer.flip();
    byte [] rdMsg = new byte[rxBuffer.limit()];
    rxBuffer.get(rdMsg);
    return new String(rdMsg);
  }
  catch (java.net.SocketTimeoutException st){return null;}
  catch (Exception ex) {throw new RuntimeException("SocketTxRx.rx(): "+ex.toString());}
}

/**
<p>Writes a message to the tx channel.</p>
@param message The message string
 */
public void tx(String message)
{

  try
  {
    txBuffer.clear();
    txBuffer.put(message.getBytes());
    txBuffer.flip();
    txChannel.write(txBuffer);
  }
  catch (Exception ex) {throw new RuntimeException("SocketTxRx.rx(): "+ex.toString());}
}

/**
<p>Cleans out any pending messages.</p>
<p>Reads socket until it times out indicating no more data.</p>
 */
public void rxFlush()
{
  while (true)
  {
     String msg = rx(500);
     if (msg == null) break;
     System.err.println("SocketTxRx.rxFlush(): "+msg+"\n");
  }
}

/**
<p>Creates a server socket channel.</p>
@param url A URL to pass the port number
 */
public void makeServerSocket(java.net.URL url)
{
  try
  {
    int port = url.getPort();
    if ( port < 1024 ) return; // Disallow privileged port numbers

    serverSocketChannel = ServerSocketChannel.open();
    serverSocketChannel.bind(new java.net.InetSocketAddress(port));
  }
  catch ( Exception ex ) { throw new RuntimeException( "SocketTxRx.makeServerSocket(): [" + url.toString() + "] " + ex.toString() ); }

  return;
}

/**
<p>Creates a socket channel.</p>
@param url A URL to pass the internet address
 */
public void makeSocket(java.net.URL url)
{
  try
  {
    int port = url.getPort();
    if ( port < 1024 ) return; // Disallow privileged ports

    socketChannel = SocketChannel.open(new java.net.InetSocketAddress(url.getHost(), port));

    rxChannel = Channels.newChannel(socketChannel.socket().getInputStream());
    txChannel = Channels.newChannel(socketChannel.socket().getOutputStream());
  }
  catch ( Exception ex ) { throw new RuntimeException( "SocketTxRx.makeSocket(): [" + url.toString() + "] " + ex.toString() ); }

  return;
}

/**
<p>Accepts a server socket connection and creates tx/rx channels.</p>
 */
public void accept()
{
  try
  {
    socketChannel = serverSocketChannel.accept();

    rxChannel = Channels.newChannel(socketChannel.socket().getInputStream());
    txChannel = Channels.newChannel(socketChannel.socket().getOutputStream());
  }
  catch ( Exception ex ) { throw new RuntimeException( "SocketTxRx.accept(): " + ex.toString() ); }
}
    
/**
<p>An example program that replies ACK_OK to any message.</p>
<p>Creates and listens on a server socket. When a connection is made,
responds with ACK_OK to all messages.</p>
<p>When the connection is broken, waits for a new connection.</p>
@param args Command-line parameters
*/
public static void main( String[] args )
{
  try
  {
    SocketTxRx tester = new SocketTxRx();
    tester.makeServerSocket(new java.net.URL("http://:7999"));
    while (true)
    {
      try
      {
	tester.accept();

        while (true)
        {
          String response = tester.rx(0);
	  int seq = Integer.valueOf(response.split("\\|")[0]);
	  System.err.println("<"+response+"\n");
	  tester.tx(seq+"|ACK_OK");
        }
      }
      catch ( Exception ex ) { System.err.println( "SocketTxRx.main(): [while] " + ex.toString() ); }  
    }
  }
  catch ( Exception ex ) { System.err.println( "SocketTxRx.main(): " + ex.toString() ); }
}

}
