package cbsw;

import java.util.*;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
<p>A utility to build bit sequences for a JTAG TAP</p>
@author Steve Wotton, Cavendish Lab (HEP), 2018-
 */
public class TapBuilder
{
  int tapLength;
  int tapPosition;
  StringBuilder tms;
  StringBuilder tdo;
  StringBuilder tdoMask; // Indicates location of TDO data
  boolean debug;

/**
<p>Constructor</p>
@param cl The TAP length (number of devices)
@param pos The position on the TAP of the target device
 */
  public TapBuilder(int cl, int pos){tapLength=cl; tapPosition=pos; debug=false;}
/**
<p>Sets the TAP length</p>
@param cl The TAP length (number of devices)
 */
  public void setTapLength(int cl){tapLength = cl;}
/**
<p>Sets the TAP position</p>
@param pos The position on the TAP of the target device
 */
  public void setTapPosition(int pos){tapPosition = pos;}
/**
<p>Sets the debug flag</p>
@param dbg true=Enable debug
 */
  public void setDebug(boolean dbg){debug=dbg;}

/**
<p>Returns the length of the sequence</p>
@return The length (-1 on error)
 */
  public int length()
  {
    int tmsLength = tms.length();
    int tdoLength = tdo.length();
    return (tmsLength == tdoLength) ? tmsLength : -1;
  }

/**
<p>Converts StringBuilder to ByteBuffer</p>
@param sb The StringBuilder
@return The ByteBuffer
*/
  public ByteBuffer toByteBuffer(StringBuilder sb)
  {
    ByteBuffer bb = ByteBuffer.allocate(32); // SW 2018-05-21 Increased from 16 
    bb.order(ByteOrder.LITTLE_ENDIAN);
    int i=sb.length()-8;
    for (;i>=0;i-=8)
    {
      bb.put((byte)(Short.parseShort(sb.substring(i,i+8),2)&0xff));
    }
    if (i>-8) bb.put((byte)(Short.parseShort(sb.substring(0,i+8),2)&0xff));

    // Pad to multiple of 4 bytes
    while (bb.position()%4 !=0) bb.put((byte)0);

    bb.position(0);

    return bb;
  }

/**
<p>Converts MASK to ByteBuffer</p>
@return The ByteBuffer
*/
  public ByteBuffer tdoMaskToByteBuffer(){ return toByteBuffer(tdoMask); }
/**
<p>Converts TDO to ByteBuffer</p>
@return The ByteBuffer
*/
  public ByteBuffer tdoToByteBuffer(){ return toByteBuffer(tdo); }
/**
<p>Convert TMS to ByteBuffer</p>
@return The ByteBuffer
*/
  public ByteBuffer tmsToByteBuffer(){ return toByteBuffer(tms); }


/**
<p>Converts TMS/TDO pairs to ByteBuffer (useful for DJTG API)</p>
@return The ByteBuffer
*/
  public ByteBuffer tmsTdoToByteBuffer()
  {
    StringBuilder tmstdo = new StringBuilder();

    int tdoLength = tdo.length();
    for (int i=0;i<tdoLength;i++)
    {
      tmstdo.append(tms.charAt(i));
      tmstdo.append(tdo.charAt(i));
    }

    if (debug) System.err.println("TMSTDO>"+tmstdo.toString()+">");

    return toByteBuffer(tmstdo);
  }

/**
<p>Clears the TMS, TDO and MASK sequences</p>
*/
  public void clear()
  {
    tms = new StringBuilder();
    tdo = new StringBuilder();
    tdoMask = new StringBuilder();
  }

/**
<p>Displays the TMS, TDO and MASK sequences on System.err</p>
*/
  public void show()
  {
    if (!debug) return;
    System.err.println("TMS>"+tms.toString()+">");
    System.err.println("TDO>"+tdo.toString()+">");
    System.err.println("MSK>"+tdoMask.toString()+">");
  }

/**
<p>IR scan StringBuffer builder</p>
<p>Insert sequence to do IR scan.</p>
<p>Valid from UPDATE-IR, UPDATE-DR or RTI</p>
@param ircode The instruction code
@param rti If true, insert transitions to RTI
*/
  public void ir(String ircode, boolean rti)
  {

      tms.insert(0,"0011");
      tdo.insert(0,"0000");
      tdoMask.insert(0,"0000");
      for (int i=0;i<tapLength;i++)
      {
	tms.insert(0,"000000");
	tdoMask.insert(0,"000000");
	if (i==tapPosition)
	{
	  tdo.insert(0,ircode); // BYPASS
	}
	else
	{
	  tdo.insert(0,"111111"); // BYPASS
	}
      }
      tms.replace(0,1,"11");
      tdo.insert(0,"0");
      tdoMask.insert(0,"0");

      if (rti)
      {
	tms.insert(0,"0");
	tdo.insert(0,"0");
	tdoMask.insert(0,"0");
      }
      return;
  }
        

/**
<p>DR scan StringBuffer builder</p>
<p>Insert sequence to do DR scan.</p>
<p>Valid from UPDATE-IR, UPDATE-DR or RTI</p>
@param tdoData The data
@param rti If true, insert transitions to RTI
*/
  public void dr(String tdoData, boolean rti)
  {

      tms.insert(0,"001");
      tdo.insert(0,"000");
      tdoMask.insert(0,"000");
      
      if ( tdoData == null ) return; //Assume we want to stay in SHIFT_DR
      
      for (int i=0;i<tapLength;i++)
      {
	if (i==tapPosition)
	{
	  for (int d=0;d<tdoData.length();d++)
	  {
	    tdoMask.insert(0,"1");
	    tms.insert(0,"0");
	  }
	  tdo.insert(0,tdoData);
	}
	else
	{
          tms.insert(0,"0");
	  tdo.insert(0,"0"); // BYPASS
	  tdoMask.insert(0,"0"); // BYPASS
	}
      }
      tms.replace(0,1,"11");
      tdo.insert(0,"0");
      tdoMask.insert(0,"0");

      if (rti)
      {
	tms.insert(0,"0");
	tdo.insert(0,"0");
	tdoMask.insert(0,"0");
      }
      return;
  }

/**
<p>Move from anywhere to RTI</p>
<p>Stay there for the requested extra number of clocks (can be 0).</p>
@param clocks Number of clocks while in RTI
*/
  public void rti(int clocks)
  {
// Five ones take us from anywhere to TLR then 0 to RTI
    tms.insert(0,"011111");
    tdo.insert(0,"000000");
    tdoMask.insert(0,"000000");
// Extra clocks
    for (int i=0;i<clocks;i++)
    {
      tms.insert(0,"0");
      tdo.insert(0,"0");
      tdoMask.insert(0,"0");
    }
  }
    
// Extra clocks with TMS 0 (can be 0)
    
/**
<p>Insert extra clocks with TMS 0 (can be 0)</p>
@param clocks Number of clocks to insert
*/
  public void clock(int clocks)
  {
    for (int i=0;i<clocks;i++)
    {
      tms.insert(0,"0");
      tdo.insert(0,"0");
      tdoMask.insert(0,"0");
    }
  }
        
}
