/*
Unmarshall GBTX register descriptors into statically accessible members of this Gbtx class.
Note that the class members are static but unfortunately not immutable because of the need to use JAXB setters.
See main() for example.
*/
package cbsw.lhcb.pdmdb;

import java.util.prefs.Preferences;
import javax.xml.bind.annotation.*;
import java.nio.file.Paths;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.net.URL;
import java.net.MalformedURLException;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import java.io.*;
import java.util.*;
import cbsw.*;
import cbsw.lhcb.pdmdb.*;

/**
<p>This class encapsulates the interactions with the CLARO.</p>
**/
@XmlAccessorType(XmlAccessType.NONE)
@XmlRootElement(name="claro")
class Claro
{
ScaDepp sca;
public static int csrSize;
public static ArrayList<RegisterDescriptor> descriptors;
public static HashMap<String,RegisterDescriptor> map;

public BitSet csrBits;
public int ec;
public int feb;
public int claro;
public int febEnableGpio;
public int writeEnableGpio;

  @XmlAttribute
  public void setCsrSize(int csrSize){Claro.csrSize = csrSize;}

  @XmlElement(name="register")
  public void setDescriptors(ArrayList<RegisterDescriptor> descriptors)
  {
    Claro.descriptors = descriptors;
  }
  public ArrayList<RegisterDescriptor> getDescriptors(){return descriptors;}
    
  static
  {
    Cbsw.unmarshal("https://www.hep.phy.cam.ac.uk/lhcb/PDMDB/xml/claro.xml",Claro.class);
    Claro.map = new HashMap<String,RegisterDescriptor>();
    for (RegisterDescriptor reg : Claro.descriptors){ Claro.map.put(reg.getName(),reg); }
  }
    
  public Claro(){csrBits = new BitSet();}

  public void init(ScaDepp sca, int ec, int feb, int claro, int febEnableGpio, int writeEnableGpio)
  {
    this.sca = sca;
    this.ec = ec;
    this.feb = feb;
    this.claro = claro;
    this.febEnableGpio = febEnableGpio;
    this.writeEnableGpio = writeEnableGpio;
  }

  public void set(File cf)
  {
    byte [] buf = new byte [csrSize];

    try
    {
      FileInputStream fi = new FileInputStream(cf);
      fi.skip(256*ec+128*feb+16*claro);
      fi.read( buf );
      fi.close();
    }
    catch (IOException ioe) { System.out.println("Csr.set(File): IOException"); System.out.println(ioe.getMessage()); }

    csrBits = BitSet.valueOf(buf);

    return;
  }

// Set a claro register by name and hex-encoded string value
// Decode special case: keys starting ch.*. are applied to all channels
    
  public void set(String key, String val)
  {
    if (key.matches("ch\\.\\*\\..*"))
    {
      for (int ch=0;ch<8;ch++) map.get(key.replace("*",String.valueOf(ch))).set(csrBits,val);
    }
    else
    {
      map.get(key).set(csrBits,val);
    }
  }
    
  public void set(String key, int val){ map.get(key).set(csrBits,String.format("%X",val)); }

  public void set(String key, boolean val){ map.get(key).set(csrBits,val); }

  public byte [] toByteArray(){ return Arrays.copyOfRange(csrBits.toByteArray(),0,16); }
    
  public static void decode(BitSet csrBitSet)
  {
    if (csrBitSet==null) return;
    for (RegisterDescriptor reg : Claro.descriptors)
    {
      System.out.println(reg.getName()+": "+reg.get(csrBitSet));
    } 
    return;
  }
    
// Prepare the SCA commands to program one CLARO
// Call sca.allocateTxRxBuffers() before calling this and sca.putBytes() afterwards to do the transaction.
// Can call this up to 16 times between allocateTxRxBuffers and putBytes().
    
// GBT-SCA-SPI manual mode is used for timing reasons.

// SCA falling TX-edge mode is used. The SCA clock filtering means
// that edges appear to be delayed at the CLARO by about 50ns compared
// to MOSI transitions so rising edge mode would probably be OK but
// falling edge is more conservative.

  public void prepareSpi(boolean wrflag)
  {
    BitSet gpioDataOut = (BitSet)sca.gpioOut.clone();

    if (wrflag)
      gpioDataOut.set(writeEnableGpio);
    else
      gpioDataOut.clear(writeEnableGpio);

    gpioDataOut.set(febEnableGpio);
    sca.infoTx("GPIO_W_DATAOUT",gpioDataOut); // Select FEB
    //    csrBits.set(127); csrBits.set(125); //Read-only bits but convenient to see on scope 
    byte [] claroBuf = Arrays.copyOfRange(csrBits.toByteArray(),0,16);
    //    for (byte b: claroBuf) System.err.println(String.format("%02x",b));
    //    if (claroBuf[15]!=0 || claroBuf[14]!=0 || claroBuf[13]!=0 ) System.err.println("CSR error");

    sca.infoTx("SPI_W_CTRL",0x1000);
    //    sca.infoTx("SPI_W_CTRL",0x1600);
    //    sca.infoTx("SPI_W_CTRL",0x1400);
    sca.infoTx("SPI_W_SS",1<<claro); // Select CLARO target

    int claroBufOffset = 0;
    if (wrflag)
    {
      sca.infoTx("SPI_W_MOSI0",Arrays.copyOfRange(claroBuf,claroBufOffset,claroBufOffset+4)); claroBufOffset+=4;
      sca.infoTx("SPI_W_MOSI1",Arrays.copyOfRange(claroBuf,claroBufOffset,claroBufOffset+4)); claroBufOffset+=4;
      sca.infoTx("SPI_W_MOSI2",Arrays.copyOfRange(claroBuf,claroBufOffset,claroBufOffset+4)); claroBufOffset+=4;
      sca.infoTx("SPI_W_MOSI3",Arrays.copyOfRange(claroBuf,claroBufOffset,claroBufOffset+4)); claroBufOffset+=4;
    }

    sca.infoTx("SPI_GO",0x00000000);

// Deselect all CLARO
    sca.infoTx("SPI_W_SS",0x00000000);

    gpioDataOut.clear(febEnableGpio);
    sca.infoTx("GPIO_W_DATAOUT",gpioDataOut); // Deselect FEB

    if (!wrflag)
    {
      byte [] buf = new byte [4];
      sca.infoTx("SPI_R_MISO0",buf);
      sca.infoTx("SPI_R_MISO1",buf);
      sca.infoTx("SPI_R_MISO2",buf);
      sca.infoTx("SPI_R_MISO3",buf);
    }

    gpioDataOut.clear(writeEnableGpio);

    return;
  }
    
// Do an SPI transaction
  public void doSpi(boolean wrflag)
  {
    sca.allocateTxRxBuffers(0x0a); // 10 SCA commands
    prepareSpi(wrflag);
    sca.putBytes(0,true);
  }

  public boolean verify(byte [] reg)
  {
    BitSet rdBits = BitSet.valueOf(reg);

    // Shift down 1 bit to account for the SPI transaction bit offset
    /*
    for (int i = rdBits.nextSetBit(0); i >= 0; i = rdBits.nextSetBit(i+1))
    {
      if (i == Integer.MAX_VALUE) break;
      rdBits.clear(i); 
      if (i == 0) continue;
      rdBits.set(i-1);
    }
    */
    BitSet maskBits = new BitSet(); maskBits.set(0,104);
    rdBits.and(maskBits);
    boolean pass = rdBits.equals(csrBits);
    if (!pass)
    {
      System.err.println("csrBits="+csrBits);
      rdBits.xor(csrBits);
      System.err.println("Differences: "+rdBits);
    }
    return pass;
  }

  public void setRandom(int irand)
  {
    for (int ch=0;ch<8;ch++)
    {
      set("ch."+ch+".threshold",irand&0x3f);
      set("ch."+ch+".enable",irand&0x1);
      set("ch."+ch+".attenuation",(irand>>6)&0x3);
      set("ch."+ch+".hysteresis-enable",(irand>>1)&0x1);
      set("ch."+ch+".threshold-offset",(irand>>2)&0x1);
      set("ch."+ch+".testpulse-disable",(irand>>3)&0x1);
    }
    set("mux-address",irand&0x7);
    set("seu-generate-enable",(irand>>3)&0x1);
    set("internal-correction",(irand>>4)&0x1);
    set("aux-testpulse",(irand>>5)&0x1);
    set("testpulse-enable",(irand>>6)&0x1);
    set("seu-count-reset",(irand>>7)&0x1);
  }

  public static void main( String[] args )
  {
    System.out.println(Claro.csrSize);
    for (RegisterDescriptor reg : Claro.descriptors){ System.out.println(reg.getName()); }  
  }

}

