/*
Unmarshall register descriptors into members of this Csr class.
*/
package cbsw;

import java.util.prefs.Preferences;
import javax.xml.bind.annotation.*;
import java.nio.file.Paths;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.net.URL;
import java.net.MalformedURLException;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import java.io.*;
import java.util.*;
import cbsw.*;

@XmlAccessorType(XmlAccessType.NONE)
@XmlRootElement(name="csr")
public class Csr
{
BitSet csrBits;
@XmlAttribute
int baseAddress;
@XmlAttribute
int csrSize;
@XmlElement(name="register")
ArrayList<RegisterDescriptor> descriptors;
HashMap<String,RegisterDescriptor> map;

  public ArrayList<RegisterDescriptor> getDescriptors(){return descriptors;}
        
  public Csr(){csrBits = new BitSet();}

  public void init()
  {
    map = new HashMap<String,RegisterDescriptor>();
    for (RegisterDescriptor reg : descriptors){ map.put(reg.getName(),reg); }
  }

  public static Csr unmarshall(String xml)
  {
    Csr csr = (Csr)Cbsw.unmarshal(xml,Csr.class);
    csr.init();  
    return csr;
  }

  public int getBaseAddress(){return baseAddress;}
  public BitSet getCsrBits(){return csrBits;}
    
  public String get( String key ) { return map.get(key).get(csrBits); }
  public String get( String key, BitSet bits ){ return map.get(key).get(bits); }
  public String get( String key, byte [] regs )
  {
    BitSet bits = BitSet.valueOf(regs);
    return map.get(key).get(bits);
  }

// Initialise the BitSet from binary file
  public void set(File cf)
  {
    byte [] buf = new byte [csrSize];

    try
    {
      FileInputStream fi = new FileInputStream(cf);
      fi.read( buf );
      fi.close();
    }
    catch (IOException ioe) { System.out.println("Csr.set(File): IOException"); System.out.println(ioe.getMessage()); }

    csrBits = BitSet.valueOf(buf);

    return;
  }
    
  public void set( String key, Boolean value ) { map.get(key).set(csrBits,value.booleanValue()); }

  public void set( String key, Integer value ) { map.get(key).set(csrBits,String.format("%X",value.intValue())); }

// Set the register bits from value stored in prefs
// key is both the XML key and the preference key name

  public void set( String key, Preferences prefs) { map.get(key).set(csrBits,prefs.get(key,"0")); }

  public void set( String key, Preferences prefs, String defaultValue) { map.get(key).set(csrBits,prefs.get(key,defaultValue)); }

  public void set( String key, Preferences prefs, Integer defaultValue) { map.get(key).set(csrBits,String.format("%X",prefs.getInt(key,defaultValue.intValue()))); }

  public void set( String key, Preferences prefs, Boolean defaultValue) { map.get(key).set(csrBits,prefs.getBoolean(key,defaultValue.booleanValue())); }

  public void set( String key, String regString){ map.get(key).set(csrBits,regString); }

  public byte [] toByteArray(int arrayOffset)
  {
    byte [] buf = new byte [arrayOffset+csrSize];
    System.arraycopy(csrBits.toByteArray(),0,buf,arrayOffset,csrBits.toByteArray().length);
    return buf;
  }

  public RegisterDescriptor getRegisterDescriptor(String key){ return map.get(key); }

  public byte [] getBytes(String key, int arrayOffset){ return map.get(key).getBytes(csrBits,arrayOffset); }
    
  public void decode()
  {
    if (csrBits==null) return;
    for (RegisterDescriptor reg : descriptors){ System.out.println(reg.getName()+": "+reg.get(csrBits)); } 
    return;
  }

// Update the bitset from hardware
  public void cache(RegisterIo io) throws cbsw.RegisterIoException
  {
    try
    {
      byte [] rdBuf = new byte [csrSize];
      io.get(baseAddress,0,rdBuf);
      csrBits = BitSet.valueOf(rdBuf);
    }
    catch ( RegisterIoException e ){ throw e; }
  }
    
// The put() methods update the underlying BitSet and write to the hardware
  public void put(String name, String val, RegisterIo io) throws cbsw.RegisterIoException
  {
    RegisterDescriptor reg = map.get(name);
    reg.set(csrBits,val);
    int off = reg.getOffset();
    try { io.put(baseAddress,off>>3,reg.getBytes(csrBits,4)); }
    catch ( RegisterIoException e ){ throw e; }
  }
    
  public void put(String name, Boolean val, RegisterIo io) throws cbsw.RegisterIoException
  {
    RegisterDescriptor reg = map.get(name);
    reg.set(csrBits,val.booleanValue());
    int off = reg.getOffset();
    try { io.put(baseAddress,off>>3,reg.getBytes(csrBits,4)); }
    catch ( RegisterIoException e ){ throw e; }
  }
    
  public void put(String name, Integer val, RegisterIo io) throws cbsw.RegisterIoException
  {
    RegisterDescriptor reg = map.get(name);
    reg.set(csrBits,String.format("%X",val.intValue()));
    int off = reg.getOffset();
    try { io.put(baseAddress,off>>3,reg.getBytes(csrBits,4)); }
    catch ( RegisterIoException e ){ throw e; }
  }
    
  public void put(String name, Preferences prefs, String defaultValue, RegisterIo io) throws cbsw.RegisterIoException
  {
    RegisterDescriptor reg = map.get(name);
    reg.set(csrBits,prefs.get(name,defaultValue));
    int off = reg.getOffset();
    try { io.put(baseAddress,off>>3,reg.getBytes(csrBits,4)); }
    catch ( RegisterIoException e ){ throw e; }
  }
    
  public void put(String name, Preferences prefs, Boolean defaultValue, RegisterIo io) throws cbsw.RegisterIoException
  {
    RegisterDescriptor reg = map.get(name);
    reg.set(csrBits,prefs.getBoolean(name,defaultValue.booleanValue()));
    int off = reg.getOffset();
    try { io.put(baseAddress,off>>3,reg.getBytes(csrBits,4)); }
    catch ( RegisterIoException e ){ throw e; }
  }
    
  public void put(String name, Preferences prefs, Integer defaultValue, RegisterIo io) throws cbsw.RegisterIoException
  {
    RegisterDescriptor reg = map.get(name);
    reg.set(csrBits,String.format("%X",prefs.getInt(name,defaultValue.intValue())));
    int off = reg.getOffset();
    try { io.put(baseAddress,off>>3,reg.getBytes(csrBits,4)); }
    catch ( RegisterIoException e ){ throw e; }
  }
    
  public static void main( String[] args )
  {
    Csr csr = (Csr)Cbsw.unmarshal("https://hep.phy.cam.ac.uk/lhcb/Chimaera2/xml/tbtrigger.xml");
    csr.init();
    for (RegisterDescriptor reg : csr.getDescriptors()){ System.out.println(reg.getName()); }  
  }

}

