package cbsw.lhcb.pdmdb;

import javax.xml.bind.annotation.*;

/**
<p>Concrete class for SCA I2C interface to DTM GBTX.</p>
<p>Settings for DTM are written to GBTXes using the SCA I2C channels
dtm is both the DTM index and the SCA I2C bus number. gbtx is both
the GBTX index and the I2C slave address.</p>
<p>Note: Each SCA-I2C transaction sends 2 pointer bytes + 14 data
bytes. GBTX has 436 1-byte registers, the first 366 are writeable.</p>
<p>The FPGA buffers are big enough to read or write all the buffers in
one call to put() or get().</p>
 **/
@XmlAccessorType(XmlAccessType.NONE)
class GbtxControl extends GbtxParent
{
  ScaDepp sca;
  int dtm;
  @XmlAttribute(name="resetGpio")
  int resetGpio;
  int i2cBus;
  @XmlAttribute(name="i2cSlaveAddress")
  int i2cSlaveAddress;
  @XmlAttribute(name="i2cSlaveSpeed")
  int i2cSlaveSpeed;

/**
<p>No-argument constructor used during unmarshalling</p>
<p>Call init after unmarshalling to complete initialisation.</p>
**/
  public GbtxControl(){super();}

/**
<p>Write a byte array to the hardware through the SCA interface.</p>
@param gbtxBuf Array of register values.
@param offset Offset to location where to start writing.
**/
  public void put(byte [] gbtxBuf, int offset)
  {
         
    int buflen = gbtxBuf.length;

    byte [] buf = new byte [4];

// For synchronisation, it may be better for transaction to be split
// into smaller blocks, each taking less than 10ms (DEPP timeout). But
// for now we do everything in one transaction and rely on not too
// fast clicking.

    int n14 = (buflen+13)/14; // Total number of 14 byte SCA I2C transactions
    if (n14<1) return; // Nothing to do
    
    sca.allocateTxRxBuffers(6*n14);
    
    int i = 0;
    for (int i14=0;i14<n14;i14++)
    {
      int remaining = buflen - 14*i14;
      int nbytes = (remaining > 13) ? 14 : remaining;
      nbytes+=2; // 2 bytes for the register offset
      
      buf[0] = (byte)((nbytes<<2)&0xff); buf[0] = (byte)(buf[0] | i2cSlaveSpeed);
      sca.infoTx("I2C_W_CTRL/"+i2cBus,buf);

      int startAddr = offset + 14*i14;
      buf[0] = (byte)(startAddr&0xff); buf[1] = (byte)((startAddr>>8)&0xff); // reg offset
      for (int j=2;j<16;j++)
      {
	if (i<buflen) buf[j%4] = gbtxBuf[i++];
        if (j%4 == 3)
        {
	  sca.infoTx("I2C_W_DAT"+j/4+"/"+i2cBus,buf);
	}
      }

      sca.infoTx("I2C_M_7B_W/"+i2cBus,i2cSlaveAddress);
    }

    sca.putBytes(0,true);
    return;
  }

/**
<p>Read a byte array from the hardware through the SCA interface.</p>
@param gbtxBuf Array of register values.
@param offset Offset to location where to start reading.
**/
  public void get(byte [] gbtxBuf, int offset)
  {
    int hdlcFrames = 4+5*((gbtxBuf.length+15)/16);
    sca.allocateTxRxBuffers(hdlcFrames);

// First we send the two byte start offset

    sca.infoTx("I2C_W_CTRL/"+i2cBus,(2<<2)|i2cSlaveSpeed);

// First two bytes are register start offset
    sca.infoTx("I2C_W_DAT0/"+i2cBus,offset&0xffff);

    sca.infoTx("I2C_M_7B_W/"+i2cBus,i2cSlaveAddress);

// Now read back registers 16 at a time (GBTX auto-increments pointer)
    sca.infoTx("I2C_W_CTRL/"+i2cBus,(16<<2)|i2cSlaveSpeed);

    int i = 0;
    while (i<(gbtxBuf.length+3))
    {
      sca.infoTx("I2C_M_7B_R/"+i2cBus,i2cSlaveAddress);

      for (int j=0;j<4;j++)
      {
	sca.infoTx("I2C_R_DAT"+(3-j)+"/"+i2cBus,0x00000000); // Reverse order
        i+=4;
      }
    }

    sca.putBytes(0,true);

// Now decode the SCA responses to extract the GBTX register data
// Data should appear in frames (5,6,7,8),(10,11,12,13),(15,16,17,18)...

    byte [] hdlcrx = new byte [16*hdlcFrames];
    sca.getBytes( 0, hdlcrx );

    int frame = 4;
    int irx = 0;
    while (frame<hdlcrx.length/16)
    {
      frame++;
      for ( int ifr=0;ifr<4;ifr++,frame++ )
      {
	for (int j=0;j<4;j++,irx++)
	{
	  int idx = sca.getRxIndex(frame,j);
          if (idx<0) continue;
	  
          gbtxBuf[irx] = hdlcrx[frame*16+idx];
	  if (irx>=gbtxBuf.length-1) break;
	}
        if (irx>=gbtxBuf.length-1) break;
      }
      if (irx>=gbtxBuf.length-1) break;
    }
    return;
  }

/**
<p>Function called to complete initialisation after unmarshalling.</p>
@param sca The SCA instance
@param dtm The DTM ID
@param i2cBus The I2C bus number
**/
  public void init(ScaDepp sca, int dtm, int i2cBus)
  {
    this.sca = sca;
    this.dtm = dtm;
    this.i2cBus = i2cBus;
    sca.gpioDirection.clear(resetGpio);
    sca.gpioOut.clear(resetGpio);
    setName("dtm"+dtm+".gbtx"+gbtx);
  }

/**
<p>Reset the DTM GBTX.</p>
<p>This is done by toggling the direction of the GPIO as the reset is
defined by a pull-up. The GPIO data out value should always be 0.</p>
**/
  public void reset()
  {
    assertReset();
    deassertReset();
  }

/**
<p>Assert DTM GBTX reset.</p>
**/
  public void assertReset()
  {

    sca.allocateTxRxBuffers(1);
    sca.gpioDirection.set(resetGpio);
    sca.infoTx("GPIO_W_DIRECTION",sca.gpioDirection);
    
    sca.putBytes(0,true);

    try { Thread.sleep(10); } catch (InterruptedException ie) {};

  }

/**
<p>Deassert DTM GBTX reset.</p>
**/
  public void deassertReset()
  {

    sca.allocateTxRxBuffers(1);
    sca.gpioDirection.clear(resetGpio);
    sca.infoTx("GPIO_W_DIRECTION",sca.gpioDirection);
    
    sca.putBytes(0,true);

// The reset is a slowly changing signal (10us) compared to the time
// taken for the SCA to process commands.

    try { Thread.sleep(100); } catch (InterruptedException ie) {};

  }

/**
<p>Perform configuration of the DTM GBTX</p>
<p>Writes the cahed register bits to the hardware.</p>
<p>Program and verify the GBLD registers.</p>
**/
  public void configure()
  {
    byte [] gbtxBuf = java.util.Arrays.copyOfRange(csrBitSet.toByteArray(),0,366);
    put(gbtxBuf,0);

    if (gbldCB.isSelected())
    {
      programGBLD();
      readGBLD();
      byte [] gbldBuf = new byte [7];
      get(gbldBuf,381);
      System.out.println("GBLD registers for DTM"+dtm+".GBTX"+gbtx);
      for (byte b : gbldBuf) System.out.println(String.format("0x%02X",b));
    }
  }

}

