/*
GbtxDepp class

Implements the put() and get() abstract methods from GbtxParent for the I2C interface.

The transactions target the DEPP.smbmaster FPGA core.

Steve Wotton, 2016-
Cavendish Lab (HEP)
*/

package cbsw.lhcb.pdmdb;

import javax.xml.bind.annotation.*;

/**
<p>Concrete class for DEPP I2C interface to master GBTX.</p>
 **/
@XmlAccessorType(XmlAccessType.NONE)
public class GbtxDepp extends GbtxParent
{
  @XmlAttribute(name="smbSlaveAddress")
  int smbSlaveAddress;

/**
<p>No-argument constructor used during unmarshalling</p>
<p>Call init after unmarshalling to complete initialisation.</p>
**/
  public GbtxDepp(){ super(); }
    
/**
<p>Constructor used during unmarshalling by GbtxAdapter</p>
<p>Call init after unmarshalling to complete initialisation.</p>
@param id The GBTX logical ID.
@param slaveAddress The I/O interface slave address (e.g. I2C).
@param txBaseAddress The DEPP base address.
@param rxBaseAddress Not used.
**/
 public GbtxDepp(int id, int slaveAddress, int txBaseAddress, int rxBaseAddress){ super(id, txBaseAddress, rxBaseAddress); smbSlaveAddress = slaveAddress; };
    
/**
<p>Write a byte array to the hardware through the DEPP interface.</p>
@param regBuf Array of register values.
@param offset Offset to location where to start writing.
**/
  public void put(byte [] regBuf, int offset)
  {

    java.nio.ByteBuffer bb = java.nio.ByteBuffer.allocate(7+regBuf.length);
    bb.order(java.nio.ByteOrder.LITTLE_ENDIAN);
    bb.clear();
    bb.putShort((short)(0x0002+regBuf.length)); // Write OP, 2+len I2C bytes
    bb.put((byte)(smbSlaveAddress&0xfe)); // SMB slave address, write op
    bb.putShort((short)(offset&0xffff)); // Reg offset
    bb.put(regBuf); // Reg values
    bb.putShort((short)0xc000); // Stop OP
    bb.flip();
    try { io.put(txBaseAddress,bb); }
    catch (cbsw.RegisterIoException ioe){System.err.println("GbtxDepp.put(): Exception - "+ioe.getMessage());}
  }

/**
<p>Read a byte array from the hardware through the DEPP interface.</p>
@param regBuf Array of register values.
@param offset Offset to location where to start reading.
**/
  public void get(byte [] regBuf, int offset)
  {
    byte [] tmpBuf = new byte [regBuf.length+1]; 
    java.nio.ByteBuffer bb = java.nio.ByteBuffer.allocate(8);
    bb.order(java.nio.ByteOrder.LITTLE_ENDIAN);
    bb.clear();
    //    bb.putShort((short)(0x8002)); // Write OP, 2 I2C bytes
    bb.putShort((short)(0x0002)); // Write OP, 2 I2C bytes
    bb.put((byte)(smbSlaveAddress&0xfe)); // SMB slave address, write op
    bb.putShort((short)(offset&0xffff)); // Reg offset
    bb.putShort((short)(0x4000 + tmpBuf.length)); // Read OP, len I2C bytes
    bb.put((byte)(smbSlaveAddress|0x01)); // SMB slave address, read op
    bb.flip();
    try
    {
      io.put(txBaseAddress,bb);
      io.get(txBaseAddress,tmpBuf);
      System.arraycopy(tmpBuf,1,regBuf,0,regBuf.length);
    }
    catch (cbsw.RegisterIoException ioe){System.err.println("GbtxDepp.put(): Exception - "+ioe.getMessage());}
  }
    
/**
<p>Reset the GBTX.</p>
<p>TCM does not have a master GBTX external reset. Does soft reset instead.</p>
**/
  public void reset()
  {
// The watchdog and timeout feature are disabled during soft configuration otherwise a timeout might cause the soft registers to be reset
    put("watchdog.enable","0");
    put("watchdog.startup-timeout-enable","0");
    
    softReset();
    
// Re-enable the watchdog and timeout feature
    put("watchdog.enable","1");
    put("watchdog.startup-timeout-enable","1");

    /*test();*/
  }

/**
<p>Fuse the GBTX.</p>
<p>Caution! This function may cause configuration e-fuses to be irreversibly fused.</p>
@param pulse Runnable that fires the EFUSEPWREN and EFUSEPROG signals.
**/
public void fuse(java.lang.Runnable pulse)
{
// Clear all currently set bits
  csrBitSet.clear();
// Load the minimal settings
  minimal();
  //  ecEport();
// Get the settings as a byte array
  byte [] gbtxBuf = java.util.Arrays.copyOfRange(csrBitSet.toByteArray(),0,367);
// Loop over the array and fuse the non-zero bytes
  int reg=0;
  System.err.println("--- Fusing non-zero bytes ---");
  try
  {
    for (byte val: gbtxBuf)
    {
      if (val!=0)
      {
// Write the fuse address and value
	byte bb [] = {(byte)(reg&0xff),(byte)((reg>>8)&0xff),val};
        put(bb,238); // Fuse address LSB, MSB, val

// Check the values were written correctly

	byte [] chk = new byte [3];
        get(chk,238);
	if (chk[0] != (byte)(reg&0xff)) throw new RuntimeException("Fuse address LSB error");
	if (chk[1] != (byte)((reg>>8)&0xff)) throw new RuntimeException("Fuse address MSB error");
	if (chk[2] != (byte)(val&0xff)) throw new RuntimeException("Fuse value error");

        System.err.println(String.format("[%03d,%03x] %02X",reg,reg,val));
// Following line causes the fusing (if 3V3 is also applied)
        try {pulse.run(); Thread.sleep(10);} catch (Exception e){}
      }
      reg++;
    }
  }
  catch ( Exception ex ) {throw new RuntimeException("GbtxDepp.fuse(): "+ex.getMessage());}
}

}

