/*
GbtxIcDepp class

Implements the put() and get() abstract methods from GbtxParent for the IC interface.

Steve Wotton, 2016-
Cavendish Lab (HEP)
*/

package cbsw.lhcb.pdmdb;

import javax.xml.bind.annotation.*;

/**
<p>Concrete class for IC interface to master GBTX.</p>
 **/
@XmlAccessorType(XmlAccessType.NONE)
public class GbtxIcDepp extends GbtxParent
{
/**
<p>No-argument constructor used during unmarshalling</p>
<p>Call init after unmarshalling to complete initialisation.</p>
**/
  public GbtxIcDepp(){ super(); }
    
/**
<p>Constructor used during unmarshalling by GbtxAdapter</p>
<p>Call init after unmarshalling to complete initialisation.</p>
@param id The GBTX logical ID.
@param slaveAddress The I/O interface slave address (e.g. I2C).
@param txBaseAddress The DEPP TX base address.
@param rxBaseAddress The DEPP RX base address.
**/
  public GbtxIcDepp(int id, int slaveAddress, int txBaseAddress, int rxBaseAddress){ super(id, txBaseAddress, rxBaseAddress); };
    
/**
<p>Write a byte array to the hardware through the DEPP interface.</p>
@param regBuf Array of register values.
@param offset Offset to location where to start writing.
**/
  public void put(byte [] regBuf, int offset)
  {
    java.nio.ByteBuffer ic = java.nio.ByteBuffer.allocate(12+regBuf.length);
    ic.order(java.nio.ByteOrder.LITTLE_ENDIAN);

// Set header bytes
    ic.putShort((short)(8+regBuf.length)); // Transaction length
    ic.putShort((short)0); // Transaction address
    ic.put((byte)0); // Dummy zero
    ic.put((byte)(0x02&0xff)); // GBTX address & write op
    ic.put((byte)(0x01&0xff)); // GBTX command, value is ignored by GBTX
    ic.putShort((short)regBuf.length); // Length
    ic.putShort((short)(offset&0xffff)); // Register address
    ic.put(regBuf); // value
    byte parity = 0;
    for (int b=6;b<ic.position();b++) parity = (byte)(parity ^ ic.get(b));
    ic.put(parity); // Parity

    try { io.put(txBaseAddress,ic); }
    catch (cbsw.RegisterIoException de) { System.out.println("GbtxIcDepp::put() IO exception"); }

    return;
  }

/**
<p>Read a byte array from the hardware through the DEPP interface.</p>
@param regBuf Array of register values.
@param offset Offset to location where to start reading.
**/
  public void get(byte [] regBuf, int offset)
  {
    try
    {
      java.nio.ByteBuffer ic = java.nio.ByteBuffer.allocate(12);
      ic.order(java.nio.ByteOrder.LITTLE_ENDIAN);

// Set header bytes
      ic.putShort((short)8); // Transaction length
      ic.putShort((short)0); // Transaction address
      ic.put((byte)0); // Dummy zero
      ic.put((byte)(0x03&0xff)); // GBTX address & read op
      ic.put((byte)(0x01&0xff)); // GBTX command, value is ignored by GBTX
      ic.putShort((short)regBuf.length); // Length
      ic.putShort((short)(offset&0xffff)); // Register address
      byte parity = 0;
      for (int b=6;b<ic.position();b++) parity = (byte)(parity ^ ic.get(b));
      ic.put(parity); // Parity

      io.put(txBaseAddress,ic);

      Thread.sleep(200);

      byte [] rxBuffer = new byte [512];
      io.get(rxBaseAddress,0,rxBuffer);

// Start of registers is offset by 7 bytes in the RX buffer
      System.arraycopy(rxBuffer,7,regBuf,0,regBuf.length);
    }
    catch (Exception ex) { throw new RuntimeException("GbtxIcDepp.get(): "+ex.getMessage()); }
  }

/**
<p>Reset the GBTX.</p>
<p>TCM does not have a master GBTX external reset. Does soft reset instead.</p>
**/
  public void reset()
  {
// The watchdog and timeout feature are disabled during soft configuration otherwise a timeout might cause the soft registers to be reset
    put("watchdog.enable","0");
    put("watchdog.startup-timeout-enable","0");
    
    softReset();
    
// Re-enable the watchdog and timeout feature
    put("watchdog.enable","1");
    put("watchdog.startup-timeout-enable","1");

  /*test();*/
  }
    
}

