package cbsw.lhcb;

import javax.swing.*;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.*;

/**
<p>Programmatic interface for SPI flash memories</p>
<p>The IO interface addresses are set in Java Preferences:</p>
<p><code>/M25Pxx/rxBase</code> Receive buffer address</p>
<p><code>/M25Pxx/txBase</code> Transmit buffer address</p>
 */
class M25PxxIo
{
  java.util.prefs.Preferences prefs;
  cbsw.RegisterIo io;
  JTextArea textArea;

  byte [] bytes;
  ByteBuffer byteBuffer;

  final int pageSize=256;
  final int txPages=4;
  static final byte PP = (byte)0x02;
  static final byte READ = (byte)0x03;
  static final byte WRDI = (byte)0x04;
  static final byte RDSR = (byte)0x05;
  static final byte WREN = (byte)0x06;
  static final byte RDID = (byte)0x9f;
  static final byte BE = (byte)0xc7;
  static final byte SE = (byte)0xd8;
    
/**
<p>Constructor</p>
@param io The IO interface
*/
  public M25PxxIo(cbsw.RegisterIo io)
  {
    prefs = java.util.prefs.Preferences.userNodeForPackage(M25PxxIo.class).node("M25PXX");
    this.io = io;
    textArea = null;
    
    bytes = new byte [2048]; // FPGA BRAM size 
    byteBuffer = ByteBuffer.wrap(bytes);
    byteBuffer.order(ByteOrder.LITTLE_ENDIAN);
  }

/**
<p>Registers a text area</p>
@param textArea The text area
*/
  public void setTextArea(JTextArea textArea){ this.textArea = textArea; }

/**
<p>Displays text in the registered text area</p>
@param txt The text to display
*/
  public void showText(String txt)
  {
      if (textArea == null)
      {
        System.out.println(txt);
      }
      else
      {
	textArea.append(txt+"\n");
      }
  }

/**
<p>Clears the transaction buffer</p>
*/
  public void clear(){byteBuffer.clear(); byteBuffer.position(4);}
    
/**
<p>Appends dataless operation to transaction buffer</p>
@param code Opcode
@param opLength Length field of the operation in bytes
*/
  public void append(byte code, short opLength)
  {
    byteBuffer.putShort(opLength);
    byteBuffer.put(code);
  }
    
/**
<p>Appends operation to transaction buffer with data</p>
@param code Opcode
@param opLength Length field of the operation in bytes
@param buf Data to write
*/
  public void append(byte code, short opLength, byte [] buf)
  {
    byteBuffer.putShort(opLength);
    byteBuffer.put(code);
    byteBuffer.put(buf);
  }
        
/**
<p>Writes instruction to buffer.</p>
<p>End of commands is marked by x"0000" in the next length field.</p>
@param offset Offset into buffer where to start writing
@param waitflag If true, activate the IO wait feature
*/
  public void op(int offset, boolean waitflag)
  {
    // Write terminator
    byteBuffer.putShort((short)0);
    int byteBufferLength = byteBuffer.position();
    // Write the header
    byteBuffer.putShort(0,(short)(byteBufferLength-4));
    if (waitflag) offset |= 0x8000;
    byteBuffer.putShort(2,(short)offset);
    byteBuffer.flip();
    byte [] buf = new byte[byteBufferLength];
    byteBuffer.get(buf);

    try { io.put( prefs.getInt("txBase",0x70), ByteBuffer.wrap(buf) ); }
    catch ( cbsw.RegisterIoException ex ) { System.out.println("op(): RegisterIo exception. "+ex.getMessage()); }

    //    showText("--> "+dataLength+" bytes");

  }
    
/**
<p>Does bulk erase operation</p>
<p>May take tens of seconds depending on memory type</p>
@param poll If true, poll until operation has completed
@param statusTF Text field where to write status
*/
  void bulkErase(boolean poll, JTextField statusTF)
  {
    clear();
    append(M25PxxIo.WREN,(short)0x0001);
    append(M25PxxIo.BE,(short)0x0001);
    op(0,false);
    
    int sl=0;
    if (poll)
    {
      byte [] status = new byte [2];
      do
      {
        try {Thread.sleep(1000);} catch (InterruptedException ie){}
        getRxBuffer(0,status);
	statusTF.setText(String.format("Erasing %dsec",sl++));
      }
      while ((status[1]&0x01) == 0x01);
    }
  }

/**
<p>Does sector erase operation</p>
@param start Start address of sector
*/
  void sectorErase(int start)
  {
    clear();
    append(M25PxxIo.WREN,(short)0x0001);
    append(M25PxxIo.SE,(short)0x0004);
    byteBuffer.put((byte)((start>>16) & 0xff)); // Addr 23..16
    byteBuffer.put((byte)((start>>8) & 0xff)); // Addr 15..8
    byteBuffer.put((byte)(start & 0xff)); // Addr 7..0
    op(0,false);
  }
    
/**
<p>Does read operation</p>
@param start Start address
*/
  void read(int start)
  {
    clear();
    append(M25PxxIo.READ,(short)0x0044);
    byteBuffer.put((byte)((start>>16) & 0xff)); // Addr 23..16
    byteBuffer.put((byte)((start>>8) & 0xff)); // Addr 15..8
    byteBuffer.put((byte)(start & 0xff)); // Addr 7..0
    for (int i=0;i<64;i++) byteBuffer.put((byte)0);
    op(0,true);
  }
    
/**
<p>Does oage program operation</p>
<p>Currently this is just a test operation that writes a fixed pattern.</p>
@param start Start address
*/
  void pp(int start)
  {
    clear();
    append(M25PxxIo.PP,(short)(pageSize+4));
    byteBuffer.put((byte)((start>>16) & 0xff)); // Addr 23..16
    byteBuffer.put((byte)((start>>8) & 0xff)); // Addr 15..8
    byteBuffer.put((byte)(start & 0xff)); // Addr 7..0
    for (int i=0;i<pageSize;i++) byteBuffer.put((byte)5); // Some dummy data for test
    op(0,true);
  }
    
/**
<p>Reads the flash data from file and programs flash.</p>
<p>It is most efficient to minimise the number of DEPP transactions
therefore buffer several pages into the same DEPP message.</p>
<p>The FPGA buffer is 16Kb so we could do up to 7 pages at a time but we
must also take account of the programming time which must not exceed
the 10ms DEPP timeout since we are using DEPP wait to synchronise with
timeout of 10ms.</p>
@param buf Buffer containing data to write
@param statusTF Test field where to display status
*/
  void program(byte [] buf, JTextField statusTF)
  {
    int npages = (buf.length + pageSize - 1)/pageSize;
    int b = 0;

    clear();

    for (int page=0; page<npages; page++)
    {
      append(M25PxxIo.WREN,(short)0x0001);

      int nwb = (buf.length - b < pageSize) ? buf.length - b : pageSize;

      append(M25PxxIo.PP,(short)(pageSize+4));
      byteBuffer.order(ByteOrder.BIG_ENDIAN);
      byteBuffer.putShort((short)page);
      byteBuffer.order(ByteOrder.LITTLE_ENDIAN);
      byteBuffer.put((byte)0);
      for (int pb=0;pb<nwb;pb++) byteBuffer.put(buf[b++]);

// Trigger the transaction every txPages pages
      if ((page%txPages == (txPages-1)) || b == buf.length)
      {
	op(0,true);
	clear();
	statusTF.setText(String.format("Programming page %d",page));
      }
    }
    statusTF.setText(String.format("Programmed %d pages",npages));
  }

/**
<p>Displays data from the RX buffer</p>
@param offset Offset in buffer from where to start reading
@param buf Buffer to receive data
 */
  public void showRxBuffer(int offset, byte [] buf)
  {
    getRxBuffer(offset, buf);

    StringBuffer hexSB = new StringBuffer();
    for ( byte b : buf ) hexSB.append(String.format("%1$02x",b));
    showText( hexSB.toString() );
  }

/**
<p>Reads data from the RX buffer</p>
@param offset Offset in buffer from where to start reading
@param buf Buffer to receive data
 */
  public void getRxBuffer(int offset, byte [] buf)
  {
      try { io.get( prefs.getInt("rxBase",0xf0), offset, buf ); }
    catch ( cbsw.RegisterIoException re ) { System.out.println("showRxBuffer(): IO exception. "+re.getDescription()); }
  }
}

