package cbsw.lhcb.pdmdb;

import java.util.prefs.Preferences;
import javax.xml.bind.annotation.*;
import java.io.*;
import java.util.*;
import cbsw.*;

/**
<p>This class encapsulates the interactions with the CLARO</p>
<p>Unmarshalls register descriptors into statically accessible members
of this class.</p>
*/
@XmlAccessorType(XmlAccessType.NONE)
@XmlRootElement(name="claro")
class Claro
{
/**<p>The SCA interface</p>*/
  ScaDepp sca;
/**<p>The size in bytes of the CLARO register</p>*/
  public static int csrSize;
/**<p>The CLARO register descriptors</p>*/
  public static ArrayList<RegisterDescriptor> descriptors;
/**<p>HashMap for register descriptors keyed by bit-field name</p>*/
  public static HashMap<String,RegisterDescriptor> map;

/**<p>The BitSet that mirrors the CLARO register settings</p>*/
  public BitSet csrBits;
/**<p>The EC ID of this CLARO</p>*/
  public int ec;
/**<p>The FEB ID of this CLARO</p>*/
  public int feb;
/**<p>The ID of this CLARO</p>*/
  public int claro;
/**<p>The GPIO number for the FEB enable for this CLARO</p>*/
  public int febEnableGpio;
/**<p>The GPIO number for the write enable for this CLARO</p>*/
  public int writeEnableGpio;

/**
<p>Sets the CLARO CSR size.</p>
<p>Called by the unmarshaller.</p>
@param csrSize The size in bytes of the CLARO register.
*/
  @XmlAttribute
  public void setCsrSize(int csrSize){Claro.csrSize = csrSize;}

/**
<p>Sets the RegisterDescriptor list.</p>
<p>Called by the unmarshaller.</p>
@param descriptors The descriptor list to set.
*/
  @XmlElement(name="register")
  public void setDescriptors(ArrayList<RegisterDescriptor> descriptors){ Claro.descriptors = descriptors; }

/**
<p>Gets the RegisterDescriptor list.</p>
<p>This is required by JAXB even though marshalling is not used in this application.</p>
@return descriptors The descriptor list.
*/    
  public ArrayList<RegisterDescriptor> getDescriptors(){ return Claro.descriptors; }

  static
  {
    Cbsw.unmarshal("https://www.hep.phy.cam.ac.uk/lhcb/PDMDB/xml/claro.xml",Claro.class);
    Claro.map = new HashMap<String,RegisterDescriptor>();
    Claro.descriptors.forEach(reg->Claro.map.put(reg.getName(),reg));
  }

/**
<p>No-argument constructor called by the unmarshaller</p>
*/
  public Claro(){csrBits = new BitSet();}

/**
<p>Completes initialisation of this instance.</p>
<p>Call after unmarshalling to complete initialisation.</p>
@param sca The SCA interface.
@param ec The EC ID for this CLARO.
@param feb The FEB ID for this CLARO.
@param claro The CLARO ID.
@param febEnableGpio The FEB enable GPIO number for this CLARO.
@param writeEnableGpio The write enable GPIO number for this CLARO.
*/
  public void init(ScaDepp sca, int ec, int feb, int claro, int febEnableGpio, int writeEnableGpio)
  {
    this.sca = sca;
    this.ec = ec;
    this.feb = feb;
    this.claro = claro;
    this.febEnableGpio = febEnableGpio;
    this.writeEnableGpio = writeEnableGpio;
  }

/**
<p>Initialises the resister mirror bits from a file.</p>
@param cf The configuration file.
*/
  public void set(File cf)
  {
    byte [] buf = new byte [csrSize];

    try
    {
      FileInputStream fi = new FileInputStream(cf);
      fi.skip(256*ec+128*feb+16*claro);
      fi.read( buf );
      fi.close();
    }
    catch (IOException ioe) { System.out.println("Csr.set(File): IOException"+ioe.toString()); }

    csrBits = BitSet.valueOf(buf);

    return;
  }

/**
<p>Sets a CLARO register mirror bit-field.</p>
<p>Special case: keys starting ch.*. are applied to all channels.</p>
@param <T> The type of the value to set.
@param key The register bit-field key name.
@param val The value to set.
*/
  public <T> void set(String key, T val)
  {
    if (key.matches("ch\\.\\*\\..*"))
    {
      for (int ch=0;ch<8;ch++) map.get(key.replace("*",String.valueOf(ch))).set(csrBits,cbsw.Csr.valueOf(val));
    }
    else
    {
      map.get(key).set(csrBits,cbsw.Csr.valueOf(val));
    }
  }

/**
<p>Returns the current register mirror settings as a byte array.</p>
@return The byte array containing the settings.
*/
  public byte [] toByteArray(){ return Arrays.copyOfRange(csrBits.toByteArray(),0,16); }

/**
<p>Displays all the register bit-field settings on System.out.</p>
@param csrBitSet The register bits.
*/
  public static void decode(BitSet csrBitSet)
  {
    if (csrBitSet==null) return;
    Claro.descriptors.forEach(reg->System.out.println(reg.getName()+": "+reg.get(csrBitSet)));
    return;
  }
    
/** 
<p>Prepares the SCA commands to program one CLARO.</p>
<p>Call sca.allocateTxRxBuffers() before calling this and
sca.putBytes() afterwards to do the transaction.</p>
<p>Can call this up to 16 times between allocateTxRxBuffers() and putBytes().</p>
<p>GBT-SCA-SPI manual mode is used for timing reasons.</p>
<p>SCA falling TX-edge mode is used. The SCA clock filtering means
that edges appear to be delayed at the CLARO by about 50ns compared to
MOSI transitions so rising edge mode would probably be OK but falling
edge is more conservative.</p>
@param wrflag If true, activate the CLARO write enable.
*/
  public void prepareSpi(boolean wrflag)
  {
    BitSet gpioDataOut = (BitSet)sca.gpioOut.clone();

    if (wrflag)
      gpioDataOut.set(writeEnableGpio);
    else
      gpioDataOut.clear(writeEnableGpio);

    gpioDataOut.set(febEnableGpio);
    sca.infoTx("GPIO_W_DATAOUT",gpioDataOut); // Select FEB
//    csrBits.set(127); csrBits.set(125); //Read-only bits but convenient to see on scope 
    byte [] claroBuf = Arrays.copyOfRange(csrBits.toByteArray(),0,16);
//    for (byte b: claroBuf) System.err.println(String.format("%02x",b));
//    if (claroBuf[15]!=0 || claroBuf[14]!=0 || claroBuf[13]!=0 ) System.err.println("CSR error");

    sca.infoTx("SPI_W_CTRL",0x1000);
//    sca.infoTx("SPI_W_CTRL",0x1600);
//    sca.infoTx("SPI_W_CTRL",0x1400);
    sca.infoTx("SPI_W_SS",1<<claro); // Select CLARO target

    int claroBufOffset = 0;
    if (wrflag)
    {
      sca.infoTx("SPI_W_MOSI0",Arrays.copyOfRange(claroBuf,claroBufOffset,claroBufOffset+4)); claroBufOffset+=4;
      sca.infoTx("SPI_W_MOSI1",Arrays.copyOfRange(claroBuf,claroBufOffset,claroBufOffset+4)); claroBufOffset+=4;
      sca.infoTx("SPI_W_MOSI2",Arrays.copyOfRange(claroBuf,claroBufOffset,claroBufOffset+4)); claroBufOffset+=4;
      sca.infoTx("SPI_W_MOSI3",Arrays.copyOfRange(claroBuf,claroBufOffset,claroBufOffset+4)); claroBufOffset+=4;
    }

    sca.infoTx("SPI_GO",0x00000000);

// Deselect all CLARO
    sca.infoTx("SPI_W_SS",0x00000000);

    gpioDataOut.clear(febEnableGpio);
    sca.infoTx("GPIO_W_DATAOUT",gpioDataOut); // Deselect FEB

    if (!wrflag)
    {
      byte [] buf = new byte [4];
      sca.infoTx("SPI_R_MISO0",buf);
      sca.infoTx("SPI_R_MISO1",buf);
      sca.infoTx("SPI_R_MISO2",buf);
      sca.infoTx("SPI_R_MISO3",buf);
    }

    gpioDataOut.clear(writeEnableGpio);

    return;
  }
    
/**
<p>Does an SPI transaction</p>
@param wrflag If true, activates the CLARO write enable
*/
  public void doSpi(boolean wrflag)
  {
    sca.allocateTxRxBuffers(0x0a); // 10 SCA commands
    prepareSpi(wrflag);
    sca.putBytes(0,true);
  }

/**
<p>Verifies the CLARO settings against the mirrored values</p>
@param reg The array of register bytes previously read from the CLARO
@return Verification status
*/
  public boolean verify(byte [] reg)
  {
    BitSet rdBits = BitSet.valueOf(reg);

// Shift down 1 bit to account for the SPI transaction bit offset
/*
for (int i = rdBits.nextSetBit(0); i >= 0; i = rdBits.nextSetBit(i+1))
{
if (i == Integer.MAX_VALUE) break;
rdBits.clear(i); 
if (i == 0) continue;
rdBits.set(i-1);
}
*/
  BitSet maskBits = new BitSet(); maskBits.set(0,104);
  rdBits.and(maskBits);
  boolean pass = rdBits.equals(csrBits);
  if (!pass)
  {
    System.err.println("csrBits="+csrBits);
    rdBits.xor(csrBits);
    System.err.println("Differences: "+rdBits);
  }
  return pass;
  }

/**
<p>Sets random values in the CLARO register for testing</p>
@param irand A random number
*/
  public void setRandom(int irand)
  {
    for (int ch=0;ch<8;ch++)
    {
      set("ch."+ch+".threshold",irand&0x3f);
      set("ch."+ch+".enable",irand&0x1);
      set("ch."+ch+".attenuation",(irand>>6)&0x3);
      set("ch."+ch+".hysteresis-enable",(irand>>1)&0x1);
      set("ch."+ch+".threshold-offset",(irand>>2)&0x1);
      set("ch."+ch+".testpulse-disable",(irand>>3)&0x1);
    }
    set("mux-address",irand&0x7);
    set("seu-generate-enable",(irand>>3)&0x1);
    set("internal-correction",(irand>>4)&0x1);
    set("aux-testpulse",(irand>>5)&0x1);
    set("testpulse-enable",(irand>>6)&0x1);
    set("seu-count-reset",(irand>>7)&0x1);
  }
}

