package cbsw.lhcb.pdmdb;

import javax.xml.bind.annotation.*;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.Color;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;
import java.util.prefs.Preferences;

/**
<p>This class encapsulates the interactions with the collection of CLAROs.</p>
**/
@XmlAccessorType(XmlAccessType.NONE)
class ClaroControl implements ActionListener
{
  Preferences prefs;
  /**<p>The SCA interface</p>*/
  ScaDepp sca;
  /**<p>List of CLARO instances</p>*/
  ArrayList<Claro> claros;
  /**<p>Number of FEBs per EC per PDMDB</p>*/
  public static final int nFebs = 2;
  /**<p>Number of CLAROs per FEB</p>*/
  public static final int nClaros = 8;
  /**<p>The EC ID of this CLARO</p>*/
  int ec;
  /**<p>GPIO number of the CLARO write enable</p>*/
  @XmlElement
  int writeEnable;
  /**<p>GPIO number of FEB1 enable</p>*/
  @XmlElement
  int feb1Enable;
  /**<p>GPIO number of FEB2 enable</p>*/
  @XmlElement
  int feb2Enable;
  /**<p>GPIO number of POR*</p>*/
  @XmlElement
  int npor;
  /**<p>GPIO number of FEB ALDO enable</p>*/
  @XmlElement
  int aldoEnable;
  /**<p>GPIO number of CLARO SEUD</p>*/
  @XmlElement
  int seud;

  private JCheckBox testEnableCB;
  public boolean testEnable = false;
  private JComboBox<String> regNameCB;
  private JTextField regValueTF;

/**
<p>No-argument constructor called by the unmarshaller</p>
 */
  public ClaroControl()
  {
    prefs = Preferences.userNodeForPackage(ClaroControl.class).node("SCA");
    claros = new ArrayList<Claro>();
  }

  public void actionPerformed( ActionEvent e )
  {
    Object ob = e.getSource();
    String command = e.getActionCommand();

    if ( command.equals("configure") )
    {
      sca.allocateTxRxBuffers(10*claros.size()); // 10 SCA commands per CLARO
      for ( Claro claro : claros ) claro.prepareSpi(true);
      sca.putBytes(0,true);
    }
    else if ( command.equals("verify") )
    {
      JButton button = (JButton)ob;
      button.setBackground(Color.yellow);
      if (verify(sca,claros))
        button.setBackground(Color.green);
      else
        button.setBackground(Color.orange);
    }
  }

/**
<p>Completes the initialisation of the instance</p>
<p>Call after unmarshalling to complete initialisation.</p>
@param sca The SCA interface
@param ec The EC ID
 */
  public void init(ScaDepp sca, int ec)
  {
    this.sca = sca;
    this.ec = ec;
    
    java.io.File cf = new java.io.File(prefs.get("claro.filename","claro.spi"));
    for (int ifeb=0;ifeb<nFebs;ifeb++)
    {
      for (int iclaro=0;iclaro<nClaros;iclaro++)
      {
	  int febEnable = (ifeb == 0) ? feb1Enable : feb2Enable;
	  Claro claro = new Claro(); claro.init(sca,ec,ifeb,iclaro,febEnable,writeEnable);
	  claro.set(cf);
	  claros.add(claro);
      }
    }
    
    sca.gpioDirection.set(writeEnable);
    sca.gpioOut.clear(writeEnable);
    sca.gpioDirection.set(feb1Enable);
    sca.gpioOut.clear(feb1Enable);
    sca.gpioDirection.set(feb2Enable);
    sca.gpioOut.clear(feb2Enable);
    sca.gpioDirection.clear(npor);
    sca.gpioOut.clear(npor);
    sca.gpioDirection.set(aldoEnable);
    sca.gpioOut.clear(aldoEnable);
    sca.gpioDirection.clear(seud);
    sca.gpioOut.clear(seud);
  }

/**
<p>Convenience method to create a GUI panel</p>
@return The panel
 */
  public JPanel makePanel()
  {
    JPanel jp = new JPanel();
    jp.setLayout(new GridBagLayout());
    GridBagConstraints gbC = new GridBagConstraints();
    gbC.fill = GridBagConstraints.VERTICAL;
    gbC.anchor = GridBagConstraints.CENTER;
    gbC.insets = new Insets(1,1,1,1);
    gbC.weightx = 1.0f;
    gbC.weighty = 1.0f;
    gbC.fill = GridBagConstraints.BOTH;

    JButton button = new JButton("CLARO Configure");
    button.addActionListener(this);
    button.setActionCommand("configure");
    gbC.gridwidth = 1;
    jp.add(button,gbC);

    button = new JButton("CLARO Verify");
    button.addActionListener(this);
    button.setActionCommand("verify");
    gbC.gridwidth = 1;
    jp.add(button,gbC);

    testEnableCB = new JCheckBox("Test enable",testEnable);
    testEnableCB.addActionListener(l->testEnable=testEnableCB.isSelected());
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    jp.add(testEnableCB,gbC);

    regNameCB = new JComboBox<String>();
    regNameCB.setEditable(true);
    regNameCB.addItem("ch.*.threshold");
    regNameCB.addItem("ch.*.attenuation");
    regNameCB.addItem("ch.*.threshold-offset");
    regNameCB.addItem("ch.*.enable");
    regNameCB.addItem("ch.*.hysteresis-enable");
    regNameCB.addItem("ch.*.testpulse-disable");
    regNameCB.addItem("testpulse-enable");
    regNameCB.addItem("aux-testpulse");
    regNameCB.addItem("internal-correction");
    regNameCB.addItem("seu-count-reset");
    regNameCB.addItem("seu-generate-enable");
    regNameCB.addItem("mux-address");
    gbC.gridwidth = 1;
    jp.add(regNameCB,gbC);
    
    regValueTF = new JTextField("0");
    gbC.gridwidth = 1;
    jp.add(regValueTF,gbC);

    button = new JButton("Set");
    button.addActionListener(l->set());
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    jp.add(button,gbC);
    
    return jp;
  }

/**
<p>Sets CLARO register value for all CLAROs</p>
<p>Called from the GUI panel</p>
 */
private void set(){  for ( Claro claro : claros ) claro.set((String)regNameCB.getSelectedItem(),regValueTF.getText()); }

/**
<p>Configures all CLAROs</p>
@return Configuration status
 */
  public boolean configure()
  {
    boolean cfgOK = false;
    if (claros.size()>16) {return cfgOK;}
    sca.allocateTxRxBuffers(10*claros.size()); // 10 SCA commands per CLARO
    for ( Claro claro : claros ) claro.prepareSpi(true);
    sca.putBytes(0,true);
    cfgOK = verify(sca, claros);
    return cfgOK;
  }
    
// Write or read CLARO configuration registers for up to 16 CLAROs.
// The FPGA buffer (36kb) is large enough for all the commands needed for both FEBs.

/*
  public static void doSpi(ScaDepp sca, ArrayList<Claro> claroList, boolean wrflag)
  {
    if (claroList.size()>16) {System.err.println("ClaroControl.doSpi(): Greater than 16 CLAROs. Bailing out..."); return;}
    sca.allocateTxRxBuffers(10*claroList.size()); // 10 SCA commands per CLARO
    for ( Claro claro : claroList ) claro.prepareSpi(wrflag);
    sca.putBytes(0,true);
    return;
  }
*/
    
/**
<p>Deselects (SPI) all CLAROs</p>
<p>Puts slave selects into safe deselected state.</p>
*/
  public void deselectAllCLARO()
  {
    sca.allocateTxRxBuffers(2);
    sca.infoTx("GPIO_W_DATAOUT",sca.gpioOut);
    sca.infoTx("SPI_W_SS",0x00000000); // Deselect all CLARO
    sca.putBytes(0,true);
  }

/**
<p>Reads CLARO configuration registers and compares with mirrored register bits.</p>
@param sca The SCA interface
@param claroList List of CLAROs to check
@return Verification status
*/
  public static boolean verify(ScaDepp sca, ArrayList<Claro> claroList)
  {
    try {Thread.sleep(10);}catch(InterruptedException ex){System.err.println("ClaroControl.verify(): sleep interrupted");};
// Do the SPI read
    sca.allocateTxRxBuffers(10*claroList.size()); // 10 SCA commands per CLARO
    for ( Claro claro : claroList ) claro.prepareSpi(false);
    sca.putBytes(0,true);

// Sleep to ensure responses have been written to FPGA buffer.

    try {Thread.sleep(10);}catch(InterruptedException ex){System.err.println("ClaroControl.verify(): sleep interrupted");};

// Now decode the SCA responses to extract and verify the CLARO register data

    sca.cache();
    
    boolean pass = true;
    for (Claro claro: claroList)
    {
      byte [] rxbuf = new byte [16];
      for ( int ifr=0;ifr<4;ifr++ )
      {
	byte [] miso = sca.getResponseBytes("SPI_R_MISO"+ifr);
	for (int j=0;j<4;j++) rxbuf[4*ifr+j] = miso[j];
      }
      if (!claro.verify(rxbuf))
      {
	System.err.println(String.format("ClaroControl.verify() [EC,FEB,CLARO]=[%d,%d,%d] Verification mismatch",claro.ec,claro.feb,claro.claro));
	pass = false;
      }
    }
    return pass;
  }

}
