package cbsw.lhcb.pdmdb;

import javax.xml.bind.annotation.*;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;

/**
<p>This class encapsulates the interactions with the DTM.</p>
**/
@XmlAccessorType(XmlAccessType.NONE)
class DtmControl implements ActionListener
{
  ScaDepp sca;
  @XmlAttribute(name="id")
  int dtm;
  @XmlAttribute(name="i2cBus")
  int i2cBus;
  @XmlElement(name="tap")
  XilS7Control xilS7;
  @XmlElement(name="gbtx")
  List<GbtxControl> gbtxList;
  @XmlElement(name="fpgaI2cSlaveAddress")
  int fpgaI2cSlaveAddress;
  JTextField csrOffsetTF;
  JTextField readValueTF, writeValueTF;

/**
<p>No-argument constructor called by the unmarshaller</p>
 */
  public DtmControl() {}

/**
<p>Convenience method to create a GUI panel</p>
@return The panel
 */
  public JPanel makePanel()
  {
    JPanel gjp = new JPanel();
    gjp.setName( "DTM."+dtm );

    gjp.setLayout(new GridBagLayout());
    GridBagConstraints gbC = new GridBagConstraints();
    gbC.fill = GridBagConstraints.VERTICAL;
    gbC.anchor = GridBagConstraints.CENTER;
    gbC.insets = new Insets(1,1,1,1);
    gbC.weightx = 1.0f;
    gbC.weighty = 1.0f;
    gbC.fill = GridBagConstraints.BOTH;

    JButton button;
    button = new JButton("Hold all reset");
    button.addActionListener(l->holdAllReset());
    button.setActionCommand("holdReset");
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    gjp.add(button,gbC);

    for ( GbtxControl gbtx : gbtxList.toArray(new GbtxControl[0]) )
    {
	gbC.gridwidth = (gbtx.gbtx < 1) ? 2 : GridBagConstraints.REMAINDER;
	gjp.add(gbtx.makePanel(),gbC);
    }

    gbC.gridwidth = GridBagConstraints.REMAINDER;
    gjp.add(xilS7.makePanel(),gbC);

    
    gbC.gridwidth = 1;
    gjp.add(new JLabel("FPGA CSR"),gbC);
    
    csrOffsetTF = new JTextField("0",10);
    gbC.gridwidth = 1;
    gjp.add(csrOffsetTF,gbC);

    button = new JButton("Read");
    button.addActionListener(this);
    button.setActionCommand("readFpga");
    gbC.gridwidth = 1;
    gjp.add(button,gbC);

    readValueTF = new JTextField("0",10);
    readValueTF.setEditable(false);
    gbC.gridwidth = 1;
    gjp.add(readValueTF,gbC);
    
    button = new JButton("Write");
    button.addActionListener(this);
    button.setActionCommand("writeFpga");
    gbC.gridwidth = 1;
    gjp.add(button,gbC);

    writeValueTF = new JTextField("0",10);
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    gjp.add(writeValueTF,gbC);
    
    return gjp;
  }

  public void actionPerformed( ActionEvent e )
  {
    Object ob = e.getSource();
    String command = e.getActionCommand();

    if ( command.equals("readFpga") )
    {
      int offset = Integer.parseInt(csrOffsetTF.getText(),16);
      byte [] regBuf = new byte [1];
      get(regBuf,offset);
      readValueTF.setText(String.format("%1$X",regBuf[0]));
    }
    else if ( command.equals("writeFpga") )
    {
      int offset = Integer.parseInt(csrOffsetTF.getText(),16);
      byte [] regBuf = {(byte)Integer.parseInt(writeValueTF.getText(),16)};
      put(regBuf,offset);
    }
  }

/**
<p>Call this method after unmarshalling to complete the initialisation</p>
@param sca The SCA interface
  */
  public void init(ScaDepp sca)
  {
    this.sca = sca;
    for ( GbtxControl gbtx : gbtxList ){ gbtx.init(sca,dtm,i2cBus); }
    xilS7.init(sca);
  }  

/**
<p>Writes a byte array to the FPGA registers through the SCA I2C interface.</p>
@param regBuf Array of register values.
@param offset Offset to location where to start writing.
*/
  public void put(byte [] regBuf, int offset){ sca.putI2c(i2cBus,fpgaI2cSlaveAddress,2,regBuf,offset); }
    

/**
<p>Reads a byte array from the FPGA registers through the SCA I2C interface.</p>
@param regBuf Array of register values.
@param offset Offset to location where to start reading.
*/
  public void get(byte [] regBuf, int offset){ sca.getI2c(i2cBus,fpgaI2cSlaveAddress,2,regBuf,offset); }
   
/**
<p>Holds both GBTXes in reset</p>
  */
  public void holdAllReset(){ for ( GbtxControl gbtx : gbtxList ) gbtx.assertReset(); }

/**
<p>Configures the DTM module</p>
<p>Holds both DTM GBTXs in reset. Deasserts reset for GBTX0 and configures it.
Deasserts reset for GBTX1 and configures it.</p>
**/
  public void configure()
  {
      holdAllReset();
      
      for (GbtxControl gbtx : gbtxList)
      {
        gbtx.reset();
        gbtx.configure();
        try {Thread.sleep(1000);} catch(InterruptedException e){} // Give some time for startup to complete
      }
      return;
  }
    
}
