package cbsw.lhcb.pdmdb;

import javax.xml.bind.annotation.*;

/**
<p>Concrete class for SCA I2C interface to DTM GBTX.</p>
<p>Settings for DTM are written to GBTXes using the SCA I2C channels
dtm is both the DTM index and the SCA I2C bus number. gbtx is both
the GBTX index and the I2C slave address.</p>
<p>Note: Each GBTX has 436 1-byte registers, the first 366 are writeable.</p>
<p>The FPGA buffers are big enough to read or write all the buffers in
one call to put() or get().</p>
 **/
@XmlAccessorType(XmlAccessType.NONE)
class GbtxControl extends GbtxParent
{
  ScaDepp sca;
  int dtm;
  @XmlAttribute(name="resetGpio")
  int resetGpio;
  int i2cBus;
  @XmlAttribute(name="i2cSlaveAddress")
  int i2cSlaveAddress;
  @XmlAttribute(name="i2cSlaveSpeed")
  int i2cSlaveSpeed;

/**
<p>No-argument constructor used during unmarshalling</p>
<p>Call init after unmarshalling to complete initialisation.</p>
**/
  public GbtxControl(){super();}

/**
<p>Writes a byte array to the hardware through the SCA interface.</p>
@param offset Offset to location where to start writing.
@param gbtxBuf Array of register values.
*/
  public void put(int offset, byte [] gbtxBuf)
  {
    sca.putI2c(i2cBus,i2cSlaveAddress,i2cSlaveSpeed,gbtxBuf,offset);
  }
    
/**
<p>Writes a byte array to the hardware through the SCA interface.</p>
@param bb Buffer of register values.
*/
  public void put(java.nio.ByteBuffer bb){ put(0,bb.array()); }
    
/**
<p>Reads a byte array from the hardware through the SCA interface.</p>
@param offset Offset to location where to start reading.
@param gbtxBuf Array of register values.
*/
  public void get(int offset, byte [] gbtxBuf)
  {
    sca.getI2c(i2cBus,i2cSlaveAddress,i2cSlaveSpeed,gbtxBuf,offset);
  }
    
/**
<p>Function called to complete initialisation after unmarshalling.</p>
@param sca The SCA instance
@param dtm The DTM ID
@param i2cBus The I2C bus number
**/
  public void init(ScaDepp sca, int dtm, int i2cBus)
  {
    this.sca = sca;
    this.dtm = dtm;
    this.i2cBus = i2cBus;
    sca.gpioDirection.clear(resetGpio);
    sca.gpioOut.clear(resetGpio);
    setName("dtm"+dtm+".gbtx"+gbtx);
  }

/**
<p>Reset the DTM GBTX.</p>
<p>This is done by toggling the direction of the GPIO as the reset is
defined by a pull-up. The GPIO data out value should always be 0.</p>
**/
  public void reset()
  {
    assertReset();
    deassertReset();
  }

/**
<p>Assert DTM GBTX reset.</p>
**/
  public void assertReset()
  {

    sca.allocateTxRxBuffers(1);
    sca.gpioDirection.set(resetGpio);
    sca.infoTx("GPIO_W_DIRECTION",sca.gpioDirection);
    
    sca.putBytes(0,true);

    try { Thread.sleep(10); } catch (InterruptedException ie) {};

  }

/**
<p>Deassert DTM GBTX reset.</p>
**/
  public void deassertReset()
  {

    sca.allocateTxRxBuffers(1);
    sca.gpioDirection.clear(resetGpio);
    sca.infoTx("GPIO_W_DIRECTION",sca.gpioDirection);
    
    sca.putBytes(0,true);

// The reset is a slowly changing signal (10us) compared to the time
// taken for the SCA to process commands.

    try { Thread.sleep(100); } catch (InterruptedException ie) {};

  }

/**
<p>Perform configuration of the DTM GBTX</p>
<p>Writes the cahed register bits to the hardware.</p>
<p>Program and verify the GBLD registers.</p>
**/
  public void configure()
  {
    try
    {
      byte [] gbtxBuf = java.util.Arrays.copyOfRange(csrBitSet.toByteArray(),0,366);
      put(0,gbtxBuf);

      if (gbldCB.isSelected())
      {
        programGBLD();
        readGBLD();
        byte [] gbldBuf = new byte [7];
        get(381,gbldBuf);
        System.out.println("GBLD registers for DTM"+dtm+".GBTX"+gbtx);
        for (byte b : gbldBuf) System.out.println(String.format("0x%02X",b));
      }
    }
    catch (Exception ex){throw new RuntimeException("GbtxControl.configure(): "+ex.toString());}
  }

}

