package cbsw.lhcb.pdmdb;

import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAttribute;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import javax.swing.*;
import javax.swing.filechooser.FileNameExtensionFilter;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.prefs.Preferences;

/**
<p>This is the abstract base class for the GBTX.</p>
<p>The class maintains a register mirror BitSet that is supposed to
faithfully reflect the GBTX register settings. Therefore, when
updating a GBTX register, the register bits in the mirror are updated
first, then the array of bytes that span the updated register
bit-field are written to the hardware. In this way, any other
registers that share the same bytes are re-written with unchanged
values.</p>
**/
@XmlAccessorType(XmlAccessType.NONE)
abstract class GbtxParent implements cbsw.Io
{
  /**<p>The RegisterIo TX base address</p>*/
  @XmlAttribute(name="txBaseAddress")
  int txBaseAddress;
  /**<p>The RegisterIo RX base address</p><p>Not used by GbtxDepp.</p>*/
  @XmlAttribute(name="rxBaseAddress")
  int rxBaseAddress;
  /**<p>The GBTX hardware address</p><p>Also used as the I2C slave address.</p>*/
  @XmlAttribute(name="id")
  int gbtx;
  /**<p>The MGT to which this GBT is connected.</p>*/
  @XmlAttribute(name="mgt")
  int mgt;

  cbsw.RegisterIo io;
  Preferences prefs;
  String name;
  /**<p>Mirrored GBTX register values</p>*/
  BitSet csrBitSet;
  /**<p>Register settings read from hardware</p>*/
  BitSet rdBitSet;

  /**<p>List of the active input e-groups</p><p>Updated by {@link readConfigurationFile()}.</p>*/
  ArrayList<String> activeInEGroups;
  /**<p>List of the active output e-groups</p><p>Updated by {@link readConfigurationFile()}.</p>*/
  ArrayList<String> activeOutEGroups;
  /**<p>List of the active e-ports</p><p>Updated by {@link readConfigurationFile()}.</p>*/
  ArrayList<String> activeEPorts;
  /**<p>List of the active phase-shifter clocks</p><p>Updated by {@link readConfigurationFile()}.</p>*/
  ArrayList<String> activePSClocks;
    
  private JTextField puStateTF;
  private JTextField cfgTF;
  byte [] gbtxReadBuf;
  JCheckBox gbldCB;

  private JComboBox<String> regNameCB;
  private JTextField regValueTF;

/**
<p>No-argument constructor called by unmarshaller</p>
*/
  public GbtxParent()
  {
    prefs = Preferences.userNodeForPackage(GbtxParent.class).node("GBTX");
    activeInEGroups = new ArrayList<String>();
    activeOutEGroups = new ArrayList<String>();
    activeEPorts = new ArrayList<String>();
    activePSClocks = new ArrayList<String>();
  }

/**
<p>Constructor called by unmarshaller via GbtxAdapter.</p>
@param id GBTX hardware ID.
@param mgt The FPGA MGT connected to this GBTX.
@param txBaseAddress RegisterIo TX base address.
@param rxBaseAddress RegisterIo RX base address.
*/
  public GbtxParent(int id, int mgt, int txBaseAddress, int rxBaseAddress)
  {
    gbtx = id;
    this.mgt = mgt;
    this.txBaseAddress = txBaseAddress;
    this.rxBaseAddress = rxBaseAddress;
    prefs = Preferences.userNodeForPackage(GbtxParent.class).node("GBTX");
    activeInEGroups = new ArrayList<String>();
    activeOutEGroups = new ArrayList<String>();
    activeEPorts = new ArrayList<String>();
    activePSClocks = new ArrayList<String>();
  }

/**
<p>Initialises the instance</p>
<p>Call after unmarshalling to complete the initialisation.</p>
@param io The RegisterIo interface
*/
  public void init(cbsw.RegisterIo io){ this.io = io; }
    
  public void setPreferencesRoot( Preferences prefs ){ this.prefs = prefs; }
/**
<p>Sets the name used to locate the configuration file and reads the file into the register mirror</p>
@param name The name
*/
  public void setName(String name){this.name = name; readConfigurationFile();}

/**
<p>Convenience method that makes a GUI panel.
@return The panel.
**/
  public JPanel makePanel()
  {
    JPanel jp = new JPanel();
      
    jp.setName( name );
    jp.setBorder( BorderFactory.createTitledBorder(name) );

    jp.setLayout(new GridBagLayout());
    GridBagConstraints gbC = new GridBagConstraints();
    gbC.fill = GridBagConstraints.VERTICAL;
    gbC.anchor = GridBagConstraints.CENTER;
    gbC.insets = new java.awt.Insets(1,1,1,1);
    gbC.weightx = 1.0f;
    gbC.weighty = 1.0f;
    gbC.fill = GridBagConstraints.BOTH;

    JButton button;
    
    button = new JButton("Save As...");
    button.addActionListener(a->Gbtx.write(csrBitSet));
    gbC.gridwidth = 1;
    jp.add(button,gbC);

    button = new JButton("Choose file");
    button.addActionListener(a->chooseConfigurationFile());
    gbC.gridwidth = 1;
    jp.add(button,gbC);

    Preferences prefs = Preferences.userNodeForPackage(GbtxParent.class).node("GBTX");
    cfgTF = new JTextField(prefs.get(name+".file-name","gbtx.gbt"));
    cfgTF.setEditable(false);
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    jp.add( cfgTF, gbC );
    
    button = new JButton("Set");
    button.addActionListener(l->set((String)regNameCB.getSelectedItem(),regValueTF.getText()));
    gbC.gridwidth = 1;
    jp.add(button,gbC);

    regNameCB = new JComboBox<String>();
    regNameCB.setEditable(true);
    regNameCB.addItem("eport\\.pa\\.[0123456]\\.config");
    regNameCB.addItem("eport\\.pa\\.[0123456]\\.reset");
    regNameCB.addItem("eport\\.pa\\.[0123456]\\.[04]\\.phase");
    regNameCB.addItem("[bypassXpll]");
    regNameCB.addItem("[enableXpll]");
    regNameCB.addItem("[minimal]");
    regNameCB.addItem("[setGbld]");
    regNameCB.addItem("[ecEport]");
    regNameCB.addItem("[clear]");
    gbC.gridwidth = 1;
    jp.add(regNameCB,gbC);
    
    regValueTF = new JTextField("0");
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    jp.add(regValueTF,gbC);

    button = new JButton("Reset");
    button.addActionListener(a->reset());
    gbC.gridwidth = 1;
    jp.add(button,gbC);

    //    button = new JButton(" Program GBLD");
    //    button.addActionListener(a->programGBLD());
    //    gbC.gridwidth = 1;
    //    jp.add(button,gbC);

    button = new JButton("Configure");
    button.addActionListener(a->configure());
    gbC.gridwidth = 1;
    jp.add(button,gbC);

    gbldCB =  new JCheckBox("GBLD",false);
    gbC.gridwidth = 1;
    jp.add( gbldCB, gbC );
  
    button = new JButton("Decode");
    button.addActionListener(a->Gbtx.decode(rdBitSet));
    //    button.addActionListener(a->Gbtx.decode(csrBitSet));
    gbC.gridwidth = 1;
    jp.add(button,gbC);

    button = new JButton("Verify");
    button.addActionListener(a->verify());
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    jp.add(button,gbC);

    puStateTF = new JTextField(10);
    puStateTF.setEditable(false);
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    jp.add( puStateTF, gbC );
    
    return jp;
  }

/**
<p>Resets the GBTX</p>
 */
public abstract void reset();

/**
<p>Writes a register bit-field to the GBTX</p>
@param regName The register bit-field key name
@param regValue Hex representation of value to write
  */
public void put(String regName, String regValue)
{
  cbsw.RegisterDescriptor reg = Gbtx.map.get(regName);
  reg.set(csrBitSet,regValue);

  String [] offsets = reg.getOffset().split(",");
  for (String offsetString : offsets)
  {
    int offset = Integer.valueOf(offsetString);
    byte [] regBuf = reg.getBytes(csrBitSet,offset);
    try { put(offset/8,regBuf); }
    catch (Exception ex){throw new RuntimeException("GbtxParent.put(): "+ex);}
  }
    
  return;
}

/**
<p>Reads a register bit-field from the GBTX</p>
@param regName The register bit-field key
@return Hex encoded String register value
 */
public String get(String regName)
{
  cbsw.RegisterDescriptor reg = Gbtx.map.get(regName);
  
  String [] offsets = reg.getOffset().split(",");
  int offset = Integer.valueOf(offsets[0]);
  int length = reg.getLength();
  int byteLow = offset/8;
  int byteHigh = (offset+length-1)/8;
  byte [] rdBuf = new byte [byteHigh-byteLow+1];
  try { get(byteLow,rdBuf); }
  catch (Exception ex){throw new RuntimeException("GbtxParent.get(): "+ex);}

// Shift the bits down to the low byte boundary
  int bitShiftDown = offset - 8*byteLow;
  BitSet rdBits = BitSet.valueOf(rdBuf);
  BitSet bits = rdBits.get(bitShiftDown,bitShiftDown+length);

// Convert to hexadecimal string
  StringBuffer sb = new StringBuffer();
  byte [] result = Arrays.copyOfRange(bits.toByteArray(),0,(length+7)/8);
  for ( byte b: result ) sb.insert(0,String.format("%02x",b));

  return sb.toString();
}

/*
<p>Writes register bit-field to GBTX</p>
<p>Called by the GUI panel</p>

private void set(){  set((String)regNameCB.getSelectedItem(),regValueTF.getText()); }
 */    
/**
<p>Writes a register bit-field to the GBTX</p>
<p>Field keys are treated as regular expressions for bit-field names
except for some special cases when the name is enclosed in []. These
update the register mirror with predefined settings but do not write
the settings to the GBTX.</p>
@param key The register bit-field key name
@param val The value to write
*/ 
  public void set(String key, String val)
  {
    if (key.equals("[bypassXpll]"))
    {
      bypassXpll();
    }
    else if (key.equals("[enableXpll]"))
    {
      enableXpll();
    }
    else if (key.equals("[minimal]"))
    {
      minimal();
    }
    else if (key.equals("[setGbld]"))
    {
      setGbld();
    }
    else if (key.equals("[ecEport]"))
    {
      ecEport();
    }
    else if (key.equals("[clear]"))
    {
      csrBitSet.clear();
    }
    /*
    else if (key.matches("eport\\.pa\\.\\*\\..*"))
    {
      for (int group=0;group<7;group++) put(key.replace("*",String.valueOf(group)),val);
    }
    */
    else
    {
      System.err.println(key);
      Gbtx.descriptors.stream().filter(reg->reg.getName().matches(key))
                               .collect(java.util.stream.Collectors.toList())
                               .forEach(reg->reg.put(csrBitSet,val,this));
      /*put(key,val);*/
    }
  }
    
/**
<p>Updates GBLD register bits of the mirrored registers</p>
<p>Does not write the values to the GBTX.</p>
**/
  public void setGbld()
  {
    Gbtx.get("gbld.slave-address").set(csrBitSet,"7E");
    Gbtx.get("gbld.control").set(csrBitSet,"87");
    Gbtx.get("gbld.pre-emphasis.width").set(csrBitSet,"2");
    Gbtx.get("gbld.modulation-current").set(csrBitSet,"20");
    Gbtx.get("gbld.bias-current").set(csrBitSet,"20");
    Gbtx.get("gbld.pre-emphasis.rising-height").set(csrBitSet,"8");
    Gbtx.get("gbld.pre-emphasis.falling-height").set(csrBitSet,"8");
    Gbtx.get("gbld.pre-emphasis.mask").set(csrBitSet,"3");
    Gbtx.get("gbld.modulation-current.mask").set(csrBitSet,"3F");
    Gbtx.get("gbld.bias-current.mask").set(csrBitSet,"FF");
    //    Gbtx.get("gbld.pre-driver.state").set(csrBitSet,"0"); // Read-only
    Gbtx.get("gbld.pre-driver.bias").set(csrBitSet,"4");
  }
    
/**
<p>Updates the mirrored register bits that enable the Xpll.</p>
<p>Does not write the values to the GBTX.</p>
**/
  public void enableXpll()
  {
    Gbtx.get("xpll.enable").set(csrBitSet,true);
    Gbtx.get("xpll.phase-detector.enable").set(csrBitSet,true);
    Gbtx.get("xpll.auto-restart.enable").set(csrBitSet,false);
    Gbtx.get("xpll.control-override").set(csrBitSet,true);
    //    Gbtx.get("xpll.frequency-trim").set(csrBitSet,"50");
    Gbtx.get("xpll.gm-select").set(csrBitSet,"A");
    Gbtx.get("xpll.cp").set(csrBitSet,"B");
    Gbtx.get("cm.refclk.select").set(csrBitSet,"1");
    return;
  }
    
/**
<p>Updates the mirrored register bits that bypass the Xpll.</p>
<p>Does not write the values to the GBTX.</p>
**/
  public void bypassXpll()
  {
    Gbtx.get("xpll.enable").set(csrBitSet,false);
    Gbtx.get("xpll.phase-detector.enable").set(csrBitSet,false);
    Gbtx.get("xpll.auto-restart.enable").set(csrBitSet,false);
    Gbtx.get("xpll.control-override").set(csrBitSet,true);
    Gbtx.get("xpll.frequency-trim").set(csrBitSet,"5D");
    Gbtx.get("xpll.gm-select").set(csrBitSet,"A");
    Gbtx.get("xpll.cp").set(csrBitSet,"0");
    Gbtx.get("cm.refclk.select").set(csrBitSet,"1");
    return;
  }
    
/**
<p>Overwrites the mirrored register bits with the minimal set</p>
<p>Does not write the values to the GBTX.</p>
<p>The minimal set should be enough to get the TCM link alive on power up.</p>
**/
  public void minimal()
  {
    csrBitSet.clear();
      //    Gbtx.get("xpll.enable").set(csrBitSet,false);
      //    Gbtx.get("xpll.phase-detector.enable").set(csrBitSet,false);
      //    Gbtx.get("xpll.auto-restart.enable").set(csrBitSet,false);
    Gbtx.get("watchdog.enable").set(csrBitSet,true);
    Gbtx.get("watchdog.startup-timeout-enable").set(csrBitSet,true);
    Gbtx.get("xpll.control-override").set(csrBitSet,true);
    Gbtx.get("xpll.frequency-trim").set(csrBitSet,"5D");
    Gbtx.get("xpll.gm-select").set(csrBitSet,"A");
    //    Gbtx.get("xpll.cp").set(csrBitSet,"0");
    Gbtx.get("cm.refclk.select").set(csrBitSet,"1");
    
    Gbtx.get("serialiser.lock-time").set(csrBitSet,"6");
    Gbtx.get("serialiser.unlock-time").set(csrBitSet,"3");
    Gbtx.get("serialiser.current").set(csrBitSet,"8");
    Gbtx.get("serialiser.resistor").set(csrBitSet,"2");
    Gbtx.get("tx.select-positive-edge").set(csrBitSet,"1");
    
    //    Gbtx.get("deserialiser.gain").set(csrBitSet,"0");
    Gbtx.get("deserialiser.resistor").set(csrBitSet,"d");
    Gbtx.get("rx.valid-headers").set(csrBitSet,"f");
    Gbtx.get("rx.max-invalid-headers").set(csrBitSet,"4");
    Gbtx.get("rx.min-valid-headers").set(csrBitSet,"8");
    Gbtx.get("rx.select-i1").set(csrBitSet,"2");
    
    Gbtx.get("rx.frequency-detector").set(csrBitSet,"2");
    Gbtx.get("rx.phase-detector").set(csrBitSet,"4");
    
    Gbtx.get("rx.switch.0").set(csrBitSet,"1");
    Gbtx.get("rx.switch.1").set(csrBitSet,"1");
    Gbtx.get("rx.switch.2").set(csrBitSet,"1");
    Gbtx.get("tx.switch.0").set(csrBitSet,"1");
    Gbtx.get("tx.switch.1").set(csrBitSet,"1");
    Gbtx.get("tx.switch.2").set(csrBitSet,"1");
    //    Gbtx.get("ps.pll.resetb").set(csrBitSet,"1");
    //    Gbtx.get("ps.pll.charge-pump").set(csrBitSet,"f");
    //    Gbtx.get("ps.pll.resistor").set(csrBitSet,"4");
    //    Gbtx.get("ps.pll.enable").set(csrBitSet,"1");
    Gbtx.get("efuse.update.enable").set(csrBitSet,"1");

    // Its probably very convenient for the EC e-port to come alive at power up too
    ecEport();

    //    Gbtx.get("cfg.do").set(csrBitSet,"aa");
    return;
  }

/**
<p>Updates the mirrored register bits that enable the EC ePort</p>
<p>Does not write the values to the GBTX.</p>
**/
  public void ecEport()
  {
    Gbtx.get("eport.pa.ec.config").set(csrBitSet,"d");
    Gbtx.get("eport.pa.ec.enable").set(csrBitSet,"1");
    Gbtx.get("eport.pa.ec.termination").set(csrBitSet,"1");
    Gbtx.get("eport.pa.coarse-lock").set(csrBitSet,"1");
    Gbtx.get("eport.ec.mode").set(csrBitSet,"1");
  }

/**
<p>Selects the GBTX register file and reads the settings into the
register mirror</p>
**/
  public void chooseConfigurationFile()
  {
    Preferences prefs = Preferences.userNodeForPackage(GbtxParent.class).node("GBTX");
    java.io.File file = cbsw.Cbsw.chooseFile(prefs.get("chooser.gbtx.dir","."),new FileNameExtensionFilter( "GBTX config files (GBT)", "gbt" ));
    prefs.put("chooser.gbtx.dir",file.getParent());
    prefs.put(name+".file-name",file.getAbsolutePath());
    cfgTF.setText(prefs.get(name+".file-name","gbtx.gbt"));
    readConfigurationFile();
  }

/**
<p>Reads the GBTX register file int the register mirror and builds
lists of active e-links and clocks</p>
**/
  public void readConfigurationFile()
  {
    Preferences prefs = Preferences.userNodeForPackage(GbtxParent.class).node("GBTX");
    System.err.println("GbtxParent.readConfigurationFile(): "+prefs.get(name+".file-name","gbtx.gbt"));
    csrBitSet = Gbtx.read(prefs.get(name+".file-name","gbtx.gbt"));

// Update list of active input eGroups
    activeInEGroups.clear();
    for (int egroup=0;egroup<7;egroup++)
    {
      String key = "eport.pa."+egroup;
      if (!Gbtx.get(key+".rate").get(csrBitSet).equals("00") ) activeInEGroups.add(key);
    }
    for (String egroup : activeInEGroups) System.err.println("GbtxParent.readConfigurationFile(): Input eGroup "+egroup+" is active");

// Update list of active output eGroups
    activeOutEGroups.clear();
    for (int egroup=0;egroup<5;egroup++)
    {
      String key = "eport."+egroup;
      if (!Gbtx.get(key+".rate").get(csrBitSet).equals("00") ) activeOutEGroups.add(key);
    }
    for (String egroup : activeOutEGroups) System.err.println("GbtxParent.readConfigurationFile(): Output eGroup "+egroup+" is active");

// Update list of active ePorts
    activeEPorts.clear();
    for (int egroup=0;egroup<7;egroup++)
    {
      for (int eport=0;eport<8;eport++)
      {
	String key = "eport.pa."+egroup+"."+eport;
	if (Gbtx.get(key+".enable").get(csrBitSet).equals("01") ) activeEPorts.add(key);
      }
    }
    for (String eport : activeEPorts) System.err.println("GbtxParent.readConfigurationFile(): ePort "+eport+" is active");

// Update list of active phase shifter clocks
    activePSClocks.clear();
    for (int clock=0;clock<8;clock++)
    {
      String key = "ps."+clock;
      if (!Gbtx.get(key+".cset").get(csrBitSet).equals("0f") ) activePSClocks.add(key);
    }
    for (String clock : activePSClocks) System.err.println("GbtxParent.readConfigurationFile(): PS clock "+clock+" is active");
  }

/**
<p>Updates the GBTX status field in the GUI panel</p>
@return Status summary string
**/
  public String showStatus()
  {
    int stateInt = Integer.parseInt(Gbtx.get("powerup-fsm-state").get(rdBitSet),16);
    String state;
    try { state = Gbtx.fsmStates.get( stateInt ); }
    catch (IndexOutOfBoundsException oobe){state = "Unknown"; System.err.println("GbtxParent.showStatus: Unknown state "+stateInt);}
    String serialNo = Gbtx.get("fuse.serial").get(rdBitSet);
    String xPllLock = Gbtx.get("xpll.locked").get(rdBitSet);  
    puStateTF.setText(String.format("[0x%s] %s xPllLock=%s",serialNo,state,xPllLock));
    return "0x" + serialNo + "|" + state + "|" + xPllLock;
  }
    
/**
<p>Writes the register mirror bits to the hardware</p>
**/
  void load()
  {

// Pack gbtx settings into array

    int buflen = 366; byte [] gbtxBuf = new byte [buflen];
    gbtxBuf = Arrays.copyOfRange(csrBitSet.toByteArray(),0,buflen);
    try { put(0,gbtxBuf); }
    catch (Exception ex){throw new RuntimeException("GbtxParent.load(): "+ex);}
        
    return;
  }
    
/**
<p>Configures the GBTX</p>
<p>Disables the watchdog and timeout feature duing the configuration
otherwise a timeout might cause the volatile registers to be
reset.</p>
<p>This puts the downlinks in 320Mbs mode since this is required for
the module tester use case for phase training.</p>
<p>Performs a soft reset.</p>
**/
public void configure()
{
  System.out.println("GbtxParent.configure(): Writing master GBTX configuration");
  
// Disable watchdog and timeout features 
  put("watchdog.enable","0");
  put("watchdog.startup-timeout-enable","0");

  load();

  //  configureEpll("3"); // Required for phase scans with module tester hardware
  
  softReset();

// Re-enable the watchdog and timeout feature
  put("watchdog.enable","1");
  put("watchdog.startup-timeout-enable","1");
}


/**
<p>Configures the Epll for all active egroups</p>
<p>Sets the PLL parameters, sets the desired data rate, resets the PLL.</p>
<p>Watchdog and timeout should be disabled during this operation.</p>
@param rate The data rate
 */
public void configureEpll(String rate)
{
// Set the configuration registers
  put("epll.rx.cp","b");
  put("epll.rx.resistor","2");
  put("epll.rx.capacitor","2");
  put("epll.rx.phase","FF");
  for (String egroup : activeOutEGroups) put(egroup+".rate","3");

  // Reset it
  put("epll.rx.reset","1");
  put("epll.rx.reset","0");
      
}

/**
<p>Reads configuration register bits from the GBTX</p>
**/
  public void readConfiguration()
  {
    try
    {
      gbtxReadBuf = new byte [436];
      get(0,gbtxReadBuf);
      rdBitSet = BitSet.valueOf(gbtxReadBuf);
    }
    catch (Exception ex){throw new RuntimeException("GbtxParent.readConfiguration(): "+ex);}
    return;
  }
    
/**
<p>Reads and verifies the GBTX registers against the mirrored values</p>
@return Verification status
**/
  public String verify()
  {

    boolean ok = true;
    try
    {
      readConfiguration();
      String status = showStatus();
      
// Compare gbtx settings

      byte [] gbtxBuf = Arrays.copyOfRange(csrBitSet.toByteArray(),0,367);

      int reg=0;
      for (byte val: gbtxBuf)
      {
        if (val != gbtxReadBuf[reg])
        {
	  ok = false;
	  System.err.println(String.format("[%03d] %02X!=%02X",reg,gbtxReadBuf[reg],val));
	}
        reg++;
      }

      // Now show all non-zero bytes
      reg=0;
      System.err.println("--- Non-zero bytes ---");
      for (byte val: gbtxBuf)
      {
        if (gbtxReadBuf[reg]!=0)
        {
	  System.err.println(String.format("[%03d,%03x] %02X",reg,reg,gbtxReadBuf[reg]));
	}
        reg++;
      }
      System.err.println("----------------------");
      
      return (ok ? "OK" : "FAIL") + "|" + status;
    }
    catch (Exception ex){ throw new RuntimeException("GbtxParent.verify(): "+ex); }
  }

/**
<p>Triggers the programming of the GBLD</p>
<p>Note: The GBLD I2C address (0x7e) must be preloaded into GBTX reg 253.</p>
<p>Trigger the transfer of the GBLD registers from the GBTX registers.</p>
**/
  public void programGBLD()
  {
    byte [] putBuf = new byte [1];
    try { put(388,putBuf); }
    catch (Exception ex){throw new RuntimeException("GbtxParent.programGBLD(): "+ex);}
    try { Thread.sleep(500); } catch (InterruptedException ie) {};

    return;
  }

/**
<p>Triggers the reading of the GBLD</p>
<p>Note: The GBLD I2C address (0x7e) must be preloaded into GBTX reg 253.</p>
<p>Trigger the transfer of the GBLD registers into the GBTX registers.</p>
**/
  public void readGBLD()
  {
    byte [] putBuf = new byte [1];
    try { put(389,putBuf); }
    catch (Exception ex){throw new RuntimeException("GbtxParent.readGBLD(): "+ex);}

    try { Thread.sleep(500); } catch (InterruptedException ie) {};

    return;
  }

/**
<p>A register read/write test</p>
<p>Experts only. May have unexpected side-effects and may change
without notice.</p>
*/
  public void test()
  {
    System.err.println("GBTX r/w test");

    try
    {
      for (byte i=0;i<5;i++)
      {

	byte [] bb = {i,i,i};
	put(238,bb); // Fusing registers
        try { Thread.sleep(50); } catch (InterruptedException ie) {};
// Check the values were written correctly

	byte [] chk = new byte [3];
        get(238,chk);
        try { Thread.sleep(50); } catch (InterruptedException ie) {};
	if (chk[0] != i) System.err.println("Fuse address LSB error "+i+" "+chk[0]);
	if (chk[1] != i) System.err.println("Fuse address MSB error "+i+" "+chk[0]);
	if (chk[2] != i) System.err.println("Fuse value error "+i+" "+chk[0]);
      }
    }
    catch (Exception ex){throw new RuntimeException("GbtxParent.test(): "+ex);}
  }

/**
<p>Performs a reset of the GBTX phase shifter logic</p>
<p>When calling this it is advisable/mandatory to disable the
watchdog/timeout feature.</p>
*/
public void resetPhaseShifter()
{
  put("ps.pll.resetb","0"); // Assert phase shifter PLL reset, active low
  for (String clock : activePSClocks) put(clock+".resetb","0"); // Assert phase shifter DLL reset, active low
  
  put("ps.pll.resetb","1"); // Deassert phase shifter PLL reset, active low

  try {Thread.sleep(10);} catch (InterruptedException ie){} // Time for PS PLL to lock before releasing DLL resets
  
  for (String clock : activePSClocks) put(clock+".resetb","1"); // Deassert Phase shifter DLL reset, active low
}

/**
<p>Performs a reset of the active GBTX ePorts</p>
<p>When calling this it is advisable/mandatory to disable the
watchdog/timeout feature.</p>
*/
public void resetEPorts()
{
  for (String egroup : activeInEGroups) put(egroup+".reset","1"); // Assert egroup PA DLL reset
  for (String eport : activeEPorts) put(eport+".reset","1"); // Assert eport reset
  put("eport.pa.ec.reset","1"); // Assert EC PA DLL reset
  
  for (String egroup : activeInEGroups) put(egroup+".reset","0"); // Deassert egroup PA DLL reset
  put("eport.pa.ec.reset","0"); // Deassert EC PA DLL reset

  try {Thread.sleep(10);} catch (InterruptedException ie){} // Time for PS PLL and PA DLLs to lock before releasing DLL resets

  for (String eport : activeEPorts) put(eport+".reset","0"); // Deassert eport reset
}

/**
<p>Performs a soft reset of the GBTX</p>
<p>When fused with the minimal configuration the GBTX startup
completes with many registers not yet programmed.  This means that
when the remaining registers are set parts of the GBTX need to be
manually reset.</p>
<p>This method manually triggers the required resets.</p>
<p>When calling this it is advisable/mandatory to disable the
watchdog/timeout feature.</p>
*/
public void softReset()
{
  resetPhaseShifter();
  resetEPorts();
}

/**
<p>Performs the automatic GBTX ePort phase training.</p>
<p>The GBTX phase scan range is 0-14 with each step UI/8 so almost
2*UI in total. This means that there are two sampling positions that
should give stable data. Empirically, the GBTX does not always choose
the same UI for all of the e-links so this method applies GBTXphase%8
to force the value into the range 0-7. This approach gives good
results for the PDMDB.</p>
@return Array of phases
*/
  public byte [] trainEports()
  {
    put("eport.pa.mode","1"); // Switch to training mode
    for (String eport : activeEPorts) put(eport+".train","1");

    try {Thread.sleep(200);} catch (InterruptedException ie){}
	
    for (String eport : activeEPorts) put(eport+".train","0");

    try {Thread.sleep(100);} catch (InterruptedException ie){}
	
    byte [] trainBuf = new byte [28];
    try { get(399,trainBuf); }
    catch (Exception ex){throw new RuntimeException("GbtxParent.trainEports(): "+ex);}

    StringBuffer sb = new StringBuffer();
    sb.append("Phases: ");
    for (int group=0;group<7;group++) sb.append(String.format("%d.0=%d %d.4=%d ",group,trainBuf[4*group]&0xf,group,trainBuf[4*group+2]&0xf));
    sb.append(String.format("\n"));

    System.err.println(sb.toString());
	
    put("eport.pa.mode","0"); // Back to static mode

// Apply the training

    for (int group=0;group<7;group++)
    {
      put("eport.pa."+group+".0.phase",String.format("%X",(trainBuf[4*group]&0xf)%8));
      put("eport.pa."+group+".4.phase",String.format("%X",(trainBuf[4*group+2]&0xf)%8));
    }

    return trainBuf;
  }

}

