package cbsw.lhcb.pdmdb;

import javax.xml.bind.annotation.*;
import java.awt.*;
import javax.swing.*;
import java.io.*;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.*;
import cbsw.gui.*;

/**
<p>This class encapsulates the I/O interface with the SCA ASIC.</p>
<p>[For historical reasons this class is called ScaDepp. It was developed using the Digilent Adept DEPP host API but it is compatible with any {@link cbsw.RegisterIo} interface, not only DEPP.]</p>
<p>Transactions with the SCA are through an HDLC command-response protocol. For efficiency, this class buffers a sequence of SCA commands which are sent as a single IO transaction to muDaq through the host interface.</p>
<p>A muDaq HDLC engine sends the commands to the SCA, one at a time, waiting for the response before sending the next. The responses are buffered by muDaq and can be read through the host interface in a single transaction. There is a one-to-one correspondence between the 16-byte command buffer slot and 16-byte response buffer slot.</p>
<p>{@link Sca} and {@link ScaRx} implement a list and map of the possible commands and responses. These descriptor lists are unmarshalled from XML files.</p>
<p>A typical code snippet for a transaction might be:</p>
<pre>
{@code
  allocateTxRxBuffers(6); // Allocate 6 slots
  infoTx("NC_W_CRB",0x3e); // Command 1 of 6
  infoTx("NC_W_CRC",0x00);
  infoTx("NC_W_CRD",0x38);
  infoTx("NC_R_CRB",0x00);
  infoTx("NC_R_CRC",0x00);
  infoTx("NC_R_CRD",0x00); // Command 6 of 6
  putBytes( 0, true ); // Send the commands to muDaq
  cache(); // Read and cache the responses buffer
  byte [] dataB = getResponseBytes("NC_R_CRB"); // Extract the data
  byte [] dataC = getResponseBytes("NC_R_CRC"); // Extract the data
  byte [] dataD = getResponseBytes("NC_R_CRD"); // Extract the data
}
</pre>
<p>For simple transactions like the above, {@link getResponseBytes(java.lang.String)} is a convenient way to extract the data from a named response. For larger or more complex tansactions it may be more appropriate to decode the {@link hdlcrx} response cache directly.</p>
*/
@XmlAccessorType(XmlAccessType.NONE)
class ScaDepp
{
/**<p>Statically accessible instance of this class.</p>*/
  static ScaDepp staticThis;
  private cbsw.RegisterIo io;
  @XmlAttribute(name="deppBaseRx")
  private int deppBaseRx;
  @XmlAttribute(name="deppBaseTx")
  private int deppBaseTx;
  private byte hdlcSeqTx = 0;
  @XmlAttribute(name="hdlcAddress")
  private byte scaHdlcAddress;
  @XmlAttribute(name="id")
  private byte scaId;
  @XmlAttribute(name="spiFrequency")
  private int spiFrequency;

/**<p>Buffer that holds a batch of SCA commands.</p>*/
  ByteBuffer scaTx;

//  java.util.stream.Stream.Builder<String> txStreamBuilder;
//  java.util.stream.Stream<String> txStream;

  private ArrayList<String> responseList;

/**<p>Flag used to enable or disable debug to the GUI text area (or console).</p>*/
  boolean debug = false;

  private JFrame scaExpertFrame;
  private JComboBox<String> groupCB, commandCB, channelCB;
  private JTextField [] dataTF;
  private JTextField rawTF;
  private JTextArea textArea;
  private JScrollPane textSP;

/**<p>Desired state of the GPIO outputs.</p>*/
  BitSet gpioOut;
/**<p>Desired settings of the GPIO direction.</p>*/
  BitSet gpioDirection;
/**<p>Desired settings of the ADC current source enables.</p>*/
  BitSet adcCurrentSource;

/**<p>The SCA response buffer cache.</p>*/
  byte [] hdlcrx;
/**<p>The current slot in the SCA response buffer cache.</p>*/
  int responseIndex;

/**
<p>No-argument constructor called by unmarshaller.</p>
<p>Sets {@link staticThis} to this instance.</p>
<p>Creates empty {@link gpioOut}, {@link gpioDirection} and {@link adcCurrentSource} BitSets that mirror the state of the corresponding SCA registers. The BitSets are populated during unmarshalling.</p>
 */
  public ScaDepp()
  {
    staticThis = this;
    gpioDirection = new BitSet(32);
    gpioOut = new BitSet(32);
    adcCurrentSource = new BitSet(32);
    responseList = new ArrayList<String>();
  }

/**
<p>Called after unmarshalling to complete initialisation.</p>
@param io The I/O interface.
 */
  public void init(cbsw.RegisterIo io){ this.io = io; }

/**
<p>Makes a standard GUI panel</p>
@return The panel
 */
  public GuiJPanel makePanel()
  {
    GuiJPanel jP = new GuiJPanel(java.util.prefs.Preferences.userNodeForPackage(ScaDepp.class).node("SCA"),"");

    scaExpertFrame = GuiJPanel.makeFrame(makeExpertPanel(),"SCA expert panel");
    scaExpertFrame.pack();

    jP.setBorder( BorderFactory.createTitledBorder("GBT-SCA commands") );
    jP.setLayout( new GridBagLayout() );
    GridBagConstraints gbC = new GridBagConstraints();
    gbC.fill = GridBagConstraints.VERTICAL;
    gbC.anchor = GridBagConstraints.CENTER;
    gbC.insets = new Insets(1,1,1,1);
    gbC.weightx = 1.0f;
    gbC.weighty = 1.0f;
    gbC.fill = GridBagConstraints.BOTH;

    JButton button = new JButton("Connect");
    button.addActionListener(a->connect());
    gbC.gridwidth = 1;
    jP.add(button,gbC);

    button = new JButton("Reset");
    button.addActionListener(a->reset());
    gbC.gridwidth = 1;
    jP.add(button,gbC);

    button = new JButton("Setup");
    button.addActionListener(a->{enableInterfaces();setup();});
    gbC.gridwidth = 1;
    jP.add(button,gbC);

    button = new JButton("Expert");
    button.addActionListener(a->scaExpertFrame.setVisible(!scaExpertFrame.isVisible()));
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    jP.add(button,gbC);

    return jP;
  }

/**
<p>Makes an GUI panel for SCA expert operations.</p>
@return The panel
 */
  public GuiJPanel makeExpertPanel()
  {
    GuiJPanel jP = new GuiJPanel(java.util.prefs.Preferences.userNodeForPackage(ScaDepp.class).node("SCA"),"");

    jP.setBorder( BorderFactory.createTitledBorder("GBT-SCA expert commands") );
    jP.setLayout( new GridBagLayout() );
    GridBagConstraints gbC = new GridBagConstraints();
    gbC.fill = GridBagConstraints.VERTICAL;
    gbC.anchor = GridBagConstraints.CENTER;
    gbC.insets = new Insets(1,1,1,1);
    gbC.weightx = 1.0f;
    gbC.weighty = 1.0f;
    gbC.fill = GridBagConstraints.BOTH;

    JButton button = new JButton("Send");
    button.addActionListener(a->sendCommand());
    gbC.gridwidth = 1;
    jP.add(button,gbC);

    groupCB = new JComboBox<String>();
    groupCB.addItem("NC");
    groupCB.addItem("I2C");
    groupCB.addItem("SPI");
    groupCB.addItem("GPIO");
    groupCB.addItem("JTAG");
    groupCB.addItem("DAC");
    groupCB.addItem("ADC");
    groupCB.addActionListener(a->updateCommandCB());
    gbC.gridwidth = 1;
    jP.add(jP.addBorder(groupCB, "Group"),gbC);

    commandCB = new JComboBox<String>();
    gbC.gridwidth = 1;
    jP.add(jP.addBorder(commandCB, "Command"),gbC);

// For I2C it is necessary to also select the channel
    channelCB = new JComboBox<String>();
    gbC.gridwidth = 1;
    jP.add(jP.addBorder(channelCB, "Channel"),gbC);

    updateCommandCB();

    JPanel dJp = new JPanel();
    dJp.setBorder( BorderFactory.createTitledBorder("D[0..3]") );
    dataTF = new JTextField [4];

    for ( int i=0;i<4;i++ )
    {
      dataTF[i] = new JTextField("0x00");
      dataTF[i].setToolTipText("Prefix 0x, 0X or # for hexadecimal, 0 for octal");
      dJp.add(dataTF[i]);
    }

    gbC.gridwidth = GridBagConstraints.REMAINDER;
    jP.add(dJp,gbC);

    rawTF = new JTextField(20);
    gbC.gridwidth = 1;
    jP.add(rawTF,gbC);

    button = new JButton("Send raw");
    button.addActionListener(a->sendRaw(rawTF.getText()));
    gbC.gridwidth = 1;
    jP.add(button,gbC);
      
    JCheckBox enableTextCB = new JCheckBox("Enable text");
    enableTextCB.addActionListener(a->debug=enableTextCB.isSelected());
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    jP.add(enableTextCB,gbC);
      
    textArea = new JTextArea(4,10); // Minimum width. Will expand to fill available width.
    textArea.setEditable(false);
    textArea.setFont(new Font("Monospaced",Font.PLAIN,12));
    textSP = new JScrollPane(textArea);
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    gbC.fill = GridBagConstraints.HORIZONTAL;
    jP.add(textSP,gbC);

    return jP;
  }

  private void sendCommand()
  {
    byte [] buf = new byte [4];
    int i = 0;
    try { for ( JTextField tf : dataTF ) buf[i++] = (byte)Integer.decode(tf.getText()).intValue(); }
    catch ( java.lang.NumberFormatException ne ){}

    String command = (String)commandCB.getSelectedItem();
    allocateTxRxBuffers(1);
    if (command.matches("^I2C.*"))
    {
      int channel = Integer.decode(((String)channelCB.getSelectedItem())).intValue()-3;
      infoTx(command+"/"+channel,buf);
    }
    else
    {
      infoTx(command,buf);
    }

    putBytes( 0, true );

  }
      
  private void updateCommandCB()
  {
    commandCB.removeAllItems();
    String groupName = (String)groupCB.getSelectedItem();
    Sca.descriptors.stream().filter(cmd->cmd.getName().matches("^"+groupName+".*")).forEach(cmd->commandCB.addItem(cmd.getName()));
    
    String commandName = (String)commandCB.getSelectedItem();
    ScaCommandDescriptor commandDescriptor = Sca.map.get(commandName);
    channelCB.removeAllItems();
    String [] channels = commandDescriptor.getChannel().split(",");
    for ( String channel : channels ) channelCB.addItem(channel);
  }

/**
<p>Allocates a buffer to hold SCA commands.</p>
@param nFrames Number of command frames.
 */
  public void allocateTxRxBuffers(int nFrames)
  {
    scaTx = ByteBuffer.allocate(nFrames*16 + 5);
    scaTx.order(ByteOrder.LITTLE_ENDIAN);
    scaTx.position(4);
    responseList.clear();
    responseIndex=0;
//    txStreamBuilder = java.util.stream.Stream.builder();
  }

/**
<p>Resets (clears) an existing SCA command buffer.</p>
 */
  public void resetBuffers()
  {
    scaTx.clear();
    scaTx.position(4);
  }

/**
<p>Displays text in the text area.</p>
@param txt The text to display.
*/
  public void showText(String txt)
  {
    if (textArea == null)
    {
      System.out.println(txt);
    }
    else
    {
      textArea.append(txt+"\n");
    }
  }

/**
<p>Replaces {@link hdlcrx} cache with bytes read from the hardware.</p>
<p>Reads the bytes through the IO interface. The number of bytes read spans the responses in responseList.</p>
*/
  public void cache()
  {
    responseIndex = 0;
    hdlcrx = new byte[16*responseList.size()];
    int tries = 0;
    boolean retry = true;
    while (retry)
    {
      try
      {
	io.get(deppBaseRx,0,hdlcrx);
	retry = false;
      }
      catch (cbsw.RegisterIoException de)
      {
	tries++;
	if (tries>50)
	{
	  retry = false;
	  System.err.println("ScaDepp.cache(): "+de);
	}
      }
    }
  }

/**
<p>Reads bytes from the IO interface.</p>
@param addr Offset to data in the hardware.
@param buf Buffer to receive the data.
@deprecated Use {@link cache()} to update {@link hdlcrx} instead.
*/
  @Deprecated
  public void getBytes(int addr, byte [] buf )
  {
    int tries = 0;
    boolean retry = true;
    while (retry)
    {
      try
      {
	io.get(deppBaseRx,addr,buf);
	retry = false;
      }
      catch (cbsw.RegisterIoException de)
      {
	tries++;
	if (tries>50)
	{
	  retry = false;
	  System.err.println("ScaDepp::getBytes(): "+de);
	}
      }
    }
  }

/**
<p>Displays the current content of the command buffer to System.err.</p>
 */
  public void showBuffer()
  {
    byte [] buf = scaTx.array();
    StringBuffer hexSB = new StringBuffer();
    for (int b=4;b<buf.length;b++)
    {
      hexSB.append(String.format("%1$02x ",buf[b]));
      if (b%16==3)
      {
        System.err.println(hexSB.toString());
        hexSB.setLength(0);
      }
    }
  }
    
/**
<p>Puts bytes to the IO interface.</p>
<p>Adds header information to the TX buffer and performs a single IO transaction to write the buffer to the hardware. The transaction is retried up to 10 times in the event of timeout.</p>
@param addr Offset to data in the hardware.
@param waitflag true=enable an IO wait feature.
*/
  public void putBytes(int addr, boolean waitflag )
  {
//    txStreamBuilder.build().forEach(s->System.err.println(s));
      
// Pad to end with zeros
    int cap = scaTx.capacity();
    for (int pad=scaTx.position();pad<cap;pad++) {scaTx.put( (byte)0 );}

// Write the header
    scaTx.position(0);
    int len = scaTx.capacity() - 4;
    scaTx.putShort( (short)len );
    if (waitflag) addr |= 0x8000;
    scaTx.putShort( (short)addr );

    byte [] buf = scaTx.array();
    if (buf.length%16 != 5){System.err.println("ScaDepp::putBytes(): Wrong buffer length."); return;}
    if (buf[buf.length-1] != 0){System.err.println("ScaDepp::putBytes(): Buffer incorrectly terminated."); return;}

    int tries = 0;
    boolean retry = true;
    while(retry)
    {
      try
      {
	io.put(deppBaseTx, scaTx);
	retry = false;
      }
      catch (cbsw.RegisterIoException de)
      {
	tries++;
	if (tries>10)
	{
	  retry = false;
	  System.err.println("ScaDepp::putBytes(): "+de+" Base="+deppBaseTx+" Addr="+addr+" Length="+buf.length);
	  showBuffer();
	}
      }
    }
  }

/**
<p>Computes the CRC for an SCA command.</p>
@param buf The command bytes.
@return int The computed CRC.
  */
  public int crc(byte [] buf)
  {
    int poly = 0x1021;
    boolean fb;

    int crc = 0xffff;

    for ( int nbyte=0; nbyte<buf.length; nbyte++ )
    {
      for ( int i = 0; i<8; i++ )
      {

        boolean crc15 = (crc & 0x8000) != 0 ? true : false;
        int nbit = i;
        boolean db = (buf[nbyte] & (1<<nbit)) != 0 ? true : false;

        fb = crc15 != db;

        crc = fb ? poly ^ (crc<<1) : crc<<1;
        crc = fb ? (crc | 0x1) : (crc & 0xfffe);
      }
    }

    // Reflect the crc bits

    int crcout = 0;
    for ( int nbit=0; nbit<16; nbit++ )
    {
      int isSet = (crc & (1<<nbit)) != 0 ? 1 : 0;
      crcout <<= 1;
      crcout |= isSet;
    }

    return (crcout&0xffff);

  }

/**
<p>Sends a raw command buffer.</p>
<p>An HDLC info transaction looks like this (each token is one byte):</p>
<code>
SOF,Addr,Control,ID,Ch,Len,Cmd,D0,D1,D2,D3,FCS,EOF
</code>
<p>For SCA, Addr=0, ID=0x00 and 0xFF are reserved.</p>
@param raw Hex encoded command string.
 */
  public void sendRaw( String raw )
  {
    allocateTxRxBuffers(1);
    
    scaTx.put( (byte)((1+raw.length()/2)<<3) ); // Length in bits including this byte

    for (int b=0;b<raw.length()/2;b++)
    {
      scaTx.put( (byte)Integer.parseInt(raw.substring(2*b,2*b+2),16) );
    }

    scaId++;
    if (scaId == -1) scaId = 0x1;
    hdlcSeqTx++;
    if (hdlcSeqTx == 0x8) hdlcSeqTx = 0x0;

    // Write the bytes to the transaction buffer
    putBytes( 0, false );

    return;
  }

/**
<p>Extracts the data from the SCA responses.</p>
<p>The user is responsible for calling {@link cache()} to update {@link hdlcrx} before calling this method.</p>
<p>Performs a search-forward operation through responseList starting at responseIndex until the next instance of the named command is found. Extracts and returns the data from the corresponding slot in {@link hdlcrx}.</p>
<p>Can be called multiple times for each call to {@link cache()}. The search will continue from the position reached by the previous call.</p>
<p>Calls to {@link cache()} or {@link allocateTxRxBuffers(int)} reset responseIndex to the first slot.</p>
@param command SCA command name.
@return The extracted data or null if response not found.
*/
  public byte [] getResponseBytes(String command)
  {
    byte [] response = new byte [4];
    for ( ;responseIndex<responseList.size();responseIndex++ )
    {
      ScaCommandDescriptor rxD = ScaRx.getDescriptor(responseList.get(responseIndex));
      if (rxD.name.equals(command))
      {
	for (int j=0;j<rxD.data.length();j++)
	{
	  if ( rxD.data.charAt(j) == '-' ) continue;
	  int idx = Integer.parseInt(rxD.data.substring(j,j+1),16);
	  response[idx%4] = hdlcrx[16*responseIndex+6+j];
	}
	responseIndex++;
	return response;
      }
    }
    return null;
  }

/**
<p>Extracts the data from the SCA responses.</p>
@param command Command name.
@param frame The response frame number.
@param rxbuf The received response data.
@param buf The extracted data.
@deprecated Use {@link getResponseBytes(String)} instead.
 */
  @Deprecated
  public void getResponseBytes(String command, int frame, byte [] rxbuf, byte [] buf)
  {
    ScaCommandDescriptor sca = ScaRx.getDescriptor( command );

    for (int j=0;j<sca.data.length();j++)
    {
      if ( sca.data.charAt(j) == '-' ) continue;
      int idx = Integer.parseInt(sca.data.substring(j,j+1),16);
      buf[idx%4] = rxbuf[16*frame+6+j];
    }
  }

/**
<p>Appends an SCA command to the TX buffer.</p>
@param command Command name.
@param info Command data encoded as an int.
@see infoTx(String,byte [])
@see putBytes(int,boolean)
*/
  public void infoTx( String command, int info)
  {
    byte [] buf = new byte [4];
    buf[0] = (byte)info;
    buf[1] = (byte)(info>>8);
    buf[2] = (byte)(info>>16);
    buf[3] = (byte)(info>>24);
    infoTx(command,buf);
    return;
  }

/**
<p>Appends an SCA command to the TX buffer.</p>
@param command Command name.
@param info Command data encoded as a BitSet.
@see infoTx(String,byte [])
@see putBytes(int,boolean)
*/
  public void infoTx( String command, BitSet info)
  {
    byte [] buf = Arrays.copyOfRange(info.toByteArray(),0,4);
    infoTx(command,buf);
    return;
  }

/**
<p>Appends an SCA command to the TX buffer.</p>
<p>An HDLC info transaction looks like this (each token is one byte):</p>
<code>
SOF,Addr,Control,ID,Ch,Len,Cmd,D0,D1,D2,D3,FCS,EOF
</code>
<p>For the SCA, Addr=0. ID=0x00 and 0xFF are reserved.</p>
@param command Command name.
@param info Command data.
@see putBytes(int,boolean)
 */
  public void infoTx( String command, byte [] info )
  {
//    txStreamBuilder.add(command);

    ScaCommandDescriptor sca = Sca.getDescriptor(command);
    ScaCommandDescriptor scaDescRx = ScaRx.getDescriptor(command);
    responseList.add(command);

    // Align the start position to next 16 byte boundary

    int startBufAddr = ((scaTx.position()+11)/16)*16+4;
    scaTx.position(startBufAddr);
    int hdlclen = 7+sca.data.length();

    scaTx.put( (byte)(hdlclen<<3) ); // Length in bits including this byte
    scaTx.put( scaHdlcAddress );
    byte hdlcSeq = (byte)((hdlcSeqTx&0x7) << 5);
    hdlcSeq |= ((hdlcSeqTx&0x7) << 1);
    scaTx.put(hdlcSeq);
    scaTx.put( scaId ); // ID
    Byte i2cChannel = Sca.getI2cChannel(command);
    if (i2cChannel == null)
    {
      scaTx.put( Integer.decode(sca.channel).byteValue() );
    }
    else
    {
      scaTx.put( i2cChannel );
    }
    scaTx.put( sca.length );
    scaTx.put( sca.code );

    if (sca.data.length()>4) {System.err.println("ScaDepp.infoTx(): Invalid tx info descriptor length.");}

    for (int j=0;j<sca.data.length();j++)
    {
      if ( sca.data.charAt(j) == '-' )
      {
	scaTx.put( (byte)0 );
      }
      else
      {
        int idx = Integer.parseInt(sca.data.substring(j,j+1),16);
        scaTx.put( info[idx%4] );
      }
    }

// The RX index array locates the returned data within a frame
    if (scaDescRx.data.length()>4) {System.err.println("ScaDepp.infoTx(): Invalid rx info descriptor length.");}

    byte [] scaBuffer = new byte [hdlclen];
    scaTx.position(startBufAddr);
    scaTx.get(scaBuffer,0,hdlclen);

    if (debug)
    {
      StringBuffer hexSB = new StringBuffer();
      for ( int j=0;j<hdlclen;j++ ) hexSB.append(String.format("%1$02x",scaBuffer[j]));
      hexSB.append( " " + command + String.format(" [seq,ch,code,crc]=[%1$02x,%2$02x,%3$02x,%4$04x]",scaBuffer[3],scaBuffer[4],scaBuffer[6],crc(Arrays.copyOfRange(scaBuffer,1,hdlclen))));
      showText( hexSB.toString() );
    }

    scaId++;
    if (scaId == -1) scaId = 0x1;
    hdlcSeqTx++;
    if (hdlcSeqTx == 0x8) hdlcSeqTx = 0x0;

    return;

  }

/**
<p>Sends supervisory frame.</p>
@param ctl The HDLC command.
*/
  public void sfTx( HdlcControl ctl )
  {
    allocateTxRxBuffers(1);
    scaTx.put( (byte)(3<<3) );
    scaTx.put( scaHdlcAddress );
    scaTx.put( ctl.code );

    byte [] scaBuffer = new byte [3];
    scaTx.position(4);
    scaTx.get(scaBuffer,0,scaBuffer.length);

    if (debug)
    {
      StringBuffer hexSB = new StringBuffer();
      for ( int j=0;j<scaBuffer.length;j++ ) hexSB.append(String.format("%1$02x",scaBuffer[j]));
      hexSB.append( " " + ctl.name() + String.format(" [crc]=[%1$04x]",crc(Arrays.copyOfRange(scaBuffer,1,3))));
      showText( hexSB.toString() );
    }

    putBytes( 0x0000, true );

    return;
  }

/**
<p>Sends SCA connect command (UF_SABM).</p>
*/
  public void connect()
  {
    hdlcSeqTx = 0;
    sfTx( HdlcControl.UF_SABM );
  }

/**
<p>Sends SCA reset command (UF_RSET).</p>
*/
  public void reset()
  {
    scaId = 0x1;
    hdlcSeqTx = 0;
    sfTx( HdlcControl.UF_RSET );
  }

/**
<p>Configures the SCA.</p>
<p>Calls {@link reset()}, {@link enableInterfaces()}, {@link setup()} and {@link readId()}.</p>
@return SCA ID.
*/
  public String configure()
  {
    reset();
    if (!enableInterfaces()) throw new RuntimeException( "FAIL|ScaDepp.configure|Failed to enable interfaces");
    setup();
    String scaId = readId();
    return scaId;
  }
    
/**
<p>Initial setup.</p>
<p>Configures some global settings needed by the PDM.</p>
<p>Sets the output state and direction of the GPIOs. Puts the SPI interface into manual slave select mode and sets the SPI frequency. Sets the I2C mode for the busses used by the PDM. Some parameters are determined from data unmarshalled during the GUI configuration.</p>
*/
  public void setup()
  {

    allocateTxRxBuffers(8);

// GPIO initial setup

    infoTx("GPIO_W_DATAOUT",gpioOut);
    infoTx("GPIO_W_DIRECTION",gpioDirection);
    
// SPI initial setup. Frequency in kHz in XML file.
    infoTx("SPI_W_CTRL",0x00000000); // SS manual mode
    int spiFrequencyDivisor = 20000/spiFrequency - 1;
    infoTx("SPI_W_FREQ",spiFrequencyDivisor&0xffff);
    infoTx("SPI_W_SS",0x00000000); // Start with nothing selected

// I2C initial setup
    infoTx("I2C_W_CTRL/0",0x00000000);
    infoTx("I2C_W_CTRL/1",0x00000000);
    infoTx("I2C_W_CTRL/2",0x00000000);

// Write the bytes to the transaction buffer
    putBytes( 0, true );

  }

/**
<p>Activates required PDMDB interfaces.</p>
<p>The following interfaces that are used by the PDM are activated:
GPIO, SPI, I2C0, I2C1, I2C2, ADC, DAC and JTAG.</p>
@return true if verification of the registers was successful, false otherwise.
*/
  public boolean enableInterfaces()
  {
    allocateTxRxBuffers(6);
// Activate interfaces
    infoTx("NC_W_CRB",0x3e); // GPIO, SPI, I2C0, I2C1 & I2C2
    infoTx("NC_W_CRC",0x00); // Other I2C disabled
    infoTx("NC_W_CRD",0x38); // ADC, DAC & JTAG

    infoTx("NC_R_CRB",0x00);
    infoTx("NC_R_CRC",0x00);
    infoTx("NC_R_CRD",0x00);
    //    infoTx("NC_W_CRA",0x55); // Was SLVS but has disappeared from manual

    putBytes( 0x0000, true );
    try { Thread.sleep(100); } catch(InterruptedException ex){}

    byte [] rxbuf = new byte [0x30];
    try { io.get( deppBaseRx, 0x0030, rxbuf ); } catch (cbsw.RegisterIoException de){System.err.println("ScaDepp::enableInterfaces(): "+de);}

    return (rxbuf[0x07]==0x3e)&&(rxbuf[0x04]==0x00)&&(rxbuf[0x17]==0x00)&&(rxbuf[0x14]==0x00)&&(rxbuf[0x27]==0x38)&&(rxbuf[0x24]==0x00);
  }

/**
<p>Reads SCA ID.</p>
@return ID as hex encoded string. 
*/
  public String readId()
  {
    allocateTxRxBuffers(1);
    infoTx("NC_R_ID",0x00);

    putBytes( 0x0000, true );
    try { Thread.sleep(100); } catch(InterruptedException ex){}

    cache();
    byte [] rxbuf = getResponseBytes("NC_R_ID");

    return String.format("0x%02x%02x%02x",rxbuf[2],rxbuf[1],rxbuf[0]);
  }

/**
<p>Sets the DAC value.</p>
@param dacId The DAC channel number.
@param dacValue The DAC value [0,255].
*/
  public void setDac(int dacId,byte dacValue)
  {
    allocateTxRxBuffers(1);
    infoTx("DAC_W_"+dacId,dacValue);
    putBytes(0,true);
    return;
  }

/**
<p>Sets the GPIO value.</p>
<p>Initial settings are based on data unmarshalled from XML during GUI configuration. Care should be exercised to avoid potential signal contention when changing this setting. For example, the GPIOs used as GBTX resets should always be deasserted (gpioValue=0).</p>
@param gpioId The GPIO number.
@param gpioValue 0=deassert, 1=assert.
*/
  public void setGpio(int gpioId, int gpioValue)
  {
    if (gpioValue==1) gpioOut.set(gpioId);
    else gpioOut.clear(gpioId);
      
    allocateTxRxBuffers(1);
    infoTx("GPIO_W_DATAOUT",gpioOut);
    putBytes(0,true);
    return;
  }
    
/**
<p>Sets the GPIO direction.</p>
<p>Initial settings are based on data unmarshalled from XML during GUI configuration. Care should be exercised to avoid potential signal contention when changing this setting.</p>
@param gpioId The GPIO number.
@param flag 0=input, 1=output.
*/
  public void setGpioDirection(int gpioId, int flag)
  {
    if (flag==1) gpioDirection.set(gpioId);
    else gpioDirection.clear(gpioId);
      
    allocateTxRxBuffers(1);
    infoTx("GPIO_W_DIRECTION",gpioDirection);
    putBytes(0,true);
    return;
  }
    
/**
<p>Sets or clears the SCA ADC current source enable.</p>
<p>Initial settings are based on data unmarshalled from XML during GUI configuration. Care should be exercised to avoid potential signal contention when changing this setting.</p>
@param adcId The SCA ADC channel.
@param flag 0=disable, 1=enable.
*/
  public void setAdcCurrentSource(int adcId, int flag)
  {
    if (flag==1) adcCurrentSource.set(adcId);
    else adcCurrentSource.clear(adcId);
      
    allocateTxRxBuffers(1);
    infoTx("ADC_W_CURR",adcCurrentSource);
    putBytes(0,true);
    return;
  }

/*
<p>Reads I2C register.</p>
<p>Only known use in DtmControl.</p>
@param i2cBus The SCA I2C bus number.
@param i2cSlaveAddress The I2C slave address.
@param off The register offset.
@return The register value.

  public byte readI2c(int i2cBus, int i2cSlaveAddress, int off)
  {
    allocateTxRxBuffers(6);

    infoTx("I2C_W_CTRL/"+i2cBus,(2<<2)|0x2);

// First two bytes are register start offset
    infoTx("I2C_W_DAT0/"+i2cBus,off);
    infoTx("I2C_M_7B_W/"+i2cBus,i2cSlaveAddress);

// Now read back register byte
    infoTx("I2C_W_CTRL/"+i2cBus,(1<<2)|0x2);
    infoTx("I2C_M_7B_R/"+i2cBus,i2cSlaveAddress);
    infoTx("I2C_R_DAT3/"+i2cBus,0x00000000);

    putBytes(0,true);

// Get the byte from the buffered responses

    cache();
    byte [] buf = getResponseBytes("I2C_R_DAT3/"+i2cBus);
    
    //    System.out.println(String.format("buf=%1$02x %2$02x %3$02x %4$02x",buf[0],buf[1],buf[2],buf[3]));
    
    return buf[0];
  }
*/
/*
<p>Writes I2C register.</p>
<p>Only known use in DtmControl.</p>
@param i2cBus The SCA I2C bus number.
@param i2cSlaveAddress The I2C slave address.
@param off The register offset.
@param reg The register value.

  public void writeI2c(int i2cBus, int i2cSlaveAddress, int off, int reg)
  {
    allocateTxRxBuffers(3);

    infoTx("I2C_W_CTRL/"+i2cBus,(3<<2)|0x2);

// First two bytes are register start offset followed by data byte
    infoTx("I2C_W_DAT0/"+i2cBus,((reg&0xff)<<16)|off);
    infoTx("I2C_M_7B_W/"+i2cBus,i2cSlaveAddress);

    putBytes(0,true);
    
    return;
  }
*/  
/**
<p>Controls the GPIOs.</p>
@param gpios gpio:direction:value|gpio:direction:value...
*/
  public void gpioControl(String [] gpios)
  {
    for (String gpio : gpios)
    {
      String [] tokens = gpio.split(":"); 
      int gpioId = Integer.decode(tokens[0]);
      if (tokens[1].equalsIgnoreCase("out"))
        gpioDirection.set(gpioId);
      else
        gpioDirection.clear(gpioId);
      
      if (tokens[2].equals("1"))
        gpioOut.set(gpioId);
      else
        gpioOut.clear(gpioId);
    }
    
    allocateTxRxBuffers(2);
    infoTx("GPIO_W_DIRECTION",staticThis.gpioDirection);
    infoTx("GPIO_W_DATAOUT",staticThis.gpioOut);
    putBytes(0,true);
  }

/**
<p>Reads the current state of the GPIOs.</p>
@return The asserted GPIOs.
*/
  public BitSet getGpioState()
  {
    allocateTxRxBuffers(1);
    infoTx("GPIO_R_DATAIN",0x00);
    putBytes(0,true);
    
// Get the GPIO state from the buffered responses

    cache();
    byte [] buf = getResponseBytes("GPIO_R_DATAIN");
    
    return BitSet.valueOf(buf);
  }
  
/**
<p>Reads ADC value.</p>
@param adc The ADC channel number.
@return The ADC value.
*/
  public int getAdc(int adc)
  {
    int framesPerADC = 3; // i.e. number of SCA commands per ADC
    allocateTxRxBuffers(framesPerADC);
    //    infoTx("ADC_W_CURR",0x00000000);
    infoTx("ADC_W_CURR",adcCurrentSource);
    infoTx("ADC_W_MUX",adc);
    byte [] go = {(byte)0x01};
    infoTx("ADC_GO",go);

    putBytes(0,true);

    //    try {Thread.sleep(10);}catch(InterruptedException ex){};

// Now decode the SCA responses to extract the ADC values

    cache();
    byte [] buf = getResponseBytes("ADC_GO");
    
    ByteBuffer rxBB = ByteBuffer.wrap(buf); rxBB.order(ByteOrder.LITTLE_ENDIAN);
    return rxBB.getInt(0);
  }

/**
<p>Writes a byte array to the hardware through the SCA interface.</p>
<p>Each SCA-I2C transaction sends 2 pointer bytes + 14 data
bytes.</p>
<p>For synchronisation, it may be better for transaction to be split
into smaller blocks, each taking less than 10ms (DEPP timeout). But
for now everything is in one transaction and rely on sleeps in the
user code where needed.</p>
@param i2cBus The SCS I2C bus number.
@param i2cSlaveAddress The I2C slave address.
@param i2cSlaveSpeed The I2C transaction speed.
@param i2cBuf Array of register values.
@param offset Offset to location where to start writing.
*/
  public void putI2c(int i2cBus, int i2cSlaveAddress, int i2cSlaveSpeed, byte [] i2cBuf, int offset)
  {
         
    int buflen = i2cBuf.length;

    byte [] buf = new byte [4];

    int n14 = (buflen+13)/14; // Total number of 14 byte SCA I2C transactions
    if (n14<1) return; // Nothing to do

    allocateTxRxBuffers(6*n14);
    
    int i = 0;
    for (int i14=0;i14<n14;i14++)
    {
      int remaining = buflen - 14*i14;
      int nbytes = (remaining > 13) ? 14 : remaining;
      nbytes+=2; // 2 bytes for the register offset
      
      buf[0] = (byte)(nbytes<<2); buf[0] = (byte)(buf[0] | i2cSlaveSpeed);
      infoTx("I2C_W_CTRL/"+i2cBus,buf);

      int startAddr = offset + 14*i14;
      buf[0] = (byte)startAddr; buf[1] = (byte)(startAddr>>8); // reg offset
      for (int j=2;j<16;j++)
      {
	if (i<buflen) buf[j%4] = i2cBuf[i++];
        if (j%4 == 3)
        {
	  infoTx("I2C_W_DAT"+j/4+"/"+i2cBus,buf);
	}
      }

      infoTx("I2C_M_7B_W/"+i2cBus,i2cSlaveAddress);

    }

    putBytes(0,true);
    return;
  }

/**
<p>Reads a byte array from the hardware through the SCA interface.</p>
@param i2cBus The I2C bus number.
@param i2cSlaveAddress The I2C slave address.
@param i2cSlaveSpeed The I2C transaction speed.
@param i2cBuf Array of register values.
@param offset Offset to location where to start reading.
*/
  public void getI2c(int i2cBus, int i2cSlaveAddress, int i2cSlaveSpeed, byte [] i2cBuf, int offset)
  {
    int hdlcFrames = 4+5*((i2cBuf.length+15)/16);
    allocateTxRxBuffers(hdlcFrames);

// First we send the two byte start offset

    infoTx("I2C_W_CTRL/"+i2cBus,(2<<2)|i2cSlaveSpeed);

// First two bytes are register start offset
    infoTx("I2C_W_DAT0/"+i2cBus,offset&0xffff);

    infoTx("I2C_M_7B_W/"+i2cBus,i2cSlaveAddress);

// Now read back registers 16 at a time (GBTX auto-increments pointer)
    infoTx("I2C_W_CTRL/"+i2cBus,(16<<2)|i2cSlaveSpeed);

    int i = 0;
    while (i<i2cBuf.length)
    {
      infoTx("I2C_M_7B_R/"+i2cBus,i2cSlaveAddress);

      for (int j=0;j<4;j++)
      {
	infoTx("I2C_R_DAT"+(3-j)+"/"+i2cBus,0x00000000); // Reverse order
        i+=4;
      }
    }

    putBytes(0,true);

// Decode the SCA responses to extract the data

    cache();

    int irx = 0;
    byte [] buf = null;
    do
    {
      for (int ifr=0;ifr<4;ifr++)
      {
	buf = getResponseBytes("I2C_R_DAT"+(3-ifr));
	if (buf == null || irx>i2cBuf.length-1) break;
	for (int j=0;j<4&&irx<i2cBuf.length;j++) i2cBuf[irx++] = buf[j];
      }
    } while (buf!=null);
    return;
  }
}

