package cbsw.lhcb.pdmdb;

import javax.xml.bind.annotation.*;
import java.util.*;
import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import java.nio.file.*;

/**
<p>This class encapsulates a sequence of Commands.</p>
<p>It also implements many of the static methods called by {@link Command}</p>
<p>The command sequence is unmarshalled from XML and executed by reflection.</p>
<p>A very basic sequence in XML that configures a PDMDB (H-type) could be:</p>
<pre>
{@code
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<testerSequence xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <command label="muDAQ configure" targetClass="TesterSequence" targetMethod="configureMuDaq"/>
  <command label="Configure master link" targetClass="TesterSequence" targetMethod="configureMasterLink" data="i2c" stopOnFail="true"/>
  <command label="SCA setup" targetClass="TesterSequence" targetMethod="setupSca"/>
  <command label="DTM0 program GBTX" targetClass="TesterSequence" targetMethod="configureDtm" data="0"/>
  <command label="DTM0 program FPGA" targetClass="TesterSequence" targetMethod="programFpga" data="0"/>
  <command label="DTM2 program GBTX" targetClass="TesterSequence" targetMethod="configureDtm" data="1"/>
  <command label="DTM2 program FPGA" targetClass="TesterSequence" targetMethod="programFpga" data="1"/>
  <command label="Train e-ports" targetClass="TesterSequence" targetMethod="trainEports" data="0?1"/>
  <command label="Configure EC0" targetClass="TesterSequence" targetMethod="configureClaros" data="0"/>
  <command label="Configure EC1" targetClass="TesterSequence" targetMethod="configureClaros" data="1"/>
  <command label="Configure EC2" targetClass="TesterSequence" targetMethod="configureClaros" data="2"/>
  <command label="Configure EC3" targetClass="TesterSequence" targetMethod="configureClaros" data="3"/>
</testerSequence>
}
</pre>
**/
@XmlAccessorType(XmlAccessType.NONE)
@XmlRootElement(name="testerSequence")
public class TesterSequence
{
private java.util.prefs.Preferences prefs;
    
@XmlElement(name="command")
private ArrayList<Command> commandList;
    
private int iterations;
private ListIterator<Command> it;
private boolean suspend=false;

private javax.swing.JTextField cmdTF;
static javax.swing.JTextArea textArea;
private javax.swing.JScrollPane textSP;
private javax.swing.JButton recordingButton;
    
private boolean recording = false;

static javax.swing.JTextField rootDirectoryTF;
static String testSubdirectory;
static Path testDirectory;
static BufferedWriter commandResponseFile;
static BufferedWriter testSummaryXml;
static Path eyeScanDataDirectory;

static TesterSequence staticThis;

/**
<p>No-argument constructor called during unmarshalling</p>
 */
public TesterSequence()
{
  prefs = java.util.prefs.Preferences.userNodeForPackage(TesterSequence.class).node("JPDMDBDepp");
  staticThis = this;
}

/**
<p>Convenience method to make a GUI panel</p>
@return The panel
 */
public javax.swing.JPanel makePanel()
{
    javax.swing.JPanel jP = new javax.swing.JPanel();
    jP.setLayout(new java.awt.GridBagLayout());
    java.awt.GridBagConstraints gbC = new java.awt.GridBagConstraints();
    gbC.fill = java.awt.GridBagConstraints.BOTH;
    gbC.anchor = java.awt.GridBagConstraints.CENTER;
    gbC.insets = new java.awt.Insets(1,1,1,1);
    gbC.weightx = 1.0f;
    gbC.weighty = 1.0f;
    
    javax.swing.JPanel dataJP = new javax.swing.JPanel();
    dataJP.setBorder(javax.swing.BorderFactory.createTitledBorder("Test data root"));
    rootDirectoryTF = new javax.swing.JTextField(prefs.get("chooser.tester.dir","ModuleTestData"));
    rootDirectoryTF.setEditable(false);
    dataJP.add(rootDirectoryTF);
    javax.swing.JButton dataJB = new javax.swing.JButton("Change");
    dataJB.addActionListener(ae->chooseDataDirectory());
    dataJP.add(dataJB);
    
    gbC.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    jP.add(dataJP,gbC);

    javax.swing.JPanel buttonJP = new javax.swing.JPanel();
    buttonJP.setLayout(new java.awt.GridLayout(0,3));
  
    for (Command cmd : commandList )
    {
      javax.swing.JButton button = new javax.swing.JButton(cmd.label);
      button.addActionListener(l->{try {cmd.go();}catch (Exception ex) {PdmdbModuleTester.readFlush();textArea.append("TesterSequence.makePanel(): "+ex+"\n");}});
      //      gbC.gridwidth = 1;
      //      jP.add(button,gbC);
      buttonJP.add(button);

      javax.swing.JTextField statusTF = new javax.swing.JTextField(60);
      statusTF.setEditable(false);
      statusTF.setFont(new java.awt.Font("Monospaced",java.awt.Font.PLAIN,12));
      //      gbC.gridwidth = java.awt.GridBagConstraints.REMAINDER;
      //      jP.add(statusTF,gbC);

      cmd.setStatusTextField(statusTF);
      cmd.setButton(button);
    }

    gbC.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    jP.add(buttonJP,gbC);
    
    cmdTF = new javax.swing.JTextField(80);
    cmdTF.addActionListener(l->{PdmdbModuleTester.exec(cmdTF.getText());});
    cmdTF.setFont(new java.awt.Font("Monospaced",java.awt.Font.PLAIN,12));
    gbC.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    jP.add(cbsw.gui.GuiJPanel.addBorder(cmdTF,"Command"),gbC);
  
    textArea = new javax.swing.JTextArea(10,10); // Minimum width. Will expand to fill available width.
    textArea.setEditable(false);
    textArea.setFont(new java.awt.Font("Monospaced",java.awt.Font.PLAIN,12));
    textSP = new javax.swing.JScrollPane(textArea);
    gbC.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gbC.fill = java.awt.GridBagConstraints.HORIZONTAL;
    jP.add(textSP,gbC);
  
    javax.swing.JPanel playJP = new javax.swing.JPanel();
    
    javax.swing.JButton button = new javax.swing.JButton("Play");
    button.addActionListener(l->{prepare();resume();});
    playJP.add(button);

    button = new javax.swing.JButton("Pause");
    button.addActionListener(l->pause());
    playJP.add(button);

    button = new javax.swing.JButton("Resume");
    button.addActionListener(l->resume());
    playJP.add(button);

    recordingButton = new javax.swing.JButton("Record");
    recordingButton.addActionListener(l->recording());
    playJP.add(recordingButton);

    button = new javax.swing.JButton("First step");
    button.addActionListener(l->{prepare();step();});
    playJP.add(button);
    
    button = new javax.swing.JButton("Step");
    button.addActionListener(l->step());
    playJP.add(button);

    button = new javax.swing.JButton("Flush RX pipe");
    button.addActionListener(l->PdmdbModuleTester.readFlush());
    playJP.add(button);

    gbC.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    jP.add(playJP,gbC);

    return jP;
}

/**
<p>Creates dialog to choose the root directory for test data</p>
<p>Timestamped subdirectories are created in this directory for each cycle of a sequence</p>
 */
public void chooseDataDirectory()
{
  javax.swing.JFileChooser chooser = new javax.swing.JFileChooser(prefs.get("chooser.tester.dir","ModuleTestData"));
  chooser.setFileSelectionMode(javax.swing.JFileChooser.DIRECTORIES_ONLY);
  int returnVal = chooser.showOpenDialog(null);
  if(returnVal != javax.swing.JFileChooser.APPROVE_OPTION) return;
  if (!chooser.getSelectedFile().isDirectory()) return;
  prefs.put("chooser.tester.dir",chooser.getSelectedFile().getAbsolutePath());
  rootDirectoryTF.setText(prefs.get("chooser.tester.dir","ModuleTestData"));
}

/**
<p>Writes a message to the command/response file.</p>
<p>Does nothing if testDirectory or commandResponseFile are null.</p>
@param msg The CharSequence to write.
 */
public static void logMessage(CharSequence msg)
{

  textArea.append(msg.toString());
  if (testDirectory == null) return;
  if (commandResponseFile == null) return;
  
  try
  {
    commandResponseFile.write(msg.toString());
    commandResponseFile.flush();
  }
  catch (IOException ioe){ throw new RuntimeException("TesterSequence.logMessage()|"+ioe); }

}

/**
<p>Creates directories and files for test data.</p>
<p>Callable by {@link Command}.</p>
<p>Call this method at the start of a test sequence.</p>
@param args Not used.
@return Command response.
*/
public static String prepareTests(String [] args)
{
  String dir = java.util.prefs.Preferences.userNodeForPackage(TesterSequence.class).node("JPDMDBDepp").get("chooser.tester.dir","ModuleTestData");
    
// Directories are created writable only by the owner but readable and traversabl by all
//  FileAttribute<Set<PosixFilePermission>> attrs = PosixFilePermissions.asFileAttribute(PosixFilePermissions.fromString("rwxr-xr-x"));
  try
  {
   testSubdirectory = "TestData-"+java.time.LocalDateTime.now().format(java.time.format.DateTimeFormatter.ofPattern("yyyyMMdd-A"));
// Create (timestamped) directory for this test cycle
   testDirectory = Files.createDirectories(Path.of(dir+File.separator+testSubdirectory));
   System.err.println("Test output directory is "+testDirectory.toString());
// Create/open the command/response file
    commandResponseFile = Files.newBufferedWriter(testDirectory.resolve(Path.of("CommandResponse.txt")),StandardOpenOption.CREATE_NEW,StandardOpenOption.WRITE);
// Create/open the test summary (XML) file
    testSummaryXml = Files.newBufferedWriter(testDirectory.resolve(Path.of("TestSummary.xml")),StandardOpenOption.CREATE_NEW,StandardOpenOption.WRITE);
    testSummaryXml.write("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n" );
    testSummaryXml.write("<test xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">\n" );
// Create directory for eye scan data files
    eyeScanDataDirectory = Files.createDirectories(testDirectory.resolve(Path.of("EyeScanData")));
    return "OK|TesterSequence.prepareTests";
  }
  catch (IOException ioe){ throw new RuntimeException("TesterSequence.prepareTests()|"+ioe); }
}

/**
<p>Closes test data files.</p>
<p>Callable by {@link Command}.</p>
<p>Call this method at the end of a test sequence to close the output files.</p>
@param args Not used.
@return Command response.
*/
public static String endOfTests(String [] args)
{
  try
  {
    testDirectory = null;
    eyeScanDataDirectory = null;
    commandResponseFile.close(); commandResponseFile = null;
    testSummaryXml.write("</test>\n" );
    testSummaryXml.close(); testSummaryXml = null;
    return "OK|TesterSequence.endOfTests";
  }
  catch (IOException ioe){ throw new RuntimeException("TesterSequence.endOfTests()|"+ioe); }
}

/**
<p>Writes lines to XML summary file.</p>
<p>If testDirectory or testSummaryXml are null, this method does nothing.</p>
@param text CharSequence to write.
 */
public static void xml(CharSequence text)
{
  if (testDirectory == null) return;
  if (testSummaryXml == null) return;
  try { testSummaryXml.write(text.toString()); testSummaryXml.flush();}
  catch (IOException ioe){ throw new RuntimeException("TesterSequence.xml(): "+ioe); }
}

/**
<p>Configures the DTM module.</p>
<p>Callable by {@link Command}.</p>
<ol>
<li>Holds both DTM GBTXs in reset.</li>
<li>Deasserts reset for GBTX0 and configures it.</li>
<li>Deasserts reset for GBTX1 and configures it.</li>
</ol>
@param args {dtmIndex}.
<p>dtmIndex is the index of the DTM in the list unmarshalled from XML, not the DTM label.</p>
@return Command response.
*/
  public static String configureDtm(String [] args)
  {
    try
    {
      StringBuffer xml = new StringBuffer();
      int dtmIndex = Integer.decode(args[0]).intValue();
      DtmControl dtm = PdmdbMaster.pdmdbMaster.dtmList.get(dtmIndex);
      dtm.configure();

      Thread.sleep(1000);
    
// After resetting the GBTXs it is recommended to reset the FPGA MGT receivers
      for (GbtxControl gbtx : dtm.gbtxList)
      {
	gbtx.trainEports();
	System.err.println("Resetting MGT "+gbtx.mgt);
        MuDaq.cfg.put("gt."+gbtx.mgt+".rxreset", true, MuDaq.muDaq.io);
        MuDaq.cfg.put("gt."+gbtx.mgt+".rxreset", false, MuDaq.muDaq.io);
      }

// Read the muDaq status to get the GBT header bits
      
      MuDaq.muDaq.status.cache(MuDaq.muDaq.io);

      String [] dtmString = new String [2];
      xml.append(String.format("<dtm testerModule=\"dtm%d\"",dtm.dtm));

      for (GbtxControl gbtx : dtm.gbtxList)
      {
	  //        int mgtId = 2 + 2*dtmIndex + gbtx.gbtx;
        String dtmGbtHeader = MuDaq.muDaq.status.get("gbt."+gbtx.mgt+".mgt.rx.header");
        String dtmVerify = gbtx.verify();
        dtmString[gbtx.gbtx] = "DTM"+dtm.dtm+".GBTX"+gbtx.gbtx+"|"+dtmVerify+"|gbtHeader="+dtmGbtHeader;
        String [] dtmTokens = dtmVerify.split("\\|");
        if (dtmTokens.length > 1 && dtmTokens[0].equals("OK")) PdmdbModuleTester.dtmGbtxId[dtm.dtm][0] = dtmTokens[1];
        xml.append(String.format(" gbtx%dId=\"%s\" gbtx%dState=\"%s\"",gbtx.gbtx,dtmTokens[1],gbtx.gbtx,dtmTokens[2]));
      }
      xml.append(" />\n");
      TesterSequence.xml(xml.toString());
    
      return "OK|TesterSequence.configureDtm|"+dtmString[0]+"|"+dtmString[1];
    }
    catch (Exception ex){ return "FAIL|TesterSequence.configureDtm()|"+ex; }

  }
    
/**
<p>Performs the automatic GBTX ePort phase training.</p>
<p>Callable by {@link Command}.</p>
@param args {dtmIndex}.
<p>dtmIndex is the index of the DTM in the list unmarshalled from XML, not the DTM label.</p>
@return Command response.
**/
  public static String trainEports(String [] args)
  {
    try
    {
      int dtmIndex = Integer.decode(args[0]).intValue();
      DtmControl dtm = PdmdbMaster.pdmdbMaster.dtmList.get(dtmIndex);
      
// Set a fixed pattern on the down links (Tester board use case)
      MuDaq.cfg.put("tfctog-enable","0",MuDaq.muDaq.io);
      MuDaq.cfg.put("tfc-d","A",MuDaq.muDaq.io);

// Set a fixed pattern on the e-links (PDMDB use case)
      byte [] regBuf = {(byte)0x01,(byte)0xaa,(byte)0xaa};
      dtm.put(regBuf,0);

      for (GbtxParent gbtx : dtm.gbtxList )
      {
	byte [] trainBuf = gbtx.trainEports();
	
	StringBuffer sb = new StringBuffer();
	sb.append("<eportPhase testerModule=\"dtm"+dtm.dtm+"\" gbtx=\""+gbtx.gbtx+"\" ");
	String timestamp = java.time.LocalDateTime.now().format(java.time.format.DateTimeFormatter.ofPattern("yyyyMMdd-A"));
	sb.append("timestamp=\""+timestamp+"\" ");
	for (int group=0;group<7;group++) sb.append(String.format("egroup"+group+"Eport0=\"%d\" egroup"+group+"Eport4=\"%d\" ",trainBuf[4*group]&0xf,trainBuf[4*group+2]&0xf));
	sb.append(String.format("/>\n"));

	TesterSequence.xml(sb);
      }
      
// Disable fixed pattern on the e-links (PDMDB use case)
      regBuf[0] = (byte)0x00;
      dtm.put(regBuf,0);
    }
    catch (Exception ioe){ return "FAIL|TesterSequence.trainEports()|"+ioe; }
    return "OK|TesterSequence.trainEports";
  }

/**
<p>Performs PD signal loopback tests (Oxford tester).</p>
<p>Callable by {@link Command}.</p>
<p>The Oxford tester board loops back the PD signals in pairs. The firmware has IOBUFs on the IOs that allow either end of the loop to be the driver and the other end the receiver. There are 84 pairs per DTM FPGA.</p>
<p>For this test, to avoid signal contention, each of the 3 DTM FPGAs must have different firmware as the IO pairing is different for each.</p>
<p>This test is not very meaningful for PDMDB-H since many of the signals are not routed on the PCB so that most of the loopbacks are broken.</p>
@param args {dtmIndex,data}.
<p>dtmIndex is the index of the DTM in the list unmarshalled from XML, not the DTM label. data is a comma-separated list of up to 14 byte values that set the level asserted by each IO.</p>
@return Command response including a hexadecimal string of the signals read back from each IO for each direction of the test.
**/
  public static String pdSignalTest(String [] args)
  {
    StringBuffer sb = new StringBuffer();
    try
    {
      int dtmIndex = Integer.decode(args[0]).intValue();
      DtmControl dtm = PdmdbMaster.pdmdbMaster.dtmList.get(dtmIndex);

      String [] dataBytes = args[1].split(",");
      
// Set the pattern
      byte [] dataBuf = new byte [dataBytes.length];
      ByteBuffer dataBB = ByteBuffer.wrap(dataBuf);
      for ( String dataByte : dataBytes ) dataBB.put((byte)Integer.decode(dataByte).byteValue());
      dtm.put(dataBuf,0x10);

// Set the direction
      byte [] dirBuf = new byte [dataBuf.length];
      dtm.put(dirBuf,0x20);
      try {Thread.sleep(100);}catch(InterruptedException ie){} // Allow tri-stated IOs to settle

// Read the looped back signals
      byte [] rdBuf = new byte [dataBuf.length];
      dtm.get(rdBuf,0x11a);

      for (byte b : rdBuf) sb.insert(0,String.format("%02x",b));

// Change the direction
      for (int i=0;i<dirBuf.length;i++) dirBuf[i] = (byte)0xff;
      dtm.put(dirBuf,0x20);
      try {Thread.sleep(100);}catch(InterruptedException ie){} // Allow tri-stated IOs to settle
      
// Read the looped back signals
      dtm.get(rdBuf,0x10c);

      sb.insert(0,"|");
      for (byte b : rdBuf) sb.insert(0,String.format("%02x",b));

    }
    catch (Exception ioe){ return "FAIL|TesterSequence.pdSignalTest()|"+ioe; }
    return "OK|TesterSequence.pdSignalTest()|"+sb;
  }

/**
<p>Programs the PDMDB FPGA.</p>
<p>Callable by {@link Command}.</p>
@param args {dtmIndex,binFile}.
<p>dtmIndex is the index of the DTM in the list unmarshalled from XML, not the DTM label. binFile is the FPGA configuration .bin file location.</p>
@return Command response.
*/
  public static String programFpga(String [] args)
  {
    int dtmIndex = Integer.decode(args[0]).intValue();
    DtmControl dtm = PdmdbMaster.pdmdbMaster.dtmList.get(dtmIndex);
    XilS7Control xil = dtm.xilS7;

    if (args.length > 1) xil.prefsSetFilename(args[1]);
    xil.program();

    HashMap<String,String> statusMap = xil.getStatus();
    StringBuffer sb = new StringBuffer();
    sb.append("<fpga testerModule=\"dtm"+dtm.dtm+"\" ");
    statusMap.forEach((k,v)->sb.append(k+"=\""+v+"\" "));
    sb.append("/>");
    TesterSequence.xml(sb);

    return "OK|TesterSequence.programFpga";
  }
    
/**
<p>Configures the muDaq registers</p>
<p>Callable by {@link Command}.</p>
@param args Not used.
@return Command response.
**/
public static String configureMuDaq(String [] args)
{
  for ( MGTControl mgt : MuDaq.muDaq.mgts ) mgt.setStateFromPreferences();
  MuDaq.muDaq.setStateConfigure();
  try { MuDaq.muDaq.io.put(MuDaq.cfg.getBaseAddress(),0,MuDaq.cfg.toByteArray(4)); }
  catch(Exception ex){System.err.println("Mudaq.configure(): "+ex); return "FAIL|MuDaq.configure";}
  MuDaq.muDaq.reset();

  try {Thread.sleep(100);}catch(InterruptedException e){}
  // Check the protocol register to decide whether configuration was successful
  //  MuDaq.muDaq.cfg.cache(MuDaq.muDaq.io);
  //  if (!cfg.get("net.ip.protocol",csr).equals("f2")) return "FAIL|MuDaq.configure";
  return "OK|TesterSequence.configureMuDaq";
}

/**
<p>Resets muDaq.</p>
<p>Callable by {@link Command}.</p>
@param args Not used.
@return Command response.
**/
  public static String resetMuDaq(String [] args)
  {
    try { MuDaq.muDaq.reset(); }
    catch ( Exception ex ){ return "FAIL|TesterSequence.resetMuDaq()|"+ex;}
    return "OK|TesterSequence.resetMuDaq()";
  }

/**
<p>Switches muDaq K7 transceiver on or off.</p>
<p>Callable by {@link Command}.</p>
@param args {mgtId,txOn,rxOn}.
<p>There are 8 MGTs with mgtId=0-7. txOn=0 to switch off the transmitter, txOn=1 to switch on. Similarly for the receiver.
@return Command response.
**/
public static String setMgtOnOff(String [] args)
{
  try
  {
    int mgtId = Integer.decode(args[0]).intValue();
    boolean txOn = Integer.decode(args[1]).intValue()==1;
    boolean rxOn = Integer.decode(args[2]).intValue()==1;
    MuDaq.cfg.put("gt."+mgtId+".rxon", rxOn, MuDaq.muDaq.io);
    MuDaq.cfg.put("gt."+mgtId+".txon", txOn, MuDaq.muDaq.io);
    return "OK|TesterSequence.setMgtOnOff()";
  }
  catch (Exception ex){throw new RuntimeException("TesterSequence.setMgtOnOff(): "+ex);}
}
    
/**
<p>Resets muDaq MGT transceiver.</p>
<p>Callable by {@link Command}.</p>
@param args {mgtId,txReset,rxReset}.
<p>There are 8 transceivers with mgtId=0-7. txReset=0 to deassert the receiver reset, txReset=1 to assert reset. Similarly for the receiver.</p> 
@return Command response.
**/
public static String setMgtReset(String [] args)
{
  try
  {
    int mgtId = Integer.decode(args[0]).intValue();
    boolean txReset = Integer.decode(args[1]).intValue()==1;
    boolean rxReset = Integer.decode(args[2]).intValue()==1;
    MuDaq.cfg.put("gt."+mgtId+".rxreset", rxReset, MuDaq.muDaq.io);
    MuDaq.cfg.put("gt."+mgtId+".txreset", txReset, MuDaq.muDaq.io);
    return "OK|TesterSequence.setMgtReset()";
  }
  catch (Exception ex){throw new RuntimeException("TesterSequence.setMgtReset(): "+ex);}
}

/**
<p>Configures the master GBTX</p>
<p>Callable by {@link Command}.</p>
<p>To allow configuration of un-fused master GBTX this function
selects either the I2C interface (&lt;gbtxI2c&gt; in XML) or the
generic interface (&lt;gbtx&gt; in XML) which can be set to either
I2C or IC</p>
@param args {interface}.
<p>interface=i2c to use the I2C interface, anything else to use the generic interface.</p>
@return Command response.
**/
public static String configureMasterLink(String [] args)
{
  try
  {
    GbtxParent gbtx = args[0].equals("i2c") ? MuDaq.muDaq.gbtxI2c : MuDaq.muDaq.gbtx;
    
    gbtx.readConfigurationFile();
    gbtx.configure();
    if (!args[0].equals("i2c")) Thread.sleep(500);
    gbtx.verify();

    PdmdbModuleTester.tcmGbtxId = gbtx.get("fuse.serial");
    String state = gbtx.get("powerup-fsm-state");
    String a5 = gbtx.get("a5");

    TesterSequence.xml(String.format("<tcm testerModule=\"tcm\" gbtxId=\"%s\" gbtxState=\"%s\" />\n",PdmdbModuleTester.tcmGbtxId,state));
  
    if (!a5.equals("a5")) return "FAIL|TesterSequence.configureMasterLink()|a5!="+a5;
    if (!state.equals("18")) return "FAIL|TesterSequence.configureMasterLink()|NOT_IDLE|FSM="+state; // 18(hex) is Idle state

    Thread.sleep(500);
    
    MuDaq.status.cache(MuDaq.muDaq.io);
    if (MuDaq.status.get("gbt.0.mgt.rx.header").equals("05"))
    {
      return "OK|TesterSequence.configureMasterLink()|TCM.GBTX|0x"+PdmdbModuleTester.tcmGbtxId;
    }
    else
    {
      return "FAIL|TesterSequence.configureMasterLink()|HEADER_INVALID";
    }
  }
  catch (Exception ex){System.err.println("TesterSequence.configureMasterLink(): "+ex); return "FAIL|TesterSequence.configureMasterLink()|Exception";}
  
}
    
/**
<p>Fuses GBTX registers</p>
<p>Callable by {@link Command}.</p>
<p>This function is called as part of the module testing sequence.</p>
<p>Caution - it irreversibly burns the TCM GBTX e-fuses.</p>
<p>The GBTX must have CONFIGSEL=1 (I2C) for this procedure and GBTFUSE3V3 must be enabled.</p>
@param args {fuseMode}
<p>fuseMode="enable" to allow fusing, "ignore" to disallow fusing, "force" to force re-fusing an already fused GBTX.</p>
@return Command response.
**/
public static String fuseMasterGbtx(String [] args)
{
  try
  {
// Check that I2C is OK by reading the "a5" register
// Return "FAIL" immediately if the value is not read correctly.
    
    MuDaq.muDaq.gbtxI2c.verify();
    BitSet rdBitSet = MuDaq.muDaq.gbtxI2c.rdBitSet; 
    String a5 = Gbtx.get("a5").get(rdBitSet);
    if (!a5.equals("a5")) return "FAIL|TesterSequence.fuseMasterGbtx()|a5!="+a5;

// Check register 366 to see the TCM appears to be already e-fused.
// If so, register 366 will contain 0x07, return "OK" immediately.
// If it is 0x00 continue with e-fusing.
// Anything else, bail out immediately.

    if (MuDaq.muDaq.gbtxI2c.gbtxReadBuf.length<367) return "FAIL|TesterSequence.fuseMasterGbtx()|Wrong register buffer length";
    byte fused = MuDaq.muDaq.gbtxI2c.gbtxReadBuf[366];
    if (fused==7 && !args[0].equals("force")) return "OK|TesterSequence.fuseMasterGbtx()|efuse.update.enable="+fused;
    if (fused!=0 && fused!=7) return "FAIL|TesterSequence.fuseMasterGbtx()|efuse.update.enable="+fused;

// Do the e-fusing. For extra protection, a preference must be set to enable it.
// Also the first argument to this method must be set to "enabled".

    java.lang.Runnable pulse = () -> MuDaq.muDaq.toggle("gbtx.efuse-program");
    java.util.prefs.Preferences prefs = java.util.prefs.Preferences.userNodeForPackage(TesterSequence.class).node("JPDMDBDepp");
    java.time.LocalDateTime fusingExpiry = java.time.LocalDateTime.parse(prefs.get("fusing-expiry","2000-01-01T12:00:00"));
    java.time.LocalDateTime now = java.time.LocalDateTime.now();
    long hoursUntilExpiry = now.until(fusingExpiry,java.time.temporal.ChronoUnit.HOURS);
    if (args[0].equals("enable") || args[0].equals("force"))
    {
      if (hoursUntilExpiry>0)
      {
        MuDaq.muDaq.cfg.put("gbtx.efuse-power-enable","1",MuDaq.muDaq.io);
        MuDaq.muDaq.gbtxI2c.fuse(pulse);
        MuDaq.muDaq.cfg.put("gbtx.efuse-power-enable","0",MuDaq.muDaq.io);
        return "OK|TesterSequence.fuseMasterGbtx()|Fusing done.";
      }
      else
      {
        return "FAIL|TesterSequence.fuseMasterGbtx()|Fusing skipped. Fusing token has expired.";
      }
    }
    else
    {
      return "OK|TesterSequence.fuseMasterGbtx()|Fusing skipped. Not enabled.";
    }
  }
  catch (Exception ex)
  {
    System.err.println("TesterSequence.fuseMasterGbtx(): "+ex);
    return "FAIL|TesterSequence.fuseMasterGbtx()|Exception during fusing";
  }

// To check that it was successful, the GBTX needs to be reset so that the fused values are loaded into the registers. This requires an interaction with the tester board so it is done as a later step of the sequence.

}

/**
<p>Configures up the SCA.</p>
<p>Callable by {@link Command}</p>
@param args Not used.
@return Command response.
*/
  public static String setupSca(String [] args)
  {
    try
    {
      String scaId = ScaDepp.staticThis.configure();
      TesterSequence.xml(String.format("<sca id=\"%s\"/>\n",scaId));
      return "OK|TesterSequence.setup()|scaId="+scaId;
    }
    catch (Exception ex){return "FAIL|TesterSequence.setupSca()|"+ex;}
  }
    
/**
<p>Sets DAC.</p>
<p>Callable by {@link Command}.</p>
@param args {channel,value}.
<p>There are 4 DACs with channel=0-3. The DACs are 8 bit with allowed range value=0-255.</p>
@return Command response.
*/
  public static String setDac(String [] args)
  {
    ScaDepp.staticThis.setDac(Integer.decode(args[0]).intValue(),(byte)Integer.decode(args[1]).intValue());
    return "OK|TesterSequence.setDac";
  }
    
/**
<p>Sets GPIO outputs.</p>
<p>Callable by {@link Command}.</p>
@param args {channel,value}.
<p>There are 32 GPIOs with channel=0-31. value=0 to deassert the GPIO output, value=1 to assert.</p>
@return Command response.
*/
  public static String setGpio(String [] args)
  {
    ScaDepp.staticThis.setGpio(Integer.decode(args[0]).intValue(),Integer.decode(args[1]).intValue());
    return "OK|TesterSequence.setGpio";
  }
    
/**
<p>Sets GPIO direction.</p>
<p>Callable by {@link Command}.</p>
@param args {channel,direction}.
<p>There are 32 GPIOs with channel=0-31. direction=0 to set the GPIO as input, direction=1 to set as output.</p>
@return Command response.
*/
  public static String setGpioDirection(String [] args)
  {
    ScaDepp.staticThis.setGpioDirection(Integer.decode(args[0]).intValue(),Integer.decode(args[1]).intValue());
    return "OK|TesterSequence.setGpioDirection";
  }
    
/**
<p>Controls the GPIOs.</p>
<p>Callable by {@link Command}.</p>
<p>Set and test multiple GPIOs in one call.</p>
@param args {testid:expected,gpio:direction:value,gpio:direction:value,...}.
<p>testid is a word to identify this test. expected is the expected value when read back (0 or 1).
gpio is the integer GPIO ID. direction can be in or out. value=0 to deassert the GPIO, value=1 to assert.</p>
@return Command response.
*/
  public static String gpioControl(String [] args)
  {
    String [] gpios = Arrays.copyOfRange(args,1,args.length);
    ScaDepp.staticThis.gpioControl(gpios);
    
// Some of the GPIOs have an RC network so give some recovery time
    try {Thread.sleep(30);} catch (InterruptedException ie){}

    BitSet gpioRead = ScaDepp.staticThis.getGpioState();
    
    StringBuffer rv = new StringBuffer("OK|TesterSequence.gpioControl");

    boolean pass = true;
    String [] testTokens = args[0].split(":");
    for (String gpio : gpios)
    {
      String [] tokens = gpio.split(":"); 
      int gpioId = Integer.decode(tokens[0]);
      boolean gpioState = gpioRead.get(gpioId);
      if (gpioState)
      {
	if (testTokens[1].equals("0")) pass = false;
	rv.append("|1");
      }
      else
      {
	if (testTokens[1].equals("1")) pass = false;
	rv.append("|0");
      }
    }
    TesterSequence.xml(String.format("<gpio test=\"%s\" pass=\"%b\" />\n",testTokens[0],pass));

    return rv.toString();
  }

/**
<p>Enables/disables ADC current source.</p>
<p>Callable by {@link Command}.</p>
@param args {channel,enable}.
<p>There are 31 ADCs for which the current source may be enabled with channel=0-30. enable=0 to disable the current source, enable=1 to enable.</p>
@return Command response.
*/
  public static String setAdcCurrentSource(String [] args)
  {
    ScaDepp.staticThis.setAdcCurrentSource(Integer.decode(args[0]).intValue(),Integer.decode(args[1]).intValue());
    return "OK|TesterSequence.setAdcCurrentSource";
  }
    
/**
<p>Acquires ADC test samples.</p>
<p>Callable by {@link Command}.</p>
@param args {samples,bins,binMinimum,binWidth,binMaximum,rmsMaximum,svg}.
@return Command response.
*/
  public static String getAdc(String [] args)
  {
    int samples;
    int bins;
    double binMinimum;
    double binWidth;
    double binMaximum;
    double maximumRMS;
    boolean svg;
    try
    {
      samples = Integer.decode(args[0]);
      bins = Integer.valueOf(args[1]);
      binMinimum = Double.valueOf(args[2]);
      binWidth = Double.valueOf(args[3]);
      binMaximum = binMinimum + bins*binWidth;
      maximumRMS = Double.valueOf(args[4]);
      svg = Boolean.valueOf(args[5]);
    }
    catch (NullPointerException npe){ return "FAIL|TesterSequence.getAdc()|Invalid argument"; }
  
    if (samples < 1) return "FAIL|TesterSequence.getAdc()|Invalid argument";

    String timestamp = java.time.LocalDateTime.now().format(java.time.format.DateTimeFormatter.ofPattern("yyyyMMdd-A"));
    
    try
    {

      StringBuffer sb = new StringBuffer();
// Bin description
      sb.append(String.format("%d,%f,%f\n",bins,binMinimum,binWidth)); // #bins, low limit, bin width
// Column headers
      for (int adc=0;adc<31;adc++) sb.append(String.format("adc%d,",adc));
      sb.append(String.format("adc31\n"));

// Data
      int adcVal;
      for (int i=0;i<samples;i++)
      {
        for (int adc=0;adc<31;adc++)
        {
	  adcVal = ScaDepp.staticThis.getAdc(adc);
	  sb.append(String.format("%d,",adcVal));
        }
	adcVal = ScaDepp.staticThis.getAdc(31);
	sb.append(String.format("%d\n",adcVal));
      }

      if (testDirectory != null)
      {
        BufferedWriter writer = Files.newBufferedWriter(testDirectory.resolve(Path.of("adc-"+timestamp+".dat")),StandardOpenOption.CREATE_NEW,StandardOpenOption.WRITE);
        writer.write(sb.toString());
        writer.close();
    
// Create the histograms (get data from the file we've just created)
        BufferedReader reader = Files.newBufferedReader(testDirectory.resolve(Path.of("adc-"+timestamp+".dat")));
        PdmdbModuleTester.plot(reader,svg,"adc");
      }
    }
    catch (IOException ioe){return "FAIL|TesterSequence.getAdc()|"+ioe;}

    return "OK|TesterSequence.getAdc()";
  }

/**
<p>Configures CLAROs for one EC.</p>
<p>Callable by {@link Command}.</p>
@param args {ecIndex}.
<p>ecIndex is the index of the EC in the list unmarshalled from XML, not the EC label.</p>
@return Command response.
*/
  public static String configureClaros(String [] args)
  {
    int ecIndex = Integer.decode(args[0]).intValue();
    
    ClaroControl claroControl = MuDaq.muDaq.pdmdbMaster.ecList.get(ecIndex).claro;
    boolean cfgOK = claroControl.configure();
    
    return "OK|TesterSequence.configureClaros|Verify:"+cfgOK;
  }
    
/**
<p>Does MGT eye scan.</p>
<p>Callable by {@link Command}.</p>
@param args {mgtId,prescale,maxErrors}.
@return Command response.
**/
public static String doEyeScan(String [] args)
{
  
  int drpsel = Integer.parseInt(args[0]);
  int prescale = Integer.parseInt(args[1]);
  if ( prescale > 31 ) prescale = 31;
  if ( prescale < 0 ) prescale = 0;
  int maxErrors = Integer.parseInt(args[2]);

  int noErrorsCount = 0;

  String rv = "OK|TesterSequence.doEyeScan({"+args[0]+","+args[1]+"})";

  try
  {
    StringBuffer stringBuffer = new StringBuffer();
    stringBuffer.append(String.format("eye%d\n",drpsel));

    DrpPanel.staticThis.eyeScanPrepare(drpsel);

    for (int x=0;x<32;x++)
    {
      for (int y=0;y<32;y++)
      {
        ArrayList<Integer> al = DrpPanel.staticThis.eyeScanDo(x,y,prescale);
        if (al.size() != 2) break;

	if (al.get(1) == 0) noErrorsCount++;

        stringBuffer.append(String.format("%d,%d,%d,%d\n",x,y,al.get(0),al.get(1))); // x,y,samples,errors
      }
    }

    if (testDirectory != null && TesterSequence.eyeScanDataDirectory != null)
    {
      BufferedWriter writer = Files.newBufferedWriter(TesterSequence.eyeScanDataDirectory.resolve(Path.of("eyeScan-"+drpsel+".txt")),StandardOpenOption.CREATE_NEW,StandardOpenOption.WRITE);
      writer.write(stringBuffer.toString());
      writer.close();
    

// Create the histograms (get data from the file we've just created)
      BufferedReader reader = Files.newBufferedReader(TesterSequence.eyeScanDataDirectory.resolve(Path.of("eyeScan-"+drpsel+".txt")));
      cbsw.gui.Histogram2D hist = new cbsw.gui.Histogram2D("");
      hist.setPaintScale(-5.5,1.5,1.4);
      try { hist.readStream(reader); } catch(cbsw.gui.Histogram2D.EOFException eof){}
      reader.close();

// Output the plots and dataset statistics
      hist.saveAsSVG(testDirectory.resolve(Path.of("eye"+drpsel+".svg")).toFile()); // Write as an SVG file
    }

    TesterSequence.xml("<eye id=\"eye"+drpsel+"\" binsWithoutErrors=\""+noErrorsCount+"\" />\n");

    rv = "OK|TesterSequence.doEyeScan({"+args[0]+","+args[1]+"})|Quality="+noErrorsCount;
    System.err.println("TesterSequence.doEyeScan(): "+noErrorsCount+" samples without errors");
  }
  catch (IOException ioe)
  {
    System.err.println("TesterSequence.doEyeScan(): "+ioe);
    rv = "FAIL|TesterSequence.doEyeScan({"+args[0]+","+args[1]+"})|IO exception";
  }

  if (noErrorsCount<maxErrors) rv = "FAIL|TesterSequence.doEyeScan({"+args[0]+","+args[1]+"})|Quality="+noErrorsCount;
  return rv;
}

/**
<p>Performs a DAC scan.</p>
<p>Callable by {@link Command}.</p>
<p>Does nothing if testDirectory is null.</p>
@param args Not used.
@return Command response.
**/
  static public String doDacScan(String [] args)
  {
    if (testDirectory == null) return "OK|TesterSequence.doDacScan()|No scan - output not enabled";
    
    Path dataDirectory;
// Create directory for phase scan data files
    try
    {
      dataDirectory = Files.createDirectories(testDirectory.resolve(Path.of("DacScanData")));
    }
    catch (Exception ex){return "FAIL|TesterSequence.doDacScan()|"+ex;}
    
// Make a single list of all the GBTXes
    List<GbtxControl> gbtxes = new ArrayList<GbtxControl>();
    for (DtmControl dtm : PdmdbMaster.pdmdbMaster.dtmList) gbtxes.addAll(dtm.gbtxList);
    
    try
    {

      MuDaq.cfg.put("tfc-d","0",ScanControl.staticThis.io); // Set a fixed pattern on the down links
      MuDaq.cfg.put("tfctog-enable","8",ScanControl.staticThis.io); // Enable the TFC test pulse
// Do scan
      String fileNameFormat = dataDirectory.toString()+"/dacScan.%d.%d.dat";
      ScanControl.staticThis.scan(ScanControl.ScanMode.DAC_SCAN,0,63,1,gbtxes,fileNameFormat);
    }
    catch ( Exception ex ){ return "FAIL|TesterSequence.doDacScan()|"+ex; }
    
// Make histograms

    ScanControl.makeDacScanPlots(dataDirectory);
    
    return "OK|TesterSequence.doDacScan()";
  }

/**
<p>Performs manual phase scan over all active links.</p>
<p>Callable by {@link Command}.</p>
<p>Does nothing if testDirectory is null.</p>
@param args Not used.
@return Command response.
 */
  static public String doPhaseScan(String [] args)
  {
    if (testDirectory == null) return "OK|TesterSequence.doPhaseScan|No scan - output not enabled";
    
// Create directory for phase scan data files
    Path dataDirectory;
    try
    {
      dataDirectory = Files.createDirectories(testDirectory.resolve(Path.of("PhaseScanData")));
    }
    catch (Exception ex){return "FAIL|TesterSequence.doPhaseScan()|"+ex;}
    String fileNameFormat = dataDirectory.toString()+"/phaseScan.%d.%d.dat";
    
// Make a single list of all the GBTXes
    List<GbtxControl> gbtxes = new ArrayList<GbtxControl>();
    for (DtmControl dtm : PdmdbMaster.pdmdbMaster.dtmList) gbtxes.addAll(dtm.gbtxList);
    
    try
    {
// Set a fixed pattern on the down links

      MuDaq.cfg.put("tfctog-enable","0",ScanControl.staticThis.io);
      MuDaq.cfg.put("tfc-d","A",ScanControl.staticThis.io);
// Do scan
      ScanControl.staticThis.scan(ScanControl.ScanMode.PHASE_SCAN,0,14,1,gbtxes,fileNameFormat);
    }
    catch ( Exception ex ){ return "FAIL|TesterSequence.doPhaseScan()|"+ex; }
    
// Make histograms

    ScanControl.makePhaseScanPlots(dataDirectory);
    
    return "OK|TesterSequence.doPhaseScan()";
  }

/**
<p>Prepares muDaq to measure bit error rate of e-links.</p>
<p>Callable by {@link Command}.</p>
@param args Not used.
@return Command response.
**/
public static String prepareBER(String [] args){ MuDaq.muDaq.prepareBER(); return "OK|TesterSequence.prepareBER";}

/**
<p>Gets the current bit error rate counters.</p>
<p>Callable by {@link Command}.</p>
@param args Not used.
@return Command response.
**/
public static String getBER(String [] args)
{
  java.nio.ByteBuffer bb = MuDaq.muDaq.getBER();
  TesterSequence.writeBERCounterXml(bb);
  
  return "OK|TesterSequence.getBER";
}

/**
<p>Performs a bit error count measurement.</p>
<p>Callable by {@link Command}.</p>
@param args Not used.
@return Command response.
**/
  public static String countBitErrors(String [] args)
  {
    try
    {
      MuDaq.muDaq.prepareBER();

      Thread.sleep(500);

      java.nio.ByteBuffer bb = MuDaq.muDaq.getBER();
      TesterSequence.writeBERCounterXml(bb);
    }
    catch (Exception ioe){ return "FAIL|TesterSequence.countBitErrors()|"+ioe; }
    return "OK|TesterSequence.countBitErrors()";
  }

/**
<p>Writes bit error rate counters to file.</p>
@param bb The buffer containing the counter values
*/
private static void writeBERCounterXml(java.nio.ByteBuffer bb)
{
// Make a single list of all the GBTXes
    List<GbtxControl> gbtxes = new ArrayList<GbtxControl>();
    for (DtmControl dtm : PdmdbMaster.pdmdbMaster.dtmList) gbtxes.addAll(dtm.gbtxList);
      
    try
    {
      for (GbtxControl gbtx: gbtxes )
      {
	bb.position((gbtx.mgt-2)*2*16);
	TesterSequence.xml(String.format("<bitErrors testerModule=\"dtm%d\" gbtx=\"%d\" ",gbtx.dtm,gbtx.gbtx));
	StringBuffer sb = new StringBuffer();
	for (int eport=0;eport<14;eport++)
	{
	  short count = bb.getShort();
	  sb.append(String.format("eport%d=\"%d\" ",eport,count));
	}
	bb.getInt(); // Swallow the padding
	TesterSequence.xml(sb.toString());
	String timestamp = java.time.LocalDateTime.now().format(java.time.format.DateTimeFormatter.ofPattern("yyyyMMdd-A"));
	TesterSequence.xml(String.format("timestamp=\"%s\" />\n",timestamp));
      }
    }
    catch (Exception ioe){ throw new RuntimeException( "TesterSequence.writeBERCounterXml(): "+ioe); }
}

/**
<p>Performs a test of the SPI interface.</p>
<p>Callable by {@link Command}.</p>
@param args Not used.
@return Command response.
 */ 
public static String spiTest(String [] args)
{
  ArrayList<Claro> claroList = MuDaq.muDaq.pdmdbMaster.ecList.get(0).claro.claros;
  ScaDepp sca = PdmdbMaster.pdmdbMaster.sca;
  
  PdmdbModuleTester.exec("TEST_SPI|11111111");
  PdmdbModuleTester.rxSeq++;
  PdmdbModuleTester.readResponse(3000);
  try {Thread.sleep(1000);} catch (InterruptedException e){}
  sca.allocateTxRxBuffers(10*claroList.size()); // 10 SCA commands per CLARO
  for ( Claro claro : claroList ) claro.prepareSpi(false);
  sca.putBytes(0,true);
  String [] spiResponse = PdmdbModuleTester.readResponse(15000);
  if (spiResponse.length > 2)
  {
    TesterSequence.xml("<spi>\n");
    for ( String ss : spiResponse )
    {
      if ( ss.startsWith("SS") )
      {
	int index = Integer.parseInt(ss.substring(2,3));
	String [] ssTokens = ss.split("[:,=]");
	xml("  <ss index=\""+index+"\" toggles=\""+ssTokens[2]+"\" clocks=\""+ssTokens[4]+"\" />\n");
      }
    }
    xml("</spi>\n");
  }
  
  return "OK|TesterSequence.spiTest";
}    

/**
<p>Writes summary of bitflip measurements to XML.</p>
<p>Callable by {@link Command}.</p>
@param args Not used.
@return Command response.
*/
public static String bitflipSummary(String [] args)
{
// Capture and save the bitflip counters

  try
  {
    MuDaq.muDaq.status.cache(MuDaq.muDaq.io);

// Make a single list of all the GBTXes
    List<GbtxControl> gbtxes = new ArrayList<GbtxControl>();
    for (DtmControl dtm : PdmdbMaster.pdmdbMaster.dtmList) gbtxes.addAll(dtm.gbtxList);
      
    xml("<bitflips>\n");
    for ( GbtxControl gbtx : gbtxes )
    {
      xml(String.format("<gbtx dtm=\"%d\" gbtx=\"%d\" flips=\"",gbtx.dtm,gbtx.gbtx));
      for (int elink=0;elink<14;elink++)
      {
	  //        int mgt = 2 + 2*gbtx.dtm + gbtx.gbtx;
        String hexCount = MuDaq.muDaq.status.get("gbt."+gbtx.mgt+".egroup."+elink/2+".elink."+elink%2+".bitflips");
// Convert the hex string to a decimal string
        try
        {
          Integer decoded = Integer.decode(hexCount);
          xml(decoded.toString());
        }
        catch (NumberFormatException nfe){xml("ovfl");}
        if (elink<13) xml(",");
      }
      xml("\" />\n");
    }
    xml("</bitflips>\n");
    return "OK|TesterSequence.bitflipSummary()";
  }
  catch (Exception ioe){ throw new RuntimeException("TesterSequence.bitflipSummary()|"+ioe);}
}

/**
<p>Writes summary of clock measurements to XML.</p>
<p>Callable by {@link Command}.</p>
@param args Not used.
@return Command response.
*/
public static String clockSummary(String [] args)
{
  if (args.length < 19) throw new RuntimeException("TesterSequence.clockSummary()|Insufficient data in response");
  xml(String.format("<clocks>\n"));
  for (String token : args)
  {
    String [] clock = token.split("[:,=]");
    if (clock.length < 2) continue;
    if (clock[0].matches("TCM_CLK[012345]") || clock[0].matches("DTM[012]_FPGA_CLK") || clock[0].matches("DTM[012]_GBTX1_CLK_OUT" ))
    {
      xml(String.format("<clock id=\"%s\" frequency=\"%s\" />\n",clock[0].toLowerCase(),clock[1]));
    }
  }
  xml(String.format("</clocks>\n"));
  return "OK|TesterSequence.clockSummary()";
}
    
/**
<p>Writes summary of power measurements to XML.</p>
<p>Callable by {@link Command}.</p>
@param args Not used.
@return Command response.
*/
public static String powerSummary(String [] args)
{
  if (args.length < 7) throw new RuntimeException("TesterSequence.powerSummary()|Insufficient data in response");
  for (String token : args)
  {
    String [] module = token.split("[:,=]");
    if (module.length < 5) continue;
    if (module[0].matches("DTM[012]") || module[0].equals("TCM"))
    {
      String timestamp = java.time.LocalDateTime.now().format(java.time.format.DateTimeFormatter.ofPattern("yyyyMMdd-A"));
      xml(String.format("<power testerModule=\"%s\" i1v5=\"%s\" i2v5=\"%s\" timestamp=\"%s\" />\n",module[0].toLowerCase(),module[2],module[4],timestamp));
    }
  }
  return "OK|TesterSequence.powerSummary()";
}
    
/**
<p>Writes summary of temperature measurements to XML.</p>
<p>Callable by {@link Command}.</p>
@param args Not used.
@return Command response.
*/
public static String temperatureSummary(String [] args)
{
  if (args.length < 7) throw new RuntimeException("TesterSequence.temperatureSummary()|Insufficient data in response");
  for (String token : args)
  {
    String [] module = token.split("[:=]");
    if (module.length < 2) continue;
    if (module[0].matches("DTM[012]") || module[0].equals("TCM"))
    {
      String timestamp = java.time.LocalDateTime.now().format(java.time.format.DateTimeFormatter.ofPattern("yyyyMMdd-A"));
      xml(String.format("<temperature testerModule=\"%s\" reading=\"%s\" timestamp=\"%s\" />\n",module[0].toLowerCase(),module[1],timestamp));
    }
  }
  return "OK|TesterSequence.temperatureSummary()";
}
    
/**
<p>Jumps to a specified step in the sequence.</p>
<p>Callable by {@link Command}.</p>
<p>Use this method to implement simple loops in a sequence.</p>
@param args {goto,iterations,wait}.
<p> goto is an integer that identifies the step in the sequence to jump to. iterations is the integer number of iterations. wait is the time in seconds to wait between iterations.</p>
@return Command response.
*/
public static String goTo(String [] args)
{
  try
  {
    if (args.length < 3) throw new RuntimeException("TesterSequence.goTo()|Insufficient data in command");
    if (staticThis.iterations == 0)
    {
      staticThis.it = staticThis.commandList.listIterator(Integer.valueOf(args[0]));
      staticThis.iterations = Integer.valueOf(args[1]);
    }
    else if (staticThis.iterations == 1)
    {
      Thread.sleep(Integer.valueOf(args[2])*1000);
      staticThis.iterations--;
    }
    else
    {
      staticThis.it = staticThis.commandList.listIterator(Integer.valueOf(args[0]));
      Thread.sleep(Integer.valueOf(args[2])*1000);
      staticThis.iterations--;
    }
    return "OK|TesterSequence.goTo()";

  }
  catch (Exception ex){ throw new RuntimeException("TesterSequence.goTo(): "+ex); }
}
    
/**
<p>Prepares to run the sequence.</p>
<p>The current sequence position is set to the beginning.</p>
 */
public void prepare()
{
  for (Command cmd : commandList ) cmd.clearStatus();

  it = commandList.listIterator();
}

/**
<p>Pauses the sequence.</p>
<p>The currently running step will continue to execute until it completes.</p>
 */
public void pause(){suspend=true;}

/**
<p>Continues the sequence after pause.</p>
 */
public void resume()
{
  suspend=false;
  TestTask tt = new TestTask();
  tt.execute();
}

/**
<p>Toggles the recording flag.</p>
 */
public void recording()
{
  recording=!recording;
  if (recording) recordingButton.setBackground(java.awt.Color.RED);
  else recordingButton.setBackground(new java.awt.Color(220,220,220));
}
    
/**
<p>Executes the next step of the sequence.</p>
 */
public void step()
{
  if (it==null) it = commandList.listIterator();
    
  try { if (it.hasNext()) it.next().go(); }
  catch (Exception ex) {PdmdbModuleTester.readFlush(); textArea.append("TesterSequence.step(): "+ex+"\n");}
}
    
/**
<p>Executes all the steps of the sequence.</p>
 */
public void go()
{
  for (Command cmd : commandList ) cmd.clearStatus();
  
  try { for (Command cmd : commandList ) cmd.go(); }
  catch (Exception ex) {PdmdbModuleTester.readFlush(); textArea.append("TesterSequence.go(): "+ex+"\n");}
}

/**
<p>Launches the sequence in its own thread (using {@link javax.swing.SwingWorker}).</p>
 */
public void launch()
{
  TestTask tt = new TestTask();
  tt.execute();
}
    
class TestTask extends javax.swing.SwingWorker<Boolean, Integer>
{

  @Override
  protected Boolean doInBackground()
  {
    boolean pass = true;
    try
    {
      while(!suspend)
      {
        for (Command cmd : commandList ) cmd.clearStatus();
	textArea.setText("");
        while(it.hasNext())
        {
          it.next().go();
          if (suspend) break;
        }
        it = commandList.listIterator();
      }
    }
    catch (Exception ex) {textArea.append("TesterSequence.TestTask.doInBackground(): "+ex+"\n");}
    return Boolean.valueOf(pass);
  }
      
  @Override
  protected void done(){ try {  } catch (Exception e){} }

  @Override
  protected void process( List<Integer> chunks ){}
}
    
}
