package cbsw.adept;

import javax.swing.*;
import java.awt.event.*;
import javax.swing.Timer;
import java.util.concurrent.locks.*;
import java.util.prefs.*;
import java.util.prefs.Preferences;
import java.util.prefs.BackingStoreException;
import java.util.*;
import java.io.*;
import java.net.*;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import cbsw.TapBuilder;
import cbsw.RegisterIoException;

/**
<p>This class encapsulates and extends the Digilent Adept2 APIs through Jni calls.</p>
<p>Most target hardware now uses the Depp2 protocol in which the
transaction buffer contains 4 header bytes. The first two bytes
contain the payload length (that is, excluding the header bytes) and
the second two bytes contain the register start offset. The byte order
for the header bytes is little-endian.</p>
<p>Note that the hardware treats the MSB of the register start offset
as a flag to activate the DEPP wait feature. Therefore the highest
allowed offset is 0x7fff. The wait feature allows a subsequent
transaction to be started immediately after the previous call returns
and it will block until the previous transaction completes. This
avoids the need for polling in many situations if the feature has been
implemented in the hardware.</p>
<p>Most of the methods described here that read or write the hardware
implement a locking mechanism to grant exclusive access to the
hardware device. This only applies for calls from within the same program.</p>
@author Steve Wotton, Cavendish Lab. (HEP), 2018-
*/
public class JniAdeptDepp implements cbsw.RegisterIo
{

  @Deprecated
  private BitSet regBits;
//  static private ReentrantLock djtgLock;
//  static private ReentrantLock deppLock;
  static private ReentrantLock lock;

  static
  {
    System.loadLibrary("JniAdeptDepp");

//    djtgLock = new ReentrantLock();
//    deppLock = new ReentrantLock();
    lock = new ReentrantLock();
  }

/**<p>Import database of known hardware from web database.</p>
@see #selectConnectedDevice
*/
  @Deprecated
  static ActionListener importHWAL = ae -> importHardware("http://www.hep.phy.cam.ac.uk/lhcb/RICHEC/xml/XilinxHardwarePrefs.xml");

  @Deprecated
  static ActionListener enumerateAL = ae -> showEnumeratedDevices();

/**<p>A convenience function to add some standard menu items to a menu.</p>
@param menu The menu to which the items are added.
*/
  public static void addMenuItems(JMenu menu)
  {
/*
JMenuItem menuItem = new JMenuItem("Import hardware...");
menuItem.addActionListener( ae -> importHardware("http://www.hep.phy.cam.ac.uk/lhcb/RICHEC/xml/XilinxHardwarePrefs.xml") );
m.add(menuItem);
*/
      
    JMenuItem menuItem = new JMenuItem("Enumerate");
    menuItem.addActionListener( ae -> showEnumeratedDevices() );
    menu.add(menuItem);
  }

/**<p>A DEPP exception that encapsulates the DEPP error codes/messages.</p>*/
  public class DeppException extends cbsw.RegisterIoException
  {
/**<p>Constructor - Create a DEPP exception that encapsulates the DEPP error codes/messages.</p>*/
    public DeppException()
    {
      super(JniAdeptDepp.dmgr.getLastError(),JniAdeptDepp.dmgr.getLastErrorName(JniAdeptDepp.dmgr.getLastError()),JniAdeptDepp.dmgr.getLastErrorDescription(JniAdeptDepp.dmgr.getLastError()));
    }
  };

  private Timer timer;
  private String deviceString;

  ActionListener timerAL = ae ->
    {
// Open the device if not already open

      try
      {
	close();
	open();
      }
      catch ( DeppException de ){System.out.println("JniAdeptDepp.timerAL(): Exception opening "+deviceString);}
    };

/**<p>Constructor</p>
@param dev Connection name of device
*/
  public JniAdeptDepp( String dev )
  {
    regBits = new BitSet(2048);
    deviceString = dev;

    timer = new Timer(1000, timerAL );
    timer.setInitialDelay(0);
    timer.setCoalesce( true );
        
  }

/**<p>Constructor</p>
@param dev Connection name of device
@param al ActionListener to replace default timer (can be null)
*/
  public JniAdeptDepp( String dev, ActionListener al )
  {
    regBits = new BitSet(2048);
    deviceString = dev;

    timer = new Timer(1000, al );
    timer.setInitialDelay(0);
    timer.setCoalesce( true );
        
  }

/**<p>Grants exclusive access to DEPP interface.</p><p>Blocks until interface is available.</p>*/
  public void lock(){lock.lock();}
/**<p>Releases access to DEPP interface.</p>*/
  public void unlock(){lock.unlock();}

/**<p>Native interface wrappers to Digilent Adept DMGR API.</p>*/
  public static class dmgr
  {
/**<p>Opens a device by connection name.</p>
<p>Native wrapper.</p>
@param dev The connection name of the device
@return Return code, 1=success, 0=fail
*/
    public static native int open( String dev );
/**<p>Closes the device.</p>
<p>Native wrapper.</p>
@return Return code (always 1)
*/
    public static native int close();
/**<p>Enumerates devices.</p>
<p>Native wrapper.</p>
@return Number of devices found
*/
    public static native int enumerate();
/**<p>Gets the connection name of an enumerated device.</p>
<p>Native wrapper.</p>
@param idx Enumeration index
@return The connection name
*/
    public static native String getEnumeratedDevice( int idx );
/**<p>Gets the capabilities of an enumerated device.</p>
<p>Native wrapper.</p>
@param idx Enumeration index
@return Capabilities string
*/
    public static native String getCapabilities( int idx );
/**<p>Gets the last error code.</p>
<p>Native wrapper.</p>
@return Error code
*/
    public static native int getLastError();
/**<p>Gets the last error name.</p>
<p>Native wrapper.</p>
@param erc Numeric code of message name
@return Error name
*/
    public static native String getLastErrorName(int erc);
/**<p>Gets the last error description.</p>
<p>Native wrapper.</p>
@param erc Numeric code of message name
@return Error description
*/
    public static native String getLastErrorDescription(int erc);
  }

/**<p>Native interface wrappers to Digilent Adept DEPP API.</p>*/
  public static class depp
  {
/**<p>Enables the DEPP channel on the opened device.</p>
<p>Native wrapper.</p>
@return Return code, 1=success, 0=fail
*/
    public static native int enable();
/**<p>Disables the DEPP channel on the opened device.</p>
<p>Native wrapper.</p>
@return Return code, 1=success, 0=fail
*/
    public static native int disable();
/**<p>Reads a set of addresses.</p>
<p>Native wrapper.</p>
@param addr Array of addresses to read
@param data Data returned from the requested addresses
@return Return code, 1=success, 0=fail
*/
    public static native int getSet( byte [] addr, byte [] data );
/**<p>Reads repeatedly from the same address.</p>
<p>Native wrapper.</p>
@param addr The address to read
@param buf Data returned
@return Return code, 1=success, 0=fail
*/
    public static native int getRepeat( byte addr, byte [] buf );
/**<p>Writes a set of address/data pairs.</p>
<p>Native wrapper.</p>
@param buf Array of address/data pairs {addr0,data0,addr1,data1...}
@return Return code, 1=success, 0=fail
*/
    public static native int putSet( byte [] buf );
/**<p>Writes repeatedly to the same address.</p>
<p>Native wrapper.</p>
@param addr The address to write
@param buf Data to write
@return Return code, 1=success, 0=fail
*/
    public static native int putRepeat( byte addr, byte [] buf );
  }

// DJTG functions

/**<p>Native interface wrappers to Digilent Adept DJTG API.</p>*/
  public static class djtg
  {
/**<p>Enables the DJTG channel on the opened device.</p>
<p>Native wrapper.</p>
@return Return code, true=success, false=fail
*/
    public static native boolean enable();
/**<p>Disables the DJTG channel on the opened device.</p>
<p>Native wrapper.</p>
@return Return code, true=success, false=fail
*/
    public static native boolean disable();
    public static native String getProperties(int port);
    public static native boolean putTmsTdi(int bits, byte [] snd, byte [] rcv);
    public static native boolean putTmsTdi(int bits, byte [] snd);
    public static native boolean putTms(int bits, byte [] snd, byte [] rcv, boolean tdi);
    public static native boolean putTms(int bits, byte [] snd, boolean tdi);
    public static native boolean putTdi(int bits, byte [] snd, byte [] rcv, boolean tms);
    public static native boolean putTdi(int bits, byte [] snd, boolean tms);
    public static native boolean clock(boolean tms, boolean tdi, int clks);
    public static byte [] packAndGo(TapBuilder tap, boolean rcv)
    {
      int bits = tap.length();
      if (bits < 1) return null;

//        tap.setDebug(true);
      tap.show();

      ByteBuffer bb = tap.tmsTdoToByteBuffer();
      tap.setDebug(false);
      boolean rc = false;
      byte [] tdi = null;
      if (rcv)
      {
	tdi = new byte [bb.capacity()/2];
	rc = JniAdeptDepp.djtg.putTmsTdi(bits,bb.array(),tdi);
      }
      else
      {
	rc = JniAdeptDepp.djtg.putTmsTdi(bits,bb.array());
      }
      if (!rc) System.err.println("JniAdeptDepp.djtg.packAndGo(): putTmsTdi() failed.");
      return tdi;
    }
  }

  @Deprecated
  public boolean OK(){ return true; }

/**<p>Writes information about enumerated devices to System.out</p>*/
  public static void showEnumeratedDevices()
  {
    int ndepp = JniAdeptDepp.dmgr.enumerate();
    if (ndepp == 0) System.out.println("JniAdeptDepp.showEnumeratedDevices(): No devices found.");
    for ( int idepp=0;idepp<ndepp;idepp++)
    {
      System.out.println(JniAdeptDepp.dmgr.getEnumeratedDevice(idepp));
    }
  }

/**<p>Starts a timer that periodically tries to connect to a device until it succeeds.</p>*/
  public void connect()
  {
    timer.start();
    return;
  }

/**<p>Stops a timer that periodically tries to connect to a device</p>
<p>After stopping the timer, closes the device.</p>*/
  public void disconnect()
  {
    timer.stop();
    close();
    return;
  }

/**<p>Opens a device</p>
@throws DeppException DEPP exception
*/
  public void open() throws DeppException
  {
    try
    {
      JniAdeptDepp.lock.lock();
      if ( JniAdeptDepp.dmgr.open( deviceString ) != 1 ){throw new DeppException();}
    }
    finally { JniAdeptDepp.lock.unlock(); }

    return;
  }

/**<p>Closes a device</p>*/
  public void close()
  {
    try
    {
      JniAdeptDepp.lock.lock();
      int rc = JniAdeptDepp.dmgr.close();
    }
    finally { JniAdeptDepp.lock.unlock(); }

    return;
  }

/**<p>Gets single byte</p>
@param addr The register address
@param buf Array to receive the data
@throws DeppException DEPP exception
*/
  public void get( byte addr, byte [] buf ) throws DeppException
  {
    try
    {
      JniAdeptDepp.lock.lock();
      byte [] addrArray = new byte[1]; addrArray[0] = addr;
      if ( JniAdeptDepp.depp.getSet( addrArray, buf ) != 1 ){throw new DeppException();}
    }
    finally { JniAdeptDepp.lock.unlock(); }

    return;
  }

/**<p>Gets multiple bytes</p>
@param addr Array of addresses to read
@param data Array to receive the data
@throws DeppException DEPP exception
*/
  public void get( byte [] addr, byte [] data ) throws DeppException
  {
    try
    {
      JniAdeptDepp.lock.lock();
      if ( JniAdeptDepp.depp.getSet( addr, data ) != 1 ){throw new DeppException();}
    }
    finally { JniAdeptDepp.lock.unlock(); }

    return;
  }

/**<p>Puts single addr,byte</p>
@param base Address to write
@param data Data to write
@throws DeppException DEPP exception
*/
  public void put( byte base, byte data ) throws DeppException
  {
    try
    {
      JniAdeptDepp.lock.lock();
      byte [] buf = new byte[2]; buf[0] = base; buf[1] = data;
      if ( JniAdeptDepp.depp.putSet( buf ) != 1 ){throw new DeppException();}
    }
    finally { JniAdeptDepp.lock.unlock(); }

    return;
  }

/**<p>Puts multiple addr,byte pairs</p>
@param buf Array of address/data pairs to write
@throws DeppException DEPP exception
*/
  public void put( byte [] buf ) throws DeppException
  {
    try
    {
      JniAdeptDepp.lock.lock();
      if ( JniAdeptDepp.depp.putSet( buf ) != 1 ){throw new DeppException();}
    }
    finally { JniAdeptDepp.lock.unlock(); }

    return;
  }

  private void bbPut(ByteBuffer bb, Integer val){bb.putInt(val);}
  private void bbPut(ByteBuffer bb, Short val){bb.putShort(val);}
  private void bbPut(ByteBuffer bb, Long val){bb.putLong(val);}
  private void bbPut(ByteBuffer bb, Byte val){bb.put(val);}

/**
<p>Writes a value to hardware in little-endian order (Depp2)</p>
<p>Consider using {@link #put(int,java.nio.ByteBuffer)} instead.</p>
<p>Integer, Short, Long and Byte values are allowed, others result in RuntimeException.</p>
@param <T> Type of value to write
@param base DEPP base address
@param addr Register offset
@param val Value to write
@throws cbsw.RegisterIoException RegisterIo exception
*/
  public <T extends Byte,Integer,Short,Long> void put( int base, int addr, T val ) throws RegisterIoException
  {
    int size = T.SIZE/8;

    java.nio.ByteBuffer bb = java.nio.ByteBuffer.allocate(4+size);
    bb.order(java.nio.ByteOrder.LITTLE_ENDIAN);
    bb.putShort((short)size);
    bb.putShort((short)addr);

    bbPut(bb,(T)val);

    try
    {
      JniAdeptDepp.lock.lock();
      if (JniAdeptDepp.depp.putRepeat( (byte)base, bb.array() ) != 1){ throw new DeppException(); }
    }
    finally { JniAdeptDepp.lock.unlock(); }
    return;
  }

/**<p>Writes array of bytes to hardware (Depp2)</p>
<p>Use {@link #put(int,java.nio.ByteBuffer)} instead.</p>
@param base DEPP base address
@param len Transaction length
@param addr Register start offset
@param buf Values to write
@param waitFlag If true, activates the DEPP wait in hardware
@throws DeppException DEPP exception
*/
  @Deprecated
  public void push(int base, int len, int addr, byte [] buf, boolean waitFlag) throws DeppException
  {
    buf[0] = (byte)len; buf[1] = (byte)(len>>8); // Length
    buf[2] = (byte)addr; buf[3] = (byte)((addr>>8)&0x7f); // Address
    if (waitFlag) buf[3] |= (byte)0x80;
    try { push((byte)base,buf); } catch (DeppException de ){ throw de; }
  }

/**<p>Writes array of bytes to hardware</p>
<p>Use {@link #put(int,java.nio.ByteBuffer)} instead.</p>
@param base DEPP base address
@param buf Values to write
@throws DeppException DEPP exception
*/
  @Deprecated
  public void push( byte base, byte [] buf ) throws DeppException
  {
    try
    {
      JniAdeptDepp.lock.lock();
      if (JniAdeptDepp.depp.putRepeat( base, buf ) != 1) {throw new DeppException();}
    }
    finally { JniAdeptDepp.lock.unlock(); }

    return;
  }

/**<p>Reads array of bytes from hardware</p>
<p>Use {@link #get(int,byte[])} instead.</p>
@param base DEPP base address
@param buf Values to write
@throws DeppException DEPP exception
*/
  @Deprecated
  public void pop( byte base, byte [] buf ) throws DeppException
  {
    try
    {
      JniAdeptDepp.lock.lock();
      if ( JniAdeptDepp.depp.getRepeat( base, buf ) != 1 ){throw new DeppException();}
    }
    finally { JniAdeptDepp.lock.unlock(); }

    return;
  }

/**<p>Writes array of bytes to legacy hardware.</p>
<p>Avoid using this for new applications. May be required for some legacy hardware.</p>
@param base DEPP base address
@param addr Register start offset
@param regs Values to write
@throws DeppException DEPP exception
*/
  public void put1(int base, int addr, byte [] regs) throws DeppException
  {
    byte [] hdr = {(byte)base, (byte)addr, (byte)(base+1), (byte)(addr>>8)};
    try
    {
      JniAdeptDepp.lock.lock();
      if (JniAdeptDepp.depp.putSet( hdr ) != 1){ throw new DeppException(); }
      if (JniAdeptDepp.depp.putRepeat( (byte)(base+2), regs ) != 1){ throw new DeppException(); }
    }
    finally { JniAdeptDepp.lock.unlock(); }
    return;
  }

/**<p>Writes array of bytes to hardware (Depp2)</p>
<p>In contrast to {@link put(int,ByteBuffer)} this method initialises
the first four header bytes. The first two bytes are filled with the
data payload length (regs.length-4) and the second two bytes are
filled with the register start offset (addr). Any values in these locations
in the array passed to this method will therefore be overwritten.</p>
@param base DEPP base address
@param addr Register start offset
@param regs Write transaction buffer, payload starting at byte 4
@throws cbsw.RegisterIoException Register I/O exception
*/
  public void put(int base, int addr, byte [] regs) throws cbsw.RegisterIoException
  {
// Write the header at the beginning of the buffer
    int idx = 0;
    int dataLength = regs.length - 4;
    regs[idx++] = (byte)dataLength;
    regs[idx++] = (byte)(dataLength>>8);
    regs[idx++] = (byte)addr;
    regs[idx++] = (byte)(addr>>8);

    try
    {
      JniAdeptDepp.lock.lock();
      if (JniAdeptDepp.depp.putRepeat( (byte)base, regs ) != 1){ throw new DeppException(); }
    }
    finally { JniAdeptDepp.lock.unlock(); }
    return;
  }

/**
<p>Writes byte buffer to hardware (Depp2)</p>
<p>The buffer should contain the payload length (2 bytes, little-endian, excludes the 4 header bytes)
followed by the register start offset (2 bytes, little-endian) followed by the data.</p>
@param base DEPP base address
@param bb Write transaction buffer
@throws cbsw.RegisterIoException Register I/O exception
*/
  public void put(int base, java.nio.ByteBuffer bb) throws cbsw.RegisterIoException
  {
    try
    {
      JniAdeptDepp.lock.lock();
      if (JniAdeptDepp.depp.putRepeat( (byte)base, Arrays.copyOfRange(bb.array(),0,bb.limit()) ) != 1){ throw new DeppException(); }
    }
    finally { JniAdeptDepp.lock.unlock(); }
    return;
  }

/**<p>Reads consecutive bytes from legacy hardware.</p>
<p>Reads regs.length bytes.</p>
@param base DEPP base address
@param addr Register start offset
@param regs Array to receive data
@throws DeppException DEPP exception
*/
  public void getBytes(int base, int addr, byte [] regs) throws DeppException
  {
    byte [] hdr = {(byte)base, (byte)addr, (byte)(base+1), (byte)(addr>>8)};
    try
    {
      JniAdeptDepp.lock.lock();
      if (JniAdeptDepp.depp.putSet( hdr ) != 1){ throw new DeppException(); }
      if (JniAdeptDepp.depp.getRepeat( (byte)(base+2), regs ) != 1){ throw new DeppException(); }
    }
    finally { JniAdeptDepp.lock.unlock(); }
    return;
  }

/**<p>Reads consecutive bytes from hardware (Depp2).</p>
<p>Reads regs.length bytes.</p>
@param base DEPP base address
@param addr Register start offset
@param regs Array to receive data
@throws cbsw.RegisterIoException Register I/O exception
*/
  public void get(int base, int addr, byte [] regs) throws cbsw.RegisterIoException
  {
    byte [] hdr = new byte [4];
    int idx = 0;
// Write the header
    int dataLength = regs.length;
    hdr[idx++] = (byte)dataLength;
    hdr[idx++] = (byte)(dataLength>>8);
    hdr[idx++] = (byte)addr;
    hdr[idx++] = (byte)(addr>>8);
    
    try
    {
      JniAdeptDepp.lock.lock();
      if (JniAdeptDepp.depp.putRepeat( (byte)base, hdr ) != 1){ throw new DeppException(); }
      if (JniAdeptDepp.depp.getRepeat( (byte)base, regs ) != 1){ throw new DeppException(); }
    }
    finally { JniAdeptDepp.lock.unlock(); }

    return;
  }

/**<p>Reads consecutive bytes from hardware (Depp2).</p>
<p>Reads regs.length bytes.</p>
<p>This version of get is intended to be called after the header has been sent.</p>
@param base DEPP base address
@param regs Array to receive data
@throws cbsw.RegisterIoException Register I/O exception
*/
  public void get(int base, byte [] regs) throws cbsw.RegisterIoException
  {
    try
    {
      JniAdeptDepp.lock.lock();
      if (JniAdeptDepp.depp.getRepeat( (byte)base, regs ) != 1){ throw new DeppException(); }
    }
    finally { JniAdeptDepp.lock.unlock(); }

    return;
  }

  @Deprecated
  public byte [] getByteArray(int size)
  {
    byte [] cfg = new byte [size];
    System.arraycopy(regBits.toByteArray(),0,cfg,0,regBits.toByteArray().length);   return cfg;
  };

  @Deprecated
  public byte [] getByteArray(int size, int off)
  {
    byte [] cfg = new byte [size];
    int bsLen = regBits.toByteArray().length;
//   System.out.println("size="+size+" off="+off+" bsLen="+bsLen );
    if (bsLen < off) return cfg; // All bits zero
    int copySize = (bsLen < (size+off)) ? bsLen-off : size;
//   System.out.println("copySize="+copySize );
    System.arraycopy(regBits.toByteArray(),off,cfg,0,copySize);
    return cfg;
  };

/**<p>Imports list of known hardware from web (legacy code)</p>
@param url The URL of the XML data
*/
  @Deprecated
  static public void importHardware(String url)
  {

    System.out.println("JniAdeptDepp.importHardware(): Importing "+url);
    
// Make a URL representing the source

    try
    {

      URL hwURL = new URL(url);
      if ( hwURL == null )
      {
        System.err.println("Error forming data source URL: " + url );
        return;
      }
      InputStream hwStream = hwURL.openStream();
      Preferences.userRoot().importPreferences(hwStream);
      hwStream.close();

    }
    catch ( MalformedURLException ue ){ System.err.println("Warning: " + ue.getMessage() ); }
    catch ( EOFException eofe ){ System.err.println( "End of file" ); }
    catch ( IOException ioe ){ System.err.println("Warning: " + ioe.getMessage() ); }
    catch (InvalidPreferencesFormatException ex) { ex.printStackTrace(); }

    return;
  }

  @Deprecated
  static public JMenu makeDeviceMenu(String title, ActionListener aL)
  {
    JMenu menu = new JMenu(title);
    try
    {
      for ( String dna : Preferences.userRoot().node("Hardware/Xilinx").childrenNames() )
      {
        JMenuItem menuItem = new JMenuItem(dna);
        menuItem.addActionListener(aL);
        menu.add(menuItem);
      }
    } catch (BackingStoreException ex) {}

    return menu;
  }

  @Deprecated
  static public String selectDevice(String filter)
  {
    ArrayList<String> hwList = new ArrayList<String>();
    if (filter == null) filter = new String("chimaera2v3");
    try
    {
      for ( String dna : Preferences.userRoot().node("Hardware/Xilinx").childrenNames() )
      {
	Preferences hwPrefs = Preferences.userRoot().node("Hardware/Xilinx").node(dna);
	if (hwPrefs.get("typeName","chimaera2v3").equals(filter))
	{
	  hwList.add(dna);
	}
      }
    } catch (BackingStoreException ex) {}

    Object [] hwArray = hwList.toArray();

    return (String) JOptionPane.showInputDialog(null, "Select hardware", "EC tester", JOptionPane.QUESTION_MESSAGE, null, hwArray, hwArray[0]);
  }

/**<p>Creates a dialog to allow the user to select a connected device</p>
@return Data describing selected device (or null if nothing selected)
*/
  static public String [] selectConnectedDevice()
  {
    ArrayList<String> hwList = new ArrayList<String>();
    int ndepp = JniAdeptDepp.dmgr.enumerate();
    for ( int idepp=0;idepp<ndepp;idepp++)
    {
      hwList.add(JniAdeptDepp.dmgr.getEnumeratedDevice(idepp));
//      System.out.println(JniAdeptDepp.dmgr.getCapabilities(idepp));
    }

    Object [] hwArray = hwList.toArray();

    if (hwArray.length < 1) return null;

    String selectedDevice = (String) JOptionPane.showInputDialog(null, "Select hardware", "Enumerated devices", JOptionPane.QUESTION_MESSAGE, null, hwArray, hwArray[0]);

    if (selectedDevice==null) return null;

    String [] deviceTokens = selectedDevice.split("/");

    if (deviceTokens.length != 2) return null;

    return deviceTokens;
  }

  @Deprecated
  static public JComboBox<String> makeDeviceComboBox(String selectedName, ActionListener aL, String filter)
  {
    if (filter == null) filter = new String("chimaera2v3");
    JComboBox<String> cB = new JComboBox<String>();
    if (aL != null) cB.addActionListener(aL);
    try
    {
      for ( String dna : Preferences.userRoot().node("Hardware/Xilinx").childrenNames() )
      {
	Preferences hwPrefs = Preferences.userRoot().node("Hardware/Xilinx").node(dna);
	if (hwPrefs.get("typeName","chimaera2v3").equals(filter))
	{
	  cB.addItem(dna);
	  if (dna.equals(selectedName))
	  {
	    cB.setSelectedItem(dna);
	  }
	}
      }
    } catch (BackingStoreException ex) {}

    return cB;
  }

  @Deprecated
  static public Preferences getDevicePreferences(Class cl, String node)
  {
    Preferences mainPrefs = Preferences.userNodeForPackage(cl).node(node);
    String hw = mainPrefs.get("hardwareId","default");
    return Preferences.userRoot().node("Hardware/Xilinx").node(hw);
  }
}
