/*
Implements Xilinx FPGA programming and other functionality through the
Digilent Adept DJTG interface.

The programming function should work for both Series7 FPGAs and Spartan6 and
probably others.

Other functions, such as XADC access is only available on some devices.
 */
package cbsw.adept.xilinx;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.filechooser.*;
import java.io.*;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.file.Files;
import java.util.*;
import java.util.stream.*;
import java.util.function.*;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;
import cbsw.*;
import cbsw.adept.*;
import cbsw.adept.JniAdeptDepp.*;
import cbsw.gui.*;
import cbsw.TapBuilder;

public class XilinxDjtg implements ActionListener
{
  Preferences prefs;
  JniAdeptDepp jniAdept;
  int chainPosition;
  ProgramThread pth;
  TapBuilder tap;
  JTextField statusTF;
  JTextField binTF;
  byte [] tdi;
  final int txPages = 1;
  final int txPageSize = 128;
  JProgressBar progressBar;
  byte [] swap;
    
public XilinxDjtg(int chainLength, int chainPosition)
{

  init();
  this.chainPosition = chainPosition;
  
  tap = new TapBuilder(chainLength,chainPosition);

}

public void init()
{
  prefs = Preferences.userNodeForPackage(XilinxDjtg.class).node("XilinxDjtg");

// Bitswap LUT
  swap = new byte [256];
  for (int i=0;i<256;i++)
  {
    int sr = i;
    for (int b=0;b<8;b++)
    {
	swap[i]<<=1;
	swap[i] |= (sr & 0x1);
	sr>>=1;
    }
  }
}
    
public void setConnectedDevice(JniAdeptDepp jniAdept){this.jniAdept = jniAdept;}
    
public void setConnectedDevice()
{
  String [] deviceTokens = JniAdeptDepp.selectConnectedDevice();
  if (deviceTokens==null || deviceTokens.length < 2)
  {
    System.out.println("No connected device found. Exiting...");
    System.exit(0);
  }
  System.out.println("Connecting to "+deviceTokens[1]);
  jniAdept = new JniAdeptDepp(deviceTokens[1]);

  try { jniAdept.open(); JniAdeptDepp.djtg.enable(); } catch (DeppException de) {System.err.println("XilinxDjtg.main(): Depp exception. "+de.getDescription());}
}

  public GuiJPanel makePanel()
  {
    GuiJPanel jP = new GuiJPanel(prefs,"");

    jP.setName( "FPGA."+chainPosition );
    jP.setBorder( BorderFactory.createTitledBorder("FPGA."+chainPosition) );

    jP.setLayout(new GridBagLayout());
    GridBagConstraints gbC = new GridBagConstraints();
    gbC.fill = GridBagConstraints.VERTICAL;
    gbC.anchor = GridBagConstraints.CENTER;
    gbC.insets = new Insets(1,1,1,1);
    gbC.weightx = 1.0f;
    gbC.weighty = 1.0f;
    gbC.fill = GridBagConstraints.BOTH;

    JButton button;

//     button = new JButton("Program");
//     button.addActionListener(ae->jprogram());
//     gbC.gridwidth = 1;
//     jP.add(button,gbC);

    button = new JButton("Load FPGA");
    button.addActionListener(this);
    button.setActionCommand("load");
    gbC.gridwidth = 1;
    jP.add(button,gbC);

    button = new JButton("Choose file");
    button.addActionListener(this);
    button.setActionCommand("choose");
    gbC.gridwidth = 1;
    jP.add(button,gbC);

    binTF = new JTextField(prefs.get("bin.file-name","pdmdb.bin"));
    binTF.setEditable(false);
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    jP.add( binTF, gbC );
    
    button = new JButton("S6 DNA");
    button.addActionListener(this);
    button.setActionCommand("s6dna");
    gbC.gridwidth = 1;
    jP.add(button,gbC);

    button = new JButton("Status");
    button.addActionListener(this);
    button.setActionCommand("status");
    gbC.gridwidth = 1;
    jP.add(button,gbC);

    statusTF = new JTextField(10);
    statusTF.setEditable(false);
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    jP.add( statusTF, gbC );
    
    button = new JButton("JSHUTDOWN");
    button.addActionListener(ae->jshutdown());
    gbC.gridwidth = 1;
    //    jP.add(button,gbC);

    button = new JButton("JSTART");
    button.addActionListener(ae->jstart());
    gbC.gridwidth = 1;
    //    jP.add(button,gbC);

    button = new JButton("TLR");
    button.addActionListener(ae->tlr());
    gbC.gridwidth = 1;
    //    jP.add(button,gbC);

    button = new JButton("RTI");
    button.addActionListener(ae->rti());
    gbC.gridwidth = 1;
    //    jP.add(button,gbC);

    progressBar = new JProgressBar(0,100);
    gbC.gridwidth = GridBagConstraints.REMAINDER;
    jP.add( progressBar, gbC );

    return jP;
  }

  public void actionPerformed( ActionEvent e )
  {
    Object ob = e.getSource();
    String command = e.getActionCommand();

    if ( command.equals("status") )
    {
      jniAdept.lock();
      JniAdeptDepp.depp.disable();
      JniAdeptDepp.djtg.enable();
      System.err.println(JniAdeptDepp.djtg.getProperties(0));
      int id = idcode();
      if (id != 0x0402e093)
      {
        double temp = xadcTemperature();
        double vccint = xadcVoltage(1);
        double vccaux = xadcVoltage(2);
        String dnaString = dna();
        statusTF.setText(String.format("DNA: %s, IDCODE: %08x, Temp[C]=%3.1f, VCCint[V]=%4.2f, VCCaux[V]=%4.2f",dnaString,id,temp,vccint,vccaux));
      }
      else
      {
        statusTF.setText(String.format("IDCODE: %08x",id));
      }
      JniAdeptDepp.djtg.disable();
      JniAdeptDepp.depp.enable();
      jniAdept.unlock();
    }
    else if ( command.equals("s6dna") )
    {
      jniAdept.lock();
      JniAdeptDepp.depp.disable();
      JniAdeptDepp.djtg.enable();

      String dnaString = s6dna();
      statusTF.setText(String.format("DNA: %s",dnaString));
	
      JniAdeptDepp.djtg.disable();
      JniAdeptDepp.depp.enable();
      jniAdept.unlock();
    }
    else if ( command.equals("choose") )
    {
      JFileChooser chooser = new JFileChooser(prefs.get("bin.dir","."));
      FileNameExtensionFilter filter = new FileNameExtensionFilter( "FPGA bin files (BIN)", "bin" );
      chooser.setFileFilter(filter);
      int returnVal = chooser.showOpenDialog(null);
      if(returnVal != JFileChooser.APPROVE_OPTION) return;
      if (!chooser.getSelectedFile().isFile()) return;
      prefs.put("bin.dir",chooser.getSelectedFile().getParent());
      prefs.put("bin.file-name",chooser.getSelectedFile().getAbsolutePath());
      binTF.setText(prefs.get("bin.file-name","pdmdb.bin"));
    }
    else if ( command.equals("load") )
    {
      pth = new ProgramThread();
      pth.start();
    }
  }
    
  public byte [] unpackData()
  {
// Now decode the responses to extract the data

    int bits = tap.length();
    if (bits < 1) return null;

    BitSet maskBS = BitSet.valueOf(tap.tdoMaskToByteBuffer());
    BitSet tdoBS =  BitSet.valueOf(tdi);
    tdoBS.and(maskBS);
    IntStream tdoIS = tdoBS.stream();
    BitSet outBS = new BitSet();
    int offset = maskBS.nextSetBit(0);
    
    tdoIS.forEach(i -> outBS.set(i-offset));

    return outBS.toByteArray();

  }
    
// 00: Temperature
// 01: VCCINT
// 02: VCCAUX
// See ug480 for more
    
  public short xadc(int ch)
  {
    tap.clear();
    tap.rti(0);
    tap.ir("110111",false); // XADC_DRP
    String drpAddr = Integer.toString((ch & 0x3ff) | 0x400,2).substring(1,11);
    tap.dr("000001"+drpAddr+"0000000000000000", true);

    JniAdeptDepp.djtg.packAndGo(tap,false);
    tdi = JniAdeptDepp.djtg.packAndGo(tap,true);

    byte [] xdata = Arrays.copyOfRange(unpackData(),0,2);

    ByteBuffer bb = ByteBuffer.wrap(xdata);
    bb.order(ByteOrder.LITTLE_ENDIAN);
    short adc = bb.getShort(0);
    return (short)((adc>>>4)&0xfff);
  }

// Get raw ADC and convert to degrees C following ug480
    
  public double xadcTemperature()
  {
    short temp = xadc(0);

    return temp*503.975/4096.0 - 273.15;
  }

// Get raw ADC and convert to volts following ug480
    
  public double xadcVoltage(int ch)
  {
    short volt = xadc(ch);

    return volt*3.0/4096.0;
  }

  public int idcode()
  {
    tap.clear();
    tap.rti(0);
    tap.ir("001001",false); // IDCODE
    tap.dr("00000000000000000000000000000000", true);

    tdi = JniAdeptDepp.djtg.packAndGo(tap,true);

    byte [] data = Arrays.copyOfRange(unpackData(),0,4);

    ByteBuffer bb = ByteBuffer.wrap(data);
    bb.order(ByteOrder.LITTLE_ENDIAN);
    return bb.getInt();
  }

// Read the device DNA
    
  public String dna()
  {
    fuseDnaSeries7();
    tdi = JniAdeptDepp.djtg.packAndGo(tap,true);

    byte [] idbuf = unpackData();

    if (idbuf == null) return "Unknown";

    BitSet dr = BitSet.valueOf(idbuf);
    BitSet drSwapped = new BitSet();
    for (int bit=0;bit<57;bit++) if (dr.get(56-bit)) drSwapped.set(bit);
    byte [] dna = Arrays.copyOfRange(drSwapped.toByteArray(),0,57);
    StringBuffer idSB = new StringBuffer("");
    for (byte b: dna) idSB.insert(0,String.format("%1$02x",b&0xff));

    return idSB.toString();

  }

  public String s6dna()
  {
    jshutdown();
    iscDnaSpartan6();
    tdi = JniAdeptDepp.djtg.packAndGo(tap,true);

    byte [] idbuf = unpackData();

    if (idbuf == null) return "Unknown";

    BitSet dr = BitSet.valueOf(idbuf);
    BitSet drSwapped = new BitSet();
    for (int bit=0;bit<57;bit++) if (dr.get(56-bit)) drSwapped.set(bit);
    byte [] dna = Arrays.copyOfRange(drSwapped.toByteArray(),0,8);
    StringBuffer idSB = new StringBuffer("");
    for (byte b: dna) idSB.insert(0,String.format("%1$02x",b&0xff));

    return idSB.toString();

  }

// FUSE_DNA is accessible while the device is running
  public void fuseDnaSeries7()
  {
    tap.clear();
    tap.rti(0);
    tap.ir("110010",true); // FUSE_DNA
    tap.dr("0000000000000000000000000000000000000000000000000000000000000000", true);
  }
        
// XSC_DNA (0x17) requires that ISC_ENABLE (0x10) is loaded first.
// Follow with ISC_DISABLE (0x16)
  public void iscDnaSeries7()
  {
    tap.clear();
    tap.rti(0);
    tap.ir("010000",true); // ISC_ENABLE
    tap.ir("010111",false); // XSC_DNA
    tap.dr("000000000000000000000000000000000000000000000000000000000", true);
    tap.ir("010110",true); // ISC_DISABLE
  }
    
// ISC_DNA (0x30) requires that ISC_ENABLE (0x10) is loaded first.
// Follow with ISC_DISABLE (0x16)
  public void iscDnaSpartan6()
  {
    tap.clear();
    tap.rti(0);
    tap.ir("010000",true); // ISC_ENABLE
    tap.ir("110000",false); // ISC_DNA
    tap.dr("000000000000000000000000000000000000000000000000000000000", true);
    tap.ir("010110",true); // ISC_DISABLE
  }
    
  public void jstart()
  {
    tap.clear();
    tap.rti(0);
    tap.ir("001100",true); // JSTART

    JniAdeptDepp.djtg.packAndGo(tap,false);

// 2000 startup clocks
    
    JniAdeptDepp.djtg.clock(false,false,2000);

    System.out.println("jstart(): Done.");
  }
    
// Move to the TLR state (from anywhere)
// TMS: 5 or more 1's take us to TLR

  public void tlr()
  {
    JniAdeptDepp.djtg.clock(true,false,16);
    System.out.println("tlr(): Done.");
  }
        

// Move to the RTI state (from anywhere)
// TMS: 5 or more 1's take us to TLR then 0 to go to RTI

  public void rti()
  {
    tlr();
    JniAdeptDepp.djtg.clock(false,false,1);
    System.out.println("rti(): Done.");
  }
        


// JSHUTDOWN followed by clocks in RTI
// Series7 requires >12 clocks to complete shutdown
// Spartan6 requires 24 clocks to complete shutdown
    
  public void jshutdown()
  {

    tap.clear();
    tap.rti(0);
    tap.ir("001101",true); // JSHUTDOWN
    tap.clock(32);

    JniAdeptDepp.djtg.packAndGo(tap,false);

    System.out.println("jshutdown(): Done.");
    
  }
        
// JPROGRAM
    
  public void jprogram()
  {

    tap.clear();
    tap.rti(0);
    tap.ir("001011",true); // JPROGRAM

    JniAdeptDepp.djtg.packAndGo(tap,false);

    System.out.println("jprogram(): Done.");
    
  }
        
// Program the FPGA.
// The CFG_IN data are shifted MSB first while instructions are LSB first.
    
  public void program()
  {

    progressBar.setValue(0);

    jniAdept.lock();
    JniAdeptDepp.depp.disable();
    JniAdeptDepp.djtg.enable();

    File f = new File(prefs.get("bin.file-name","pdmdb.bin"));
    InputStream fi = null;
    try { fi = Files.newInputStream(f.toPath()); } catch (IOException ioe){;}

    int totalPages = ((int)f.length())/txPageSize;

    //    jshutdown(); // JSHUTDOWN instruction
    jprogram(); // JPROGRAM instruction

// Move the TAP to the CFG_IN DR state. Right hand bit shifted first.

    tap.clear();
    tap.ir("000101",false);
    tap.dr(null,false); // Should be in SHIFT-DR after this

    JniAdeptDepp.djtg.packAndGo(tap,false);

// Shift pagewise.

    int page = 0; // A page is a group of commands sent as one transaction

    byte [] buf = new byte [txPageSize];
    ByteBuffer rbb = ByteBuffer.wrap(buf);
    byte [] tdo = new byte [txPageSize];

    int rd = 0;
    try
    {
    do
    {
      rd = fi.read(buf);
      for (int j=0;j<txPageSize;j++)
      {
	// Convert to unsigned
	int ub = buf[j]>=0 ? buf[j] : 256 + buf[j];
	// Bitswap each configuration data byte
	tdo[j] = swap[ub];
      }
      
      page++;

// Trigger the transaction every txPages pages
      if (page%txPages == 0)
      {
	JniAdeptDepp.djtg.putTdi(8*txPageSize,tdo,false);
// Show progress every now and then
//        if (page%(500*txPages)==0) System.out.println("page="+page);
        if (page%(100*txPages)==0) progressBar.setValue(100*page/totalPages);
      }
    } while (rd == txPageSize);
    } catch (IOException ioe){;}

// After shifting in config data we go back to RTI
// However we first stay in SHIFT-DR for 32 clocks which allows for any bypassed devices in the chain. Don't expect to be more than 126 of these. 

    byte [] tms = {(byte)0x00,(byte)0x00,(byte)0x00,(byte)0x60};
    JniAdeptDepp.djtg.putTms(128,tms,false);

// Now do the startup procedure
    jstart();
    tlr();
    
    JniAdeptDepp.djtg.disable();
    JniAdeptDepp.depp.enable();
    jniAdept.unlock();

    progressBar.setValue(100);
    System.out.println("program(): Done.");
  }
    
  public class ProgramThread extends Thread
  {
    public void run()
    {
      program();
    }
  }

ActionListener exitAL = new ActionListener () { public void actionPerformed(ActionEvent e)
{
  //   depp.close();
   System.exit(0);
   return;
}
};

  public JMenu makeMainMenu(String title)
  {
    JMenu menuBarMenu = new JMenu(title);

// Add some standard DEPP menuitems
    JniAdeptDepp.addMenuItems(menuBarMenu);

// Add GUI Import/Export menuitems
    GuiPreferences.addMenuItems(menuBarMenu);

    menuBarMenu.addSeparator();

    JMenuItem menuItem = new JMenuItem("Exit");
    menuItem.addActionListener( exitAL );
    menuBarMenu.add(menuItem);
    
    return menuBarMenu;
  }

public static void main( String[] args ) {

// The pause here is for the splash screen which closes as soon as the first swing window is made visible.

  try
  {
    Thread.sleep(1000);
  }
  catch ( InterruptedException inte )
  {
    System.err.println("Warning: " + inte.getMessage() );
  }

// Force SwingApplet to come up in the System L&F

//  String laf = UIManager.getSystemLookAndFeelClassName();
  String laf = UIManager.getCrossPlatformLookAndFeelClassName();
  try 
  {
    UIManager.setLookAndFeel(laf);
  }
  catch (UnsupportedLookAndFeelException exc)
  {
    System.err.println("Warning: UnsupportedLookAndFeel: " + laf);
  }
  catch (Exception exc)
  {
    System.err.println("Error loading " + laf + ": " + exc);
  }

// Minor tweaks to the LAF
  
  UIManager.put("Button.background",new Color(220,220,220));
  UIManager.put("TextField.background",new Color(240,240,240));
  UIManager.put("TextArea.background",new Color(240,240,240));
  UIManager.put("Table.background",new Color(240,240,240));
  
  XilinxDjtg csr = new XilinxDjtg(1,0);
  csr.setConnectedDevice();
  JFrame frame = GuiJPanel.makeFrame(csr.makePanel(),"XilinxDjtg");
  frame.getJMenuBar().add(csr.makeMainMenu("Main"));
  frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
  frame.pack();
  frame.setVisible(true);

  return;
}

}
